/* -*- Mode: C -*-
 * Lift simulation program
 * based on lifts.t of the Tempura programs collection
 * however after 5 rewrites this program
 * has no resemblance anymore with lifts.t
 */

load "../library/conversion".
load "../library/tcl".
/* tcl lifts-simple-paper */
set print_states=false. 
set verbosity=1.

define nlift  = 3. /* number of lifts  */
define nfloor = 6. /* number of floors */

define s =  0. /* stop */
define u =  1. /* up   */
define d = -1. /* down */

define c = 0. /* close */
define o = 1. /* open */



define ev(x,X) = {
   len (x-1) and keep{X=false};
   skip and X=true
}.

define keep_stable(X) = exists x:{x=X and keep {X=x}}.

define array(L,n) = {
  list(L,n) and stable(struct(L))
}.

define tassign(l,X,x) = {
  if l=1 then {skip and X:=x}
                      else { {len l-1 and stable{X}};
                             {skip and X:=x}
                           }
}.

/* Status[i] : status of lift i, 
          [s,o] : stop and doors are open,
          [u,o] : direction is up and doors are open,
          [u,c] : direction is up and doors are closed,
          [d,o] : direction is down and doors are open,
          [d,c] : direction is down and doors are closed,
   Position[i]           : positive integer, position of lift i,
   GotoReq[i][j]         : boolean, button for floor j in lift i is pressed,
   GotoReqLight[i][j]    : boolean, light for a call for lift i to floor j,
   GotoPrevReq[i][j]     : boolean, previous value of GotoReqLight[i][j],
   CallReqUp[j]          : boolean, up button on floor j is pressed,
   CallReqUpLight[j]     : boolean, light for an up call from floor j,  
   CallUp[j]             : boolean, previous value of CallReqUpLight[j],
   CallReqDown[j]        : boolean, down button on floor j is pressed,
   CallReqDownLight[j]   : boolean, light for an down call from floor j,
   CallDown[j]           : boolean, previous value of CallReqDownLight[j],
   DoorOpen[i]           : boolean, "open door" button in lift i is pressed
*/

/* run */ define lift_main() = {
 exists Status,Position,GotoReq,GotoReqLight,GotoPrevReq,CallReqUp,
       CallUp,CallReqUpLight,CallReqDown,CallDown,
       CallReqDownLight,DoorOpen : {

define main_control() = {
    array(Status,       nlift) and
    array(Position,     nlift) and
    array(GotoReq,      nlift) and 
    array(GotoReqLight, nlift) and 
    array(DoorOpen,     nlift) and
    array(GotoPrevReq,  nlift) and
    array(CallReqUp,        nfloor) and
    array(CallUp,           nfloor) and
    array(CallReqUpLight,   nfloor) and
    array(CallReqDown,      nfloor) and
    array(CallDown,         nfloor) and
    array(CallReqDownLight, nfloor) and
    
    forall i<nfloor : {
     CallUp[i] = false and
     CallDown[i] = false and
     CallReqUpLight[i]   is (CallUp[i]   or CallReqUp[i]  )  and
     CallReqDownLight[i] is (CallDown[i] or CallReqDown[i]) and 
     CallUp[i] gets 
      (CallReqUpLight[i]   and 
        ~exists j<nlift : (Position[j]=i and Status[j][0]=u)) and
     CallDown[i] gets 
      (CallReqDownLight[i] and 
        ~exists j<nlift : (Position[j]=i and Status[j][0]=d))
    } 
    
  } and
  
  
		     
define lift(i) = {
    

define automaton(i) = {      
      while true do {
       if Status[i] = [s,o] then {
        skip and stable Position[i] and 
        if called_up(i) or called_down(i) then {
          if  called_up(i)   then {
            Status[i]:=if DoorOpen[i] then [u,o] else [u,c]
          } else { 
             Status[i]:=if DoorOpen[i] then [d,o] else [d,c]
            }  
         } else Status[i]:=[s,o]
       } and 
       if Status[i] = [u,c]  then {
         skip and 
         if stop_up(i) or ~called_up(i) then {
           stable Position[i] and 
           if stop_up(i) then {
             if called_up(i) then Status[i]:=[u,o]
                             else {if called_down(i) then Status[i]:=[d,o]
                                                     else Status[i]:=[s,o]
                                }
           } else Status[i]:=[d,c]
          } else {  
              skip and Status[i]:=[u,c] and 
              Position[i] := Position[i]+1
            }
       } and
       if Status[i] = [d,c]  then {
         skip and 
         if stop_down(i) or ~called_down(i) then {
           stable Position[i] and 
           if stop_down(i) then {
             if called_down(i) then Status[i]:=[d,o]
                               else {if called_up(i) then Status[i]:=[u,o]
                                                     else Status[i]:=[s,o]
                                    } 
           } else Status[i]:=[u,c] 
         } else {
            skip and Status[i]:=[d,c] and
            Position[i] := Position[i]-1
          }
       } and
       if Status[i] = [u,o]  then {
        {skip and  stable{Status[i]} and  stable{Position[i]} };
        {skip and stable Position[i] and 
         if called_up(i) then {
          Status[i]:=if DoorOpen[i] then [u,o] else [u,c]
         }
           else {
            if called_down(i) then {
             Status[i]:=if DoorOpen[i] then [d,o] else [d,c]
            } else Status[i]:=[s,o]
           }
        }
       } and
       if Status[i] = [d,o]  then {
        {skip and  stable{Status[i]} and  stable{Position[i]} };
        {skip and stable Position[i] and 
         if called_down(i) then {
          Status[i]:=if DoorOpen[i] then [d,o] else [d,c]
         }
           else { 
            if called_up(i) then { 
             Status[i]:=if DoorOpen[i] then [u,o] else [u,c]
            } else Status[i]:=[s,o]
           }
        }
       }
     }
    } and

define called(i) = {
       calls(i,0,nfloor) 
    } and
    
define called_up(i) = {
       calls(i,Position[i]+1,nfloor) 
    } and
    
define called_down(i) = {
     calls(i,0,Position[i])  
    } and

define calls(i,x,y) = {
     exists j<y-x:( GotoReqLight[i][x+j] or 
                    CallReqUpLight[x+j] or CallReqDownLight[x+j]) 
    } and

define stop_up(i) = {
     (GotoReqLight[i][Position[i]] or 
      CallReqUpLight[Position[i]]) 
    } and

define stop_down(i) = {
     (GotoReqLight[i][Position[i]] or  
      CallReqDownLight[Position[i]]) 
    } and

    Position[i] = 0 and
    array(GotoReq[i],nfloor) and
    array(GotoReqLight[i],nfloor) and
    array(GotoPrevReq[i],nfloor) and
    forall j<nfloor : {
      GotoPrevReq[i][j] = false and
      GotoReqLight[i][j] is  (GotoPrevReq[i][j]  or GotoReq[i][j]) and
      GotoPrevReq[i][j] gets (GotoReqLight[i][j] and ~(Position[i]=j))     
    } and
    Status[i] = [s,o] and 
    automaton(i)
  } and

  
define display_x() = {
    if Status=Status and GotoReqLight=GotoReqLight and 
       GotoPrevReq=GotoPrevReq and CallDown=CallDown and
       CallReqDownLight=CallReqDownLight and 
       CallUp=CallUp and DoorOpen=DoorOpen and
       CallReqUpLight=CallReqUpLight and Position=Position   
    then {   
      forall l<nlift : {
         tcl("lift",[l,Position[l],Status[l][1]])
      } and
      forall j<nfloor : {
         tcl("floor",[j,CallReqUpLight[j],CallReqDownLight[j]])
      } and
      forall j<nfloor : {
       forall l<nlift : {
         tcl("button",[l,j,GotoReqLight[l][j]])
       }
      } and
      forall l<nlift : {
         tcl("lstop",[l,DoorOpen[l]])
      }
    }
  } and

   
     forall i<nfloor : {
      if i<nfloor-1 then {
       while true do {
       ev(7+Random mod 5,   CallReqUp[i])
       }
      } else CallReqUp[i] is false 
     }  and
   
   
   
     forall i<nfloor : {
      if i>0 then {
       while true do {
       ev(7+Random mod 7, CallReqDown[i])
       }
      } else CallReqDown[i] is false 
     } and

     forall j<nlift : {
       while true do {
       /* {len j+2 and DoorOpen[j] is true };skip;*/
        {len 7+j and DoorOpen[j] is false };skip
        
       }
     } and 
   
   
     forall i<nfloor : {
      forall j<nlift : {
       while true do {
       ev(7+ Random mod 9,    GotoReq[j][i])
     }
    }
   } and
 
  len 150 and 
  main_control() and
  forall i<nlift : { lift(i) } and
  tcl("init",[nlift,nfloor]) and always tclbreak() and
  always display_x()
}
}.




