/* -*- Mode: C -*-
 * Lift simulation program                          
 * based on lifts.t of the Tempura programs collection
 * however after 5 rewrites this program
 * has no resemblance with lifts.t anymore
 */

set print_states=false. 
set verbosity=1.

define nlift  = 5. /* number of lifts */
define nfloor = 3. /* number of floors */
define doorClosingDuration = 2.  /* time needed to close a door */
define doorOpeningDuration = 2.  /* time needed to open a door */
define doorOpenDuration = 5.     /* required open time of a door */
define interFloorMovingTime = 2. /* time needed for a lift to move to the 
                                    next/previous floor */ 

define s =  0. /* stop */
define u =  1. /* up   */
define d =  2. /* down */

define c = 0. /* close */
define o = 1. /* open */

load "../library/conversion".
load "../library/tcl".
/* tcl lifts5-paper */

define ev(x,X) = {
   len (x-1) and keep{X=false};
   skip and X=true
}.

define keep_stable(X) = exists x:{x=X and keep {X=x}}.

define array(L,n) = {
  list(L,n) and stable(struct(L))
}.

define tassign(l,X,x) = {
  if l=1 then {skip and X:=x}
                      else { {len l-1 and stable{X}};
                             {skip and X:=x}
                           }
}.


define tassign_i(l,X,x,Y,y) = {
  if l=1 then {skip and X:=if Y then y else x}
         else {
          if Y then {skip and X:=y}
               else {
                {skip and X:=x};tassign_i(l-1,X,x,Y,y)
               }
         }
}.  


/* Status[i] : status of lift i, 
          [s,o] : stop and doors are open,
          [u,o] : direction is up and doors are open,
          [u,c] : direction is up and doors are closed,
          [d,o] : direction is down and doors are open,
          [d,c] : direction is down and doors are closed,
   Position[i]           : positive integer, position of lift i,
   GotoReq[i][j]         : boolean, button for floor j in lift i is pressed,
   GotoReqLight[i][j]    : boolean, a call for lift i to floor j (light),
   GotoPrevReq[i][j]     : boolean, previous value of GotoReqLight[i][j],
   CallReqUp[j]          : boolean, up button on floor j is pressed,
   CallReqUpLight[j]     : boolean, an up call from floor j (light),  
   CallUp[j]             : boolean, previous value of CallReqUpLight[j],
   CallReqDown[j]        : boolean, down button on floor j is pressed,
   CallReqDownLight[j]   : boolean, an down call from floor j (light),
   CallDown[j]           : boolean, previous value of CallReqDownLight[j],
   DoorOpen[i]           : boolean, "open door" button in lift i is pressed
*/

/* run */ define lift_main() = {
 exists Status,Position,T,GotoReq,GotoReqLight,GotoPrevReq,CallReqUp,
        CallUp,CallReqUpLight,CallReqDown,CallDown,
        CallReqDownLight,DoorOpen : {

define main_control() = {
    array(Status,       nlift) and
    array(Position,     nlift) and
    array(T,            nlift) and
    array(GotoReq,      nlift) and 
    array(GotoReqLight, nlift) and 
    array(DoorOpen,     nlift) and
    array(GotoPrevReq,  nlift) and
    array(CallReqUp,        nfloor) and
    array(CallUp,           nfloor) and
    array(CallReqUpLight,   nfloor) and
    array(CallReqDown,      nfloor) and
    array(CallDown,         nfloor) and
    array(CallReqDownLight, nfloor) and
    
    forall i<nfloor : {
     CallUp[i] = false and
     CallDown[i] = false and
     CallReqUpLight[i]   is (CallUp[i]   or CallReqUp[i]  )  and
     CallReqDownLight[i] is (CallDown[i] or CallReqDown[i])  and 
     CallUp[i] gets 
       (CallReqUpLight[i] and 
         ~exists j<nlift : ( T[j]=interFloorMovingTime-1 and 
                             Position[j]=i and Status[j][0]=u)) and
     CallDown[i] gets 
       (CallReqDownLight[i] and 
         ~exists j<nlift : ( T[j]=interFloorMovingTime-1 and 
                             Position[j]=i and Status[j][0]=d))
    } 
    
  } and
  
  
		     
define lift(i) = {
    

define automaton(i) = {      
      while true do {
      if T[i]=interFloorMovingTime-1 then {
       if Status[i] = [s,o] then {
         stable Position[i] and stable{T[i]} and
         if called_up(i) or called_down(i) then {
          if  called_up(i)  then {
            tassign_i(doorClosingDuration,Status[i],[u,c],DoorOpen[i],[u,o])
          } else {
             tassign_i(doorClosingDuration,Status[i],[d,c],DoorOpen[i],[d,o])
            }  
        } else {
           skip and Status[i]:=[s,o]
          }
       } and 
       if Status[i] = [u,c]  then {
         if stop_up(i) or ~called_up(i)  then {
          stable Position[i] and stable{T[i]} and 
          if stop_up(i) then {
           if called_up(i) 
           then tassign(doorOpeningDuration,Status[i],[u,o])
           else { 
            if called_down(i) 
            then tassign(doorOpeningDuration,Status[i],[d,o])
            else tassign(doorOpeningDuration,Status[i],[s,o])
           }
         } else {
            skip and Status[i]:=[d,c]
           }
          } else {
             len interFloorMovingTime and Status[i] gets [u,c] and
             T[i] gets (T[i]-1) mod interFloorMovingTime and
             Position[i] gets Position[i]+ if T[i]=0 then 1 else 0
            }
       } and
       if Status[i] = [d,c]  then {
         if stop_down(i) or ~called_down(i) then {
          stable Position[i] and stable{T[i]} and
          if stop_down(i) then {
           if called_down(i) 
           then tassign(doorOpeningDuration,Status[i],[d,o])
           else { 
            if called_up(i) 
            then tassign(doorOpeningDuration,Status[i],[u,o])
            else tassign(doorOpeningDuration,Status[i],[s,o])
           }
          } else {
            skip and Status[i]:=[u,c]
            }
         } else {
            len interFloorMovingTime and Status[i] gets [d,c] and
            T[i] gets (T[i]-1) mod interFloorMovingTime and
            Position[i] gets Position[i]- if T[i]=0 then 1 else 0
           }
       } and
       if Status[i] = [u,o]  then {
        {len doorOpenDuration and  stable{Status[i]} and 
         stable{T[i]} and stable{Position[i]} };
        {stable Position[i] and stable{T[i]} and
         if called_up(i) then {
           tassign_i(doorClosingDuration,Status[i],[u,c],DoorOpen[i],[u,o])
          }
           else {
            if called_down(i) then {
             tassign_i(doorClosingDuration,Status[i],[d,c],DoorOpen[i],[d,o])
            } else {skip and Status[i]:=[s,o]}
           }
        }
       } and
       if Status[i] = [d,o]  then {
        {len doorOpenDuration and  stable{Status[i]} and 
         stable{T[i]} and stable{Position[i]} };
        {stable Position[i] and stable{T[i]} and
         if called_down(i) then {
           tassign_i(doorClosingDuration,Status[i],[d,c],DoorOpen[i],[d,o])
         }
          else {
           if called_up(i) then {
            tassign_i(doorClosingDuration,Status[i],[u,c],DoorOpen[i],[u,o]) 
           } else {skip and Status[i]:=[s,o]}
          }
        }
       }
      } else {
        skip and format("error") and
        stable{T[i]} and stable{Status[i]} and stable{Position[i]}
       }
     }
    } and

define called(i) = {
       calls(i,0,nfloor) 
    } and
    
define called_up(i) = {
       calls(i,Position[i]+1,nfloor) 
    } and
    
define called_down(i) = {
     calls(i,0,Position[i])  
    } and

define calls(i,x,y) = {
      exists j<y-x:( GotoReqLight[i][x+j] or 
                     (T[i]=interFloorMovingTime-1 and CallReqUpLight[x+j]) or  
                     (T[i]=interFloorMovingTime-1 and CallReqDownLight[x+j])
                   ) 
    } and

define stop_up(i) = {
     (GotoReqLight[i][Position[i]] or 
      CallReqUpLight[Position[i]] 
     )  
    } and

define stop_down(i) = {
     (GotoReqLight[i][Position[i]] or  
      CallReqDownLight[Position[i]] 
     )   
    } and

    Position[i] = 0 and
    array(GotoReq[i],nfloor) and
    array(GotoReqLight[i],nfloor) and
    array(GotoPrevReq[i],nfloor) and
    forall j<nfloor : {
      GotoPrevReq[i][j] = false and
      GotoReqLight[i][j] is  (GotoPrevReq[i][j]  or GotoReq[i][j]) and
      GotoPrevReq[i][j] gets 
       (GotoReqLight[i][j] and ~(T[i]=interFloorMovingTime-1 and Position[i]=j))     
    } and
    Status[i] = [s,o] and T[i]=interFloorMovingTime-1 and
    automaton(i)
  } and

define display_x() = {
    if Status=Status and GotoReqLight=GotoReqLight and 
       GotoPrevReq=GotoPrevReq and CallDown=CallDown and
       CallReqDownLight=CallReqDownLight and 
       CallUp=CallUp and DoorOpen=DoorOpen  and
       CallReqUpLight=CallReqUpLight and Position=Position 
    then {   
      forall l<nlift : {
         tcl("lift",[l,Position[l],Status[l][1],T[l],Status[l][0]]) and
         tcl("lstop",[l,DoorOpen[l]]) and
         tcl("dir",[l,Status[l][0]]) 
      } and
      forall j<nfloor : {
         tcl("floor",[j,CallReqUpLight[j],CallReqDownLight[j]])
      } and
      forall j<nfloor : {
       forall l<nlift : {
         tcl("button",[l,j,GotoReqLight[l][j]])
       }
      }   
    }
  } and

  


    
   
     forall i<nfloor : {
      if i<nfloor-1 then {
       while true do {
       ev(7+ Random mod 5,   CallReqUp[i])
       }
      } else CallReqUp[i] is false 
     }  and
   
   
   
     forall i<nfloor : {
      if i>0 then {
       while true do {
       ev(7+Random mod 7, CallReqDown[i])
       }
      } else CallReqDown[i] is false 
     } and
   
   forall j<nlift : {
       while true do {
       /* {len j+2 and DoorOpen[j] is true };skip;*/
        {len 7+j and DoorOpen[j] is false };skip
        
       }
     } and

     forall i<nfloor : {
      forall j<nlift : {
       while true do {
       ev(7+Random mod 6,    GotoReq[j][i])
     }
    }
   } and
 
  len 250 and 
  main_control() and
  forall i<nlift : lift(i) and
  tcl("init",[nlift,nfloor,interFloorMovingTime]) and 
  always tclbreak() and 
  always display_x()
}
}.



