/* -*- Mode: C -*-
    This the version without default value nor delegation

   (c) 11-12/05/03 -- F. Siewe
*/

/* Constants */
define simulationTime = 10.
define nbSubject = 4.
define nbObject = 2.
define nbAction = 3.

define bob = 2.
define alice = 3.
define manager = 0.
define developer = 1.

define managerDir = 0.
define developerDir = 1.

define read = 0.
define write = 1.
define execute = 2.

define owner = [manager, developer].

/* groups/roles must come before individuals (in numbering)
*/
define role = [[],[],[manager,developer], [developer]].
define actionObj = [[read,write,execute], [read, write,execute]].
define subjectName = ["Manager", "Developer", "Bob", "Alice"].
define objectName = ["/home/manager", "/home/developer"].
define actionName = ["read", "write", "execute"].
define message = ["permitted", "denied", "accepted", "rejected"].

/* Miscellaneous */
define array(L,n) = {
    List(L,n) and stable(struct(L))
}.

define clock(Timer, startTime, rate) = {
   Timer= startTime and Timer gets Timer + rate
}.

define member(x,l,i) = if i>=|l| then false 
                     else if l[i]=x then true
                     else member(x,l,i+1).

define index(s,o,a) = ((s * nbObject +o) * nbAction +a) . 

define setSize(x, a, n) = if type(x)="string" then
                         if |x|>=n then x[0..n]
                         else setSize(x+a,a, n)
                      else x.
  
define out(s1, o1, a1, x) = {
   exists s, o, a, status:{
      s = setSize(subjectName[s1]," ", 10) and o = setSize(objectName[o1]," ", 16) and 
      a = setSize(actionName[a1]," ", 8) and status = message[x] and
      format("%s %s %s %s\n", s, o, a, status)
   }
}.

define bool2int(x) = if x=true then 0 else 1.

/* Rules */
define roleAutho(s,o, a, i, X) = if s>=|role| then false
                               else if i>=|role[s]| then false
                               else if X[index(role[s][i], o, a)]=true
                               then true
                               else  roleAutho(s,o, a, i+1, X).

 
/* Owner can access*/  
define rule1(AuthoP, AuthoN) = {
   always{
     forall s<nbSubject:{
       forall o<nbObject:{
         forall i<|actionObj[o]|:{
            if (owner[o]=s or roleAutho(s,o,actionObj[o][i],0, AuthoP)) then 
               AuthoP[index(s,o,actionObj[o][i])] = true
            else
               AuthoP[index(s,o,actionObj[o][i])] = false           
         }
       }
     }
   }
}.

define rule2(AuthoP, AuthoN) = {
   always{
     forall s<nbSubject:{
       forall  o<nbObject:{
         forall i<|actionObj[o]|:{
            if roleAutho(s,o,actionObj[o][i],0, AuthoN) then 
               AuthoN[index(s,o,actionObj[o][i])] = true
            else
               AuthoN[index(s,o,actionObj[o][i])] = false
         }
       }
     }
   }
}.


/* Decision rules */

/* Closed policy
*/
define closed(AuthoP, AuthoN, Autho) = {
   always{
     forall s<nbSubject:{
       forall o<nbObject:{
         forall i<|actionObj[o]|:{
           if AuthoP[index(s,o,actionObj[o][i])] = true then 
             Autho[index(s,o,actionObj[o][i])] = true
           else
             Autho[index(s,o,actionObj[o][i])] = false
         }
       }
     }
   }   
}.

/* Open policy
*/
define open(AuthoP, AuthoN, Autho) = {
   always{
     forall s<nbSubject:{
       forall o<nbObject:{
         forall i<|actionObj[o]|:{
           if AuthoN[index(s,o,actionObj[o][i])] ~= true then 
             Autho[index(s,o,actionObj[o][i])] = true
           else
             Autho[index(s,o,actionObj[o][i])] = false
         }
       }
     }
   }   
}.

/* Hybrid with denials take precedence
*/
define denialTakePrec(AuthoP, AuthoN, Autho) = {
   always{
     forall s<nbSubject:{
       forall o<nbObject:{
         forall i<|actionObj[o]|:{
           if AuthoP[index(s,o,actionObj[o][i])] = true and 
              AuthoN[index(s,o,actionObj[o][i])] ~= true then 
              Autho[index(s,o,actionObj[o][i])] = true
           else
              Autho[index(s,o,actionObj[o][i])] = false
         }
       }
     }
   }   
}.

/* Hybrid with no conflict allowed
*/
define conflictNotAllowed(AuthoP, AuthoN, Autho, Error) = {
   always{
     forall s<nbSubject:{
       forall o<nbObject:{
         forall i<|actionObj[o]|:{
           if AuthoP[index(s,o,actionObj[o][i])] = true and 
              AuthoN[index(s,o,actionObj[o][i])] ~= true then 
              Error[index(s,o,actionObj[o][i])] = true
           else
             Error[index(s,o,actionObj[o][i])] = false and 
             if AuthoN[index(s,o,actionObj[o][i])] = true then 
                Autho[index(s,o,actionObj[o][i])] = true
             else
                Autho[index(s,o,actionObj[o][i])] = false
         }
       }
     }
   }   
}.

/* Policies */
define policy1(AuthoP, AuthoN, Autho) = {
   rule1(AuthoP, AuthoN) and 
   rule2(AuthoP, AuthoN) and
   closed(AuthoP, AuthoN, Autho) 
}.
define policy2(AuthoP, AuthoN, Autho) = {
   rule1(AuthoP, AuthoN) and 
   rule2(AuthoP, AuthoN) and
   open(AuthoP, AuthoN, Autho)
}.

define policy3(AuthoP, AuthoN, Autho) = {
   rule1(AuthoP, AuthoN) and 
   rule2(AuthoP, AuthoN) and
   denialTakePrec(AuthoP, AuthoN, Autho)
}.

define acp(Autho) = {
   exists AuthoP, AuthoN:{
      array(AuthoP, nbSubject*nbObject*nbAction) and 
      array(AuthoN, nbSubject*nbObject*nbAction) and

      {policy1(AuthoP, AuthoN, Autho) and len(simulationTime/2);skip;
      policy1(AuthoP, AuthoN, Autho) and len(simulationTime/2-1)}   
   }
}.

/* Reference monitor */
define rm(Request, Autho) = {
   format("%s\n", setSize("-","-",48)) and
   format("%s %s %s %s\n", setSize("Subjects"," ", 10), setSize("Objects"," ", 16), setSize("Actions"," ", 8), "Request ...") and 
   format("%s\n", setSize("-","-",48)) and 
   chopstar{
      skip and forall s<nbSubject:{
         forall o<nbObject:{
            forall i<|actionObj[o]|:{
               if Request[index(s,o,actionObj[o][i])] = true then{
                  if Autho[index(s,o,actionObj[o][i])] = true then 
                     out(s,o,actionObj[o][i],2)
                  else
                     out(s,o,actionObj[o][i],3)
               } 
            }
         }
      }
   }

}.


define acm(Request, Autho) = {
   format("%s\n", setSize("-","-",50)) and
   format("%s %s %s %s\n", setSize("Subjects"," ", 10), setSize("Objects"," ", 16), setSize("Actions"," ", 8), "Authorization") and 
   format("%s\n", setSize("-","-",50)) and 
   chopstar{
      skip and forall s<nbSubject:{
         forall o<nbObject:{
            forall i<|actionObj[o]|:{
               out(s,o,actionObj[o][i], bool2int(Autho[index(s,o,actionObj[o][i])]))                 
            }
         }
      }
   }

}.

/* Behaviour of Subjects */
/* Bob requests to read managerDir every 3 time units, carries out 
   a local computation, and then requests to write in 
   developerDir every 7 time units. 
*/
define bobReq(Request, Timer) = {
   always{ 
      forall i < |actionObj[managerDir]|:{
         if actionObj[managerDir][i]=read and Timer mod 3 =0 then 
            Request[index(bob,managerDir,actionObj[managerDir][i])]=true
         else 
            Request[index(bob,managerDir,actionObj[managerDir][i])]=false
      }
   } and
   always{ 
      forall i < |actionObj[developerDir]|:{
         if actionObj[developerDir][i]=write and Timer mod 7 =1 then 
            Request[index(bob,developerDir,actionObj[developerDir][i])]=true
         else 
            Request[index(bob,developerDir,actionObj[developerDir][i])]=false
      }
   }
        
}.

/* Alice randomly requests to write in managerDir or developerDir.
*/

define aliceReq(Request, Timer) = {
   always{ 
      forall i < |actionObj[managerDir]|:{
         if actionObj[managerDir][i]=write and Timer >= Random mod simulationTime then 
            Request[index(alice,managerDir,actionObj[managerDir][i])]=true
         else 
            Request[index(alice,managerDir,actionObj[managerDir][i])]=false
      }
   } and
   always{ 
      forall i < |actionObj[developerDir]|:{
         if actionObj[developerDir][i]=write and Timer>=random mod simulationTime then 
            Request[index(alice,developerDir,actionObj[developerDir][i])]=true
         else 
            Request[index(alice,developerDir,actionObj[developerDir][i])]=false
      }
   }
        
}.

/* No request from manager 
*/
define managerReq(Request) = {
   always{
      forall o<nbObject:{
         forall i<|actionObj[o]|:{  
            Request[index(manager,o,actionObj[o][i])]=false
         }
      }
   }
}.

/* No request from developer 
*/
define developerReq(Request) = {
   always{
      forall o<nbObject:{
         forall i<|actionObj[o]|:{  
            Request[index(developer,o,actionObj[o][i])]=false
         }
      }
   }
}.

define users(Request, Timer) = {
   bobReq(Request, Timer) and 
   aliceReq(Request, Timer) and 
   managerReq(Request) and 
   developerReq(Request)
}.



/* Simulation */
/* run */ define test() = {
   exists Autho, Request, Timer:{  
      array(Autho, nbSubject*nbObject*nbAction) and 
      array(Request, nbSubject*nbObject*nbAction) and 
      set random_seed=1 and
      users(Request, Timer) and 
      
      acp(Autho) and  
      rm(Request, Autho) and 
      len(simulationTime) and 
      clock(Timer, 0, 1)
   }
}.






