/* -*- Mode: C -*-
 * Tempura interpreter for Tempura.  6 May 84;
 * Updated 24 Apr 85
 *
 *  This is intended to be a simple interpreter for Tempura
 *  and is written in Tempura itself.
 */



/* Reduce a statement */
define reduce(Env,e) = {
    if e[0] = "and" then reduce_and(Env,e[1 to |e|])
    else if e[0] = "=" then reduce_eq(Env,e[1 to |e|])
    else if e[0] = "if" then reduce_if(Env,e[1 to |e|])
    else if e[0] = "stable" then reduce_stable(Env,e[1 to |e|])
    else if e[0] = "box" then reduce_box(Env,e[1 to |e|])
    else if e[0] = ":=" then reduce_set(Env,e[1 to |e|])
    else if e[0] = "halt" then reduce_halt(Env,e[1 to |e|])
    else if e[0] = "skip" then reduce_skip(Env,e[1 to |e|])
    else if e[0] = "is" then reduce_is(Env,e[1 to |e|])
    else if e[0] = "semi" then reduce_semi(Env,e[1 to |e|])
    else if e[0] = "while" then reduce_while(Env,e[1 to |e|])
    else if e[0] = "for-times" then reduce_cycle(Env,e[1 to |e|])
    else if e[0] = "input" then reduce_input(Env,e[1 to |e|])
    else if e[0] = "output" then reduce_output(Env,e[1 to |e|])
    else format("bad statement\n")
}.

/* ["and" e1 e2 ... eN] */
define reduce_and(Env,elist) = {
    forall i<|elist| : reduce(Env,elist[i])
}.

/* ["if" B e1 e2] */
define reduce_if(Env,elist) = {
    if getval(Env,elist[0])
    then reduce(Env,elist[1])
    else reduce(Env,elist[2])
}.

/* ["=" e1 e2] */
define reduce_eq(Env,elist) = {
    Env[getref(Env,elist[0])] = getval(Env,elist[1])
}.

/* Stable values */

/* ["stable" e] */
define reduce_stable(Env,plist) = {
	true
}.


/* ***Linear-time temporal operators*** */


/* ["box" W] */
define reduce_box(Env,elist) = {
    box(reduce(Env,elist[0]))
}.
	   
	   
/* unit assignment */
/* [":=" V e] */
define reduce_set(Env,elist) = {
    Env[getref(Env,elist[0])] := getval(Env,elist[1])
}.

	   
/* ["halt" W] */
define reduce_halt(Env,elist) = {
    halt(getval(Env,elist[0]))
}.
	   
/* ["skip"] */
define reduce_skip(Env,elist) = {
    skip
}.
	   
/* ["is" V e] */
define reduce_is(Env,elist) = {
    Env[getref(Env,elist[0])] is getval(Env,elist[1])
}.



/* ***Semicolon, while, repeat-until, etc.*** */

/* ["semi" W1 W2] */
define reduce_semi(Env,elist) = {
    reduce(Env,elist[0]) ; reduce(Env,elist[1])
}.

       
	   
/* ["while" e W] */
define reduce_while(Env,elist) = {
    while getval(Env,elist[0]) do reduce(Env,elist[1])
}.

/* ["for-times" e W] */
define reduce_cycle(Env,elist) = {
    for (getval(Env,elist[0])) times do reduce(Env,elist[1])
}.

/* ["len" e] */
define reduce_len(Env,elist) = {
    len(getval(Env,elist[0]))
}.


/* ***Some features for accessing the system*** */
/* /* def sys *			/* /

/* Input a value */
/* ["input" e] */
define reduce_input(Env,elist) = {
    input(Env[getref(Env,elist[0])])
}.


/* Output a value */
/* ["output" e] */
define reduce_output(Env,elist) = {
    format("%t\n",getval(Env,elist[0]))
}.


/* Determine the value of something */
define getval(Env,e) = {
    if is_integer(e) then e
    else if is_string(e) then getval_id(Env,e)
    else getval_op(Env,e)
}.

define getval_id(Env,e) = {
    Env[getref_id(Env,e)]
}.	   

/* get the value of an operator */
define getval_op(Env,e) = {
    if e[0] = "=" then getval_eq(Env,e[1 to |e|])
    else if e[0] = "+" then getval_plus(Env,e[1 to |e|])
    else if e[0] = "if" then getval_if(Env,e[1 to |e|])
    else if e[0] = "<" then getval_lt(Env,e[1 to |e|])
    else if e[0] = "and" then getval_and(Env,e[1 to |e|])
    else if e[0] = "not" then getval_not(Env,e[1 to |e|])
    else if e[0] = "or" then getval_or(Env,e[1 to |e|])
    else format("Error\n")
}.


/* Logical operators */
	   
/* ["not" e] */
define getval_not(Env,elist) = {
    not getval(Env,elist[0])
}.	   

/* ["and" e1 e2 ... eN] */
define getval_and(Env,elist) = {
    forall i<|elist| : getval(Env,elist[i])
}.

/* ["or" e1 e2] */
define getval_or(Env,elist) = {
    getval(Env,elist[0]) or getval(Env,elist[1])
}.


/* ["if" B e1 e2] */
define getval_if(Env,elist) = {
    if getval(Env,elist[0]) 
    then getval(Env,elist[1])
    else getval(Env,elist[2])
}.


/* ["=" e1 e2] */
define getval_eq(Env,elist) = {
    getval(Env,elist[0]) = getval(Env,elist[1])
}.

/* ["<" e1 e2] */
define getval_lt(Env,elist) = {
    getval(Env,elist[0]) < getval(Env,elist[1])
}.



/* Arithmetic operators */


/* ["+" e1 e2] */
define getval_plus(Env,elist) = {
    getval(Env,elist[0]) + getval(Env,elist[1])
}.





/* Get the reference to something */

define getref(Env,id) = {
    getref_id(Env,id)
}.

define getref_id(Env,id) = {
    myref(id)
}.



/* end sys */





define is_string(S) = {
    type(S) = "string"
}.

define myref(id) = {
    if id = "X" then 0
    else if id = "Y" then 1
    else if id = "Z" then 2
    else format("Bad id to look up: %t\n",id)
}.

/* Test data. */

define boot_test(pgm) = {
    exists Env : {
	fixed_list(Env,3) and
	reduce(Env,pgm)
    }
}.

define boot_test1 =
    ["and",
	["input","X"],["input","Y"],
	["box",["output","Y"]],
	["for-times","X",
	    ["and",["skip"],[":=","Y",["+","Y","Y"]]]
	]		  
    ].

define boot_test2 =
    ["and",
	["=","X",3],["=","Y",4],
	["box",["output","Y"]],
	["for-times","X",
	    ["and",["skip"],[":=","Y",["+","Y","Y"]]]
	]		  
    ].



/* end of Tempura interpreter source */

