/* -*- Mode: C -*- 
 * Reading and writing to cyclic buffer. Roger Hale. 
 *
 */

define alter(L,I,Val) = {
    L[0..I] gets L[0..I] and
    L[I] gets Val and
    L[I+1..|L|] gets L[I+1..|L|]
}.

define buffer(Buff,n) = {
    fixed_list(Buff,n+2) and
    forall i<n+2 : {Buff[i]=0}
}.

define no_sent(Buff) = Buff[0].

define no_recd(Buff) = Buff[1].

define buff_data(Buff) = Buff[2..|Buff|].

define buff_size(Buff) = |Buff|-2.

define put(Data,Buff) = exists data : {
    data = Data and
    halt(no_sent(Buff) < no_recd(Buff)+buff_size(Buff)) and
    stable buff_data(Buff) and 
    stable no_sent(Buff);
    skip and 
    alter(buff_data(Buff),no_sent(Buff) mod buff_size(Buff),data) and
    no_sent(Buff) := no_sent(Buff)+1
}.

define no_put(Buff) = {
  stable buff_data(Buff) and stable no_sent(Buff)
}.

define get(Data,Buff) = {
    halt(no_recd(Buff) < no_sent(Buff)) and 
    stable no_recd(Buff);
    skip and 
    Data := buff_data(Buff)[no_recd(Buff) mod buff_size(Buff)] and 
    no_recd(Buff) := no_recd(Buff) + 1
}.

define no_get(Buff) = stable no_recd(Buff).

define write(data,Buff) = {
    for i<|data| do put(data[i],Buff);
    no_put(Buff)
}.

define slow_write(data,Buff) = {
    for i<|data| do {
	len(2) and no_put(Buff);
	put(data[i],Buff)
    };
    no_put(Buff)
}.

define read(data,Buff) = {
    for i<|data| do get(data[i],Buff)
}.

define slow_read(data,Buff) = {
    for i<|data| do {
	len(2) and no_get(Buff);
	get(data[i],Buff)
    }
}.

/* run */ define buff_test(input_data) = exists output_data,Buff : {
    buffer(Buff,2) and
    fixed_list(output_data,|input_data|) and {
	write(input_data[0..|input_data|/2],Buff) and
	slow_read(output_data[0..|output_data|/2],Buff);
	slow_write(input_data[|input_data|/2..|input_data|],Buff) and
	read(output_data[|output_data|/2..|output_data|],Buff)
    } and
    format("Input data = %t\n",input_data) and
    always format("Buffer = %t\n",buff_data(Buff)) 
   and fin format("Output data = %t\n",output_data)
}.

load "library".
