/* -*- Mode: C -*-
 * Two-way odd-even mergesort for 
 * linearly connected processors (Kronsjo) 
 *
 */

define noop=0.
define shiftleft=1.
define shiftright=2.
define compareleft=3.
define compareright=4.

define merge(C,Phase) = {
    if |C|=2 then {
	compare(C[0],C[1]) and Phase gets "compare"
    }
    else {
        unshuffle(C) and Phase gets "unshuffle";
	exists Phase1,Phase2 : {
	    merge(C[0..|C|/2],Phase1) and
	    merge(C[|C|/2..|C|],Phase2) and
	    wnext {Phase is Phase1+","+Phase2}
	};
        shuffle(C) and Phase gets "shuffle";
        compare_odd_even(C) and Phase gets "odd-even compare"
    }
}.

define shuffle(C) = exists n : {
    n = |C| and
    for i<n/2-1 do {
	idle(C[0..n/2-i-1]) and
	forall j<i+1 : {
	    swap(C[n/2-i+2*j-1],C[n/2-i+2*j])
	} and
	idle(C[n/2+i+1..n])
    }
}.

define unshuffle(C) = exists n : {
    n = |C| and
    for i<n/2-1 do {
	idle(C[0..i+1]) and
	forall j<n/2-i-1 : {
	    swap(C[i+2*j+1],C[i+2*j+2])
	} and
	idle(C[n-i-1..n])
    }
}.

define compare_odd_even(C) = exists n : {
    n = |C| and
    C[0] = "noop" and
    C[n-1] = "noop" and
    forall i<n/2-1 : compare(C[2*i+1],C[2*i+2])
}.

define swap(Ci,Cj) = {
    skip and Cj = "shiftright"  and Ci = "shiftleft"
}.

define compare(Ci,Cj) = {
    skip and Ci = "compareright"  and Cj = "compareleft"
}.

define idle(Cij) = {
    forall k<|Cij| : Cij[k] = "noop"
}.

define compare_net(n,C,D) = 
    forall i<n : {
	D[i] gets (if C[i]="shiftright" then D[i-1]
	           else if C[i]="shiftleft" then D[i+1]
	           else if C[i]="compareleft" then max(D[i],D[i-1])
	           else if C[i]="compareright" then min(D[i],D[i+1])
		   else D[i])
    }.

define processor(C,L,M,R) = {
    M gets (if C="shiftleft" then L
	    else if C="shiftright" then R
	    else if C="compareleft" then max(M,L)
	    else if C="compareright" then min(M,R)
	    else M)
}.

define merge_test(data) = exists n,C,D,Phase : {
    n = |data| and
    fixed_list(C,n) and
    fixed_list(D,n) and
    D = data and
    Phase = "" and
    compare_net(n,C,D) and
/*    forall i<n : processor(C[i],D[i-1 mod n],D[i],D[i+1 mod n]) and*/
    merge(C,Phase) and
/*    always format("Data=%t    : %s\n",D,Phase)*/
    keep format("Data=%t, Inst=%t\n",D,C) and
    fin format("Data=%t\n",D)
}.

define min(X,Y) = (if X<=Y then X else Y).
define max(X,Y) = (if X>=Y then X else Y).

define merge_test1 = [1,3,4,6,0,2,5,7].
define merge_test2 = [0,1,2,3,4,5,6,7].
define merge_test3 = merge_test2+merge_test2.


load "lists".
