/* -*- Mode: C -*-
 * Tempura routines for lists.  14 Jan 85. 
 * The declaration list(L,n) specifies that L is a list of fixed (static)
 * size n, whereas List(L,n) specifies that L is a list of size <= n (as yet
 * unspecified). 
 */

/* Declare a variable to always hold a FIXED-LENGTH list. */
define fixed_list(L,n) = {
    list(L,n) and stable struct(L)
}.

/* Declare a variable to always hold a VARIABLE-LENGTH list. */
define flex_list(L,n) = {
    List(L,n) and stable struct(L)
}.

define head(L) = L[0].
define tail(L) = L[1..|L|].

/* exchange two elements of a list */
define swap_list(L,i,j) = {
    forall k < |L|: {
	if k=i then L[k]:=L[j]
	else if k=j then L[k]:=L[i]
	else L[k]:=L[k]
    }
}.

/* extend a list by one element */
define extend_list(L,d) = {
    list(next L, |L|+1) and
    forall i<|L|+1: if i<|L| then L[i]:=L[i] else L[i]:=d
}.

define part_step(L,key,I,J) = {
    if L[I] < key
    then {
	stable L
	and I:=I+1 and J:=J
    }
    else {
	swap_list(L,I,J-1)
	and I:=I and J:=J-1
    }
}.

/* partition a list into two parts */
define partition(L,key,left_len) = {
    exists I,J : {
	I = 0 and J=|L|
	and while I<J do {
	    skip and
	    part_step(L,key,I,J)
	};
	{
	    empty and left_len=I
	}
    }
}.

/* test routine for partitioning */
define part_test(part_list) = {
    exists L,key,left_len,State_Num: {
	fixed_list(L,|part_list|)
	and L=part_list and key=part_list[0]
	and partition(L,key,left_len)
	and State_Num = 0 and State_Num gets State_Num+1
	and box format("State%3D:  List=%t\n",State_Num,L)
	and fin format("Number of list elements < first element = %d\n",
	    left_len)
    }
}.

/* partitioning used by quicksort routines */
define quick_partition(L,pivot) = {
    {
	partition(L[0 to |L|-1],L[|L|-1],pivot)
	and stable L[|L|-1]
    };
    {
	skip and swap_list(L,pivot,|L|-1)
    }
}.

