/* -*- Mode: C -*-
 * The following multiplier is based on an LCF version developed by Mike
 * Gordon and described in Cambridge Comp. Lab. Tech. Rep. 41, 1983.
 * Note: This version has had some things reordered.  Also, the redundant
 * signals 'L2' (is 'N'), 'Out' (is 'M') and 'T' (is 'Done')
 * have been eliminated. 
 */

define mux(Switch, In1, In2, Out) = {
    always {if Switch then Out=In1 else Out=In2}
}.

define reg(In, Out) = {Out gets In}.

define flipflop(In, Out) = {Out gets In}.

define dec(In, Out) = {Out is (if In=0 then 0 else In-1)}.

define adder(In1,In2,Out) = {Out is In1+In2}.

define zero_test(In,Out) = {Out is (In=0)}.

define or_gate(In1,In2,Out) = {Out is (In1 or In2)}.

define zero(Out) = {Out is 0}.

define mult_imp(in1,in2,M,N,Done) = {
    exists B1,B2,B3,B4,L1,L2,L3,L4,L5,L6,L7,L8,L9,L10 : {
	zero_test(in1,B4) and
	zero(L10) and
	mux(B4,L10,in2,L9) and
	zero_test(in2,B2) and
	dec(in1,L6) and
	adder(L9,M,L8) and
	dec(N,L3) and
	mux(Done,L9,L8,L7) and
	reg(L7,M) and
	mux(Done,in1,L3,L1) and
	reg(L1,N) and
	dec(L3,L4) and
	mux(Done,L6,L4,L5) and
	zero_test(L5,B1) and
	or_gate(B1,B2,B3) and
	flipflop(B3,Done)
    }
}.

/* run */ define main() =
  exists I,M,N,in1,in2,Done: {
	input(I) and
	in1 = I[0] and
	in2 = I[1] and
	Done = true and
	M = 0 and N = 0 and
	next {halt(Done)} and
	mult_imp(in1,in2,M,N,Done) and
	always format("Done=%5t  M=%2d  N=%2d\n",Done,M,N)
}.

/* run */ define mult_test(in1,in2) =
  exists M,N,Done: {
	Done = true and
	M = 0 and N = 0 and
	next {halt(Done)} and
	mult_imp(in1,in2,M,N,Done) and
	always format("Done=%5t  M=%2d  N=%2d\n",Done,M,N)
}.

