/* -*- Mode: C -*-
 * General-purpose routines for process synchonization.
 * Initially written:  9 Jan 85.
 * Updates: 14 Jan 85; 31 Jan 85; 11 Feb 85; 16 Apr 85
 * Ben Moszkowski
 */



/* Declare a signal */
define make_signal(Sig) = {
    flex_list(Sig,2)
    and |Sig|=1
    and Sig[0]=false
}.

/* Declare a signal list */
define signal_list(Sig_list,n) = {
    fixed_list(Sig_list,n) and
    forall i<n: {
	make_signal(Sig_list[i])
    } 
}.

/* Status of a signal */
define signal_status(Sig) = {
    (|Sig|=2) and (Sig[0]=Sig[1])
}.

/* Execute some subprocesses */
define exec_subprocesses(Sig_list,n) = {
    signal_list(Sig_list,n)
    and {
	start_subprocesses(Sig_list);
	skip;
	terminate_subprocesses(Sig_list);
	skip
    }
}.

/* Start up all subprocesses */
define start_subprocesses(Sig_list) = {
    for i<|Sig_list| do {
	halt (Sig_list[i][0]) 
    }
    and forall i<|Sig_list|: {
	box (if |Sig_list[i]|=2 then Sig_list[i][1]=empty )
    /*and box format("start.Sig_list=%t  k=%t\n",Sig_list,k) */
    }
}.

/* Terminate all subprocesses */
define terminate_subprocesses(Sig_list) = {
    for i<|Sig_list| do {
	halt (~Sig_list[i][0])
    }
    and forall i<|Sig_list|: {
	box (if |Sig_list[i]|=2 then Sig_list[i][1]=not empty)
    /* and  box format("terminate.Sig_list=%t  k=%t\n",Sig_list,k)*/}
}.

/* begin a process */
define begin_process(Sig) = {
    {
	skip and |Sig|:=2 and Sig[0]:=true
    };
    {
	halt(Sig[1]) and stable |Sig| and stable Sig[0]
    };
    {
	skip and |Sig|:=1 and stable Sig[0]
    }
}.


/* finish a process */
define finish_process(Sig) = {
    {
	skip and |Sig|:=2 and Sig[0]:=false
    };
    {
	halt(~Sig[1]) and stable |Sig| and stable Sig[0]
    };
    {
	skip and |Sig|:=1 and stable Sig[0]
    }
}.

define sum_tree_process(Sig,Tree) = process {
    {
	begin_process(Sig) and stable Tree
    };
    {
	par_sum_tree(Tree) and stable Sig
    };
    {
	finish_process(Sig) and stable Tree
    }
}.

define par_sum_tree(Tree) = {
    if is_integer(Tree) then empty
    else {
	exists Sig_list: {
	    exec_subprocesses(Sig_list,2)
	    and sum_tree_process(Sig_list[0],Tree[0])
	    and sum_tree_process(Sig_list[1],Tree[1])
	    and stable struct(Tree)
            /*and box format("par-sum.Sig_list=%t k=%t\n",Sig_list,k)*/
	};
	{
	    skip and Tree:= Tree[0]+Tree[1]
	}
    }
}.

define process_test(t) = {
    exists Tree : {
	Tree=t 
	and par_sum_tree(Tree)
        and box format("Tree=%t\n",Tree)
    }
}.

define process_test1 = 1.

define process_test2 = [1,2].

define process_test3 = [[1,[2,3]],[4,5]].

define process_test4 = [[[1,1],[1,1]],[[1,1],[1,1]]].

define process_test5 = [[process_test3,process_test3],process_test3].
