/* -*- Mode: C -*-
 * Routines for buffers with projection.  
 * Ben Moszkowski  
 * 24 Apr 85.
 */

/* suppress(buffer,
buffer_closed,
sender_side,
receiver_side,
receiver_side_data,
end_of_receiver_side). 
*/


/* Declare a buffer */
define buffer(Buffer) = {
    fixed_list(Buffer,2)
    and flex_list(Buffer[0],1) and Buffer[0]=[]
    and flex_list(Buffer[1],1) and Buffer[1]=[[]]
}.

/* Test if buffer is closed */
define buffer_closed(Buffer) = {
    |Buffer[0]|=0 and |Buffer[1]|=1 and |Buffer[1][0]|=0
}.

/* Access the buffers's sender side */
define sender_side(Buffer) = {
    Buffer[0]
}.

/* Access the buffers's receiver side */
define receiver_side(Buffer) = {
    Buffer[1]
}.

/* Access data on receiver side */
define receiver_side_data(Side) = {
    Side[0][0]
}.

/* Test if end of receiver side */
define end_of_receiver_side(Side) = {
    |Side[0]|=0
}.

/* open buffer side to receive input */
define open_receiver_side(Side) = {
    {
	skip and |Side|:=0
    };
    receiver_pause("open")
}.

/*  open buffer to send output */
define open_sender_side(Side) = {
    empty
}.

/* advance a buffer to the next input */
define advance_receiver_side(Side) = {
    {
	skip and |Side|:=0
    };
    receiver_pause("advance")
}.

/* advance a sender side to the next output. */
define advance_sender_side(Side,data) = {
    {
	skip and |Side|:=1 and Side[0]:=[data]
    };
    sender_pause("advance")
}.

/*  close receiver side */
define close_receiver_side(Side) = {
    empty
}.

/* close sender side */
define close_sender_side(Side) = {
    {
	skip and |Side|:=1 and Side[0]:=[]
    };
    sender_pause("close")
}. 

/* Have receiver pause */
define receiver_pause(text) = {
    skip
    and format("Receiver pausing to do %s operation.\n",text)
    and wnext(format("Receiver resuming.\n"))
}.

/* Have sender pause */
define sender_pause(text) = {
    skip
    and format("Sender pausing to do %s operation.\n",text)
    and wnext(format("Sender resuming.\n"))
}.

/* Scan the sender side of a buffer between states of projection */
define scan_sender_side(Buffer) = {
    if |Buffer[0]|=0 then skip
    else {
	{
	    halt(|Buffer[1]|=0) and stable Buffer[0]
	};
	{
	    skip and |Buffer[0]|:=0
	}
    }
}.

/* Scan the receiver side of a buffer between states of projection */
define scan_receiver_side(Buffer) = {
    if |Buffer[1]|=1 then skip
    else {
	{
	    halt(|Buffer[0]|=1) and stable Buffer[1]
	};
	{
	    skip and Buffer[1]:=Buffer[0]
	}
    }
}.

/* Give status of buffer */
define buffer_status(Buffer) = {
    Buffer[0]
}.

/* process for writing a list onto a side */
define list_to_side_process(Ready,Done,data_input,Buffer) = process {
    {
	{
	scan_sender_side(Buffer)
	proj
	list_to_side(data_input,Buffer[0])
	}
	and stable Ready
    };
    {
	skip and Ready:=true and stable Buffer[0]
    };
    {
	(halt Done) and stable Buffer[0] and stable Ready
    }
}.

/* Body of list_to_side_process */
define list_to_side(data_input,Side) = {
    open_sender_side(Side);
    for data in data_input do {
	advance_sender_side(Side,data);
	{
	    len(3) and stable Side /* intentional delay */
	}
    };
    close_sender_side(Side)
}.

/* process for reading a side into a list */
define side_to_list_process(Ready,Done,Buffer,data_output) = 
   process {
    {
	{
	scan_receiver_side(Buffer)
	proj
	side_to_list(Buffer[1],data_output)
	}
	and stable Ready
    };
    {
	skip and Ready:=true and stable Buffer[1]
    };
    {
	(halt Done) and stable Buffer[1] and stable Ready
    }
}.

/*  Body of side_to_list_process */
define side_to_list(Side,data_output) = {
    open_receiver_side(Side);
    exists L,N: {
	L=[] and N=0 and
	{
    	    while not end_of_receiver_side(Side) do {
		{
		    skip and stable Side
		    and L:=[receiver_side_data(Side),L]
		    and N:=N+1
		};
		{
		    advance_receiver_side(Side)
		    and stable L and stable N
		}
	    };
	    {
		list(data_output,N) and
		while N>0 do {
		    skip and L:=L[1] and N:=N-1
		    and data_output[N-1]=L[0]
		}
		and stable Side
	    }
	}
    };
    close_receiver_side(Side)
}.

/* test routine for projected buffers */
define buffer_test(data_input) = {
    exists Ready1,Ready2,Done,Buffer,data_output: {
	buffer(Buffer)
	and Ready1=false and Ready2=false
	and (Done is (Ready1 and Ready2))
	and (halt Done)
	and list_to_side_process(Ready1,Done,data_input,Buffer)
	and side_to_list_process(Ready2,Done,
	    Buffer,data_output)
	and always format("Buffer=%t\n",Buffer)
	and fin(format("output list = %t\n",data_output))
    }
}.

/* test routine for buffers showing status */
define buffer_status_test(data_input) = {
    exists Sig_list,Buffer,data_output: {
	buffer(Buffer)
	and exec_subprocesses(Sig_list,2)
	and list_to_side_process(Sig_list[0],data_input,sender_side(Buffer))
	and side_to_list_process(Sig_list[1],
	    receiver_side(Buffer),data_output)
	and format("Buffer_status\n",buffer_status(Buffer))
	and always format("%8t\n",buffer_status(Buffer))
	and fin(format("output list = %t\n",data_output))
    }
}.

