/* -*- Mode: C -*-
 * Functional spec. of slotted ring having 2-bit "packets". 
 * By Roger Hale,
 * 8 Nov 84. 
 */

/* run */ define ring_test() = {
    exists Reset,Send,DataIn,DataOut : {
	DataOut = false 
	and {			/* Initialisation */
	    len(3) 
	    and Reset is true
	    and DataIn is false
	    and Send is false
	    ;			/* Active phase. */
	    for i<2 do {
		skip and DataIn := ~DataIn and Send := true
		;
		len(2) and Reset is false and stable(Send) and stable(DataIn)
	    }
	    ;
	    for i<2 do {
		skip and DataIn := ~DataIn and Send := false
		;
		len(2) and Reset is false and stable(Send) and stable(DataIn)
	    }
	    ;
	    for i<2 do {
		skip and DataIn := ~DataIn and Send := true
		;
		len(2) and Reset is false and stable(Send) and stable(DataIn)
	    }
	    ;
	    for i<2 do {
		skip and DataIn := ~DataIn and Send := true
		;
		len(2) and Reset is true and stable(Send) and stable(DataIn)
	    }
	}
	and format(" %5s %5s %5s %5s\n","I","T","R","O")
	and box format(" %5s %5s %5s %5s\n",
			wave(DataIn),wave(Send),wave(Reset),wave(DataOut))
	and ring(Reset,Send,DataIn,DataOut)
    }
}.

define ring(Reset,Send,DataIn,DataOut) = {
    exists L1,L2,L3 : {
        L1 = false and L2 = false and L3 = false 
       	and monitor_node(L3,Reset,L1)
	and transmitter_node(L1,Send,DataIn,L2)
 	and receiver_node(L2,L3,DataOut)
    }
}.

define monitor_node(In,Reset,Out) = {
					/* This node maintains either one or
					   zero packets on the ring, according
					   to whether Reset is false or true.
					   */
    exists PacketOnRing : {	/* PacketOnRing maintains a record of whether
				   or not there is a packet on the ring. */
	PacketOnRing = false
	and Out gets (~Reset /\ (In \/ ~PacketOnRing))
	and PacketOnRing gets (~Reset \/ (PacketOnRing /\ ~In))
    }
}.

define transmitter_node(In,Send,Data,Out) = {
				/* Send is a flag to indicate whether or
				   not the signal Data should be written into
				   the next packet. */
    exists InPkt : {
	InPkt = false
	and Out gets (if (Send /\ InPkt) then Data else In)
	and InPkt gets (~InPkt /\ In)
    }
}.

define receiver_node(In,Out,Data) = {
					/* The value of the last data received
				          is maintained in Data. */
    exists InPkt : {
	InPkt = false
	and Out gets In
	and InPkt gets (~InPkt /\ In)
	and Data gets (if InPkt then In else Data)
    }
}.

define delay (I,O,n) = {
    if n<=0
    then O is I
    else if n=1
    then O gets I 
    else exists D : {
	List(D,n-1)
	and forall i<n-1 : { D[i] = false } 
	and D[0] gets I
	and forall i<n-2 : { D[i+1] gets D[i] }
	and O gets D[n-2]
    }
}.

define wave(A) = {
    if ~A then "|  " else "  |"
}.
