/* -*- Mode: C -*-
 * Asynchronous evaluator using streams.  
 * 30 Jan 85.
 * Updated 12 Feb 85; 19 Apr 85
 *     Ben Moszkowski
 *
 * The process eval_process(Signal,Polish_receiver,answer) takes the reverse-
 * Polish input from the port Polish_receiver and evaluates it, returning
 * the answer at the end.
 *
 */

/* Uses stream routines. */
/* load(streams). */


/* Evaluator process */
define eval_process(Signal,Polish_receiver,answer) = process {
    {
	begin_process(Signal)
	and stable Polish_receiver
    };
    {
	eval_expr() and stable Signal
    };
    {
	finish_process(Signal)
	and stable Polish_receiver
    }
} where


/* Body of eval_process */
define eval_expr() = {
    open_polish_port();
    eval_polish();
    if end_of_polish() then close_polish_port()
    else eval_error("Evaluator terminated before end of expression.")
} and


/* Evaluate a reverse polish expression */
define eval_polish() = exists Eval_stack: {
    Eval_stack = [] and
    while not end_of_polish() do {
	{
	    skip and stable Polish_receiver and
	    if is_integer(polish_data()) then
		Eval_stack := [polish_data(),Eval_stack]
	    else if polish_data() = "+" then
		Eval_stack := [Eval_stack[0]+Eval_stack[1][0],
			Eval_stack[1][1]]
	    else eval_error(["Unknown operator: ", polish_data()])
	};
	{
	    get_polish() and stable Eval_stack
	}
    };
    {
	if |Eval_stack|=2 and |Eval_stack[1]|=0 then {
	    empty and answer=Eval_stack[0]
	}
	else eval_error("Incomplete polish input port.")
    }
} and

/* Open the polish port */
define open_polish_port() = {
    open_receiver_port(Polish_receiver)
} and
    

/* Test whether more polish input */
define end_of_polish() = {
    end_of_receiver_port(Polish_receiver)
} and

/* Current polish token */
define polish_data() = {
    receiver_port_data(Polish_receiver)
} and

/* Get a new polish token */
define get_polish() = {
    advance_receiver_port(Polish_receiver)
} and

/* Close the polish port */
define close_polish_port() = {
    close_receiver_port(Polish_receiver)
} and
    

/* Terminate if error detected */
define eval_error(text) = {
    format("**Error detected by evaluator:%t\n",text) and
    false
}.

    
/*  Test routine */
define eval_test(polish_input) = {
    exists Polish_stream,answer,Sig_list: {
	stream(Polish_stream)
	and format("Evaluator started.  Polish input list = %t\n",
	    polish_input)
	and exec_subprocesses(Sig_list,2,0)
        and exec_subprocesses(Sig_list,2,1)
	and list_to_port_process(Sig_list[0],
	    polish_input,sender_port(Polish_stream))
	and eval_process(Sig_list[1],
	    receiver_port(Polish_stream),answer)
	and box format("Polish_status=%t\n",stream_status(Polish_stream))
	and fin format("Evaluator finished.  Answer = %t\n", answer)
    }
}.

/*  Test of lexer, parser and eval routines.
The lexer and parser are found in other files.
 */
define sys_test(char_input) = {
    exists Char_stream,Token_stream,Polish_stream,answer,Sig_list: {
	stream(Char_stream) and stream(Token_stream)
	and stream(Polish_stream)
	and format("Input string = %t\n", char_input)
	and exec_subprocesses(Sig_list,4)
	and list_to_port_process(Sig_list[0],
	    char_input,sender_port(Char_stream))
	and lexer_process(Sig_list[1],
	    receiver_port(Char_stream),sender_port(Token_stream))
	and parser_process(Sig_list[2],
	    receiver_port(Token_stream),sender_port(Polish_stream))
	and eval_process(Sig_list[3],
	    receiver_port(Polish_stream),answer)
	and format(
	    "%8s%8s%8s\n","Char","Token","Polish")
	and always format(
	    "%8t%8t%8t\n",
	    stream_status(Char_stream),stream_status(Token_stream),
	    stream_status(Polish_stream))
	and fin format("Final answer = %t\n", answer)
    }
}.

	
/* Test data */
define eval_test1 = [1].

define eval_test2 = [1,2,"+"].

define eval_test3 = [1,2,"+",3,4,"+","+"].

define eval_test4 = [1,2,3,4,5,"+","+","+",6,7,"+","+","+",10,"+"].

/*  End of port-eval file */
