/* -*- Mode: C -*-
 * Routines for streams.  
 * Ben Moszkowski  
 * 15 Jan 85.
 * Updated 23 Jan 85; 29 Jan 85; 30 Jan 85; 11 Feb 85; 14 Feb 85;
 *         19 Apr 85.
 */

/* suppress(stream,
stream_closed,
sender_port,
receiver_port,
receiver_port_data,
end_of_receiver_port). 
*/


/* Declare a stream */
define stream(Stream) = {
    fixed_list(Stream,2)
    and forall i<2: {
	port(Stream[i])
	and box {
	    if |Stream[i]|=2 then
		if |Stream[1-i]|=0 then |Stream[i][1]|=0
		else Stream[i][1]=Stream[1-i][0]
	}
    }
    and Stream[0][0]=[]
    and Stream[1][0]=[[]]
}.

/* Test if stream is closed */
define stream_closed(Stream) = {
    |Stream[0]|=0 and |Stream[1]|=0
}.

/* Access the streams's sender port */
define sender_port(Stream) = {
    Stream[0]
}.

/* Access the streams's receiver port */
define receiver_port(Stream) = {
    Stream[1]
}.

/* Declare a port */
define port(Port) = {
    flex_list(Port,2)
    and forall j<2: {
	flex_list(Port[j],1)
	and flex_list(Port[j][0],1)
    }
    and |Port|=1
}.

/* Access data on receiver port */
define receiver_port_data(Port) = {
    Port[0][0][0]
}.

/* Test if end of receiver port */
define end_of_receiver_port(Port) = {
    |Port[0][0]|=0
}.

/* open stream to receive input */
define open_receiver_port(Port) = {
    {
	skip and |Port|:=2 and |Port[0]|:=0
    };
    {
	halt(|Port[1]|=1) and stable |Port| and stable Port[0]
    };
    {
	skip and |Port|:=1 and Port[0]:=Port[1]
    }
}.

/*  open stream to send output */
define open_sender_port(Port) = {
    empty
}.

/* advance a stream to the next input */
define advance_receiver_port(Port) = {
    {
	skip and |Port|:=2 and |Port[0]|:=0
    };
    {
	halt(|Port[1]|=1) and stable |Port| and stable Port[0]
    };
    {
	skip and |Port|:=1 and Port[0]:=Port[1]
    }
}.

/* advance a sender port to the next output. */
define advance_sender_port(Port,data) = {
    {
	skip and |Port|:=2 and |Port[0]|:=1
	and |Port[0][0]|:=1 and Port[0][0][0]:=data
    };
    {
	halt(|Port[1]|=0) and stable |Port| and stable Port[0]
    };
    {
	skip and |Port|:=1 and |Port[0]|:=0
    }
}.

/*  close receiver port */
define close_receiver_port(Port) = {
    empty
}.

/* close sender port */
define close_sender_port(Port) = {
    {
	skip and |Port|:=2 and |Port[0]|:=1 and |Port[0][0]|:=0
    };
    {
	halt(|Port[1]|=0) and stable |Port| and stable Port[0]
    };
    {
	skip and |Port|:=1 and stable Port[0]
    }
}. 

/* Give status of stream */
define stream_status(Stream) = {
    if |sender_port(Stream)|<2 or |receiver_port(Stream)|<2 then []
    else if |sender_port(Stream)[0]|=0 or |receiver_port(Stream)[0]|=1 then []
    else sender_port(Stream)[0]
}.

/* process for writing a list onto a port */
define list_to_port_process(Signal,data_input,Port) = process {
    {
	begin_process(Signal) and stable Port
    };
    {
	list_to_port(data_input,Port) and stable Signal
    };
    {
	finish_process(Signal) and stable Port
    }
}.

/* Body of list_to_port_process */
define list_to_port(data_input,Port) = {
    open_sender_port(Port);
    for data in data_input do {
	advance_sender_port(Port,data)
    };
    close_sender_port(Port)
}.

/* process for reading a port into a list */
define port_to_list_process(Signal,Port,data_output) = process {
    {
	begin_process(Signal) and stable Port
    };
    {
	port_to_list(Port,data_output) and stable Signal
    };
    {
	finish_process(Signal) and stable Port
    }
}.

/*  Body of port_to_list_process */
define port_to_list(Port,data_output) = {
    open_receiver_port(Port);
    exists L,N: {
	L=[] and N=0 and
	{
    	    while not end_of_receiver_port(Port) do {
		{
		    skip and stable Port
		    and L:=[receiver_port_data(Port),L]
		    and N:=N+1
		};
		{
		    advance_receiver_port(Port)
		    and stable L and stable N
		}
	    };
	    {
		list(data_output,N) and
		while N>0 do {
		    skip and L:=L[1] and N:=N-1
		    and data_output[N-1]=L[0]
		}
		and stable Port
	    }
	}
    };
    close_receiver_port(Port)
}.

/* test routine for streams */
define stream_test(data_input) = {
    exists Sig_list,Stream,data_output: {
	stream(Stream)
	and exec_subprocesses(Sig_list,2) 
        and exec_subprocesses(Sig_list,2)    
	and list_to_port_process(Sig_list[0],data_input,sender_port(Stream))
	and port_to_list_process(Sig_list[1],
	    receiver_port(Stream),data_output)
	and box format("Stream=%8t\n",Stream)
	and fin(format("output list = %t\n",data_output))
    }
}.

/* test routine for streams showing status */
define stream_status_test(data_input) = {
    exists Sig_list,Stream,data_output: {
	stream(Stream)
	and exec_subprocesses(Sig_list,2)
        and exec_subprocesses(Sig_list,2)
	and list_to_port_process(Sig_list[0],data_input,sender_port(Stream))
	and port_to_list_process(Sig_list[1],
	    receiver_port(Stream),data_output)
	and format("Stream_status\n",stream_status(Stream))
	and box format("%8t\n",stream_status(Stream))
	and fin(format("output list = %t\n",data_output))
    }
}.

