/* -*- Mode: C -*-
 * test of a memory variable
 */

/* declares a memory cell X shared by L processes */
/* X[0] is used to memorise the old value         */
/* X[1] contains the current write value          */
/* X[2+i] is write flag for process i             */
/*        false: process i doesn't write          */
/*        true: process i does write              */
define memory(X,L) = {
   list(X, 2+L) and stable(struct(X)) and
   /* memorise the current value of X */
   X[0] gets read(X)
}.

/* the current value of memory cell X */
/* if there is no write then it is the old value */
/* otherwise it is the new value                 */
define read(X) = {
  if forall i<|X|-2: { X[2+i]=false } then X[0]
  else X[1]
}.

/* process P writes Y to memory cell X */
define write(P,X,Y) = {
  X[2+P]=true and X[1]=Y 
  and format("write %d by %d\n",Y,P)
}.

/* process P writes Y to memory cell X in the next state*/
define next_write(P,X,Y) = {
  next write(P,X,Y)
}.

/* process P doesn't change the memory cell X */
define no_write(P,X) = {
 X[2+P]=false 
 and format("no write by %d\n",P)
}.

/* run */ define test() = {
  exists X,Z : {
   /* lets have a memory variable X shared by two processes */
   memory(X,2) and 
   /* print the value of X in every state */
   always ( format("X=%d\n",read(X)) ) and

  { /* process 0 */
    /* state 0: no write */
    /* state 1: write value of X in state 0  plus 1 */
    {skip and no_write(0,X) and  next_write(0,X, read(X)+1)};skip;
    /* state 2: no write */
    {empty and no_write(0,X)};skip;
    /* state 3: no write */
    {empty and no_write(0,X)};skip;
    /* state 4: write 2 */
    {empty and write(0,X,2)};skip;
    /* state 5: no write */
    {empty and no_write(0,X)}
    /* state 6: write 4 */
    /* uncomment the following and the one in process 1 to have racing */
    ;skip;{empty and write(0,X,4)} 
  }
  and
  { /* process 1 */
    /* state 0: write 10 */
    {empty and write(1,X, 10)};skip;
    /* state 1: no write */
    {empty and no_write(1,X)};skip;
    /* state 2: no write */
    {empty and no_write(1,X)};skip;
    /* state 3: no write */
    {empty and no_write(1,X)};skip;
    /* state 4: write 2 */
    {empty and write(1,X, 2)};skip;
    /* state 5: no write */
    {empty and no_write(1,X)}
    /* state 6: write 5 */
     ;skip;{empty and write(1,X,5)} 
  }

 }

}.
