/*                               -*- Mode: C -*- 
*/
define sm_open           = 5.
define sm_closed         = 0.
define sm_closing        = 3.

define outside           = 0.
define buying            = 1.
define queuing           = 2.
define served            = 3.

define cp_closed         = 0.
define cp_open_busy      = 1.
define cp_open_free      = 2.
define cp_closing_busy   = 3.
define cp_closing_free   = 4.

define notthere          = 10000.
define notserved         = -1.

define ncashpoints = 2.
define nclients    = 4.

load "../library/conversion".
load "../library/tcl".
/* tcl supermarket */

/* run */ define supermarket() = {


 exists Supermarket, Client, Qtime, 
        Cashpoint, Timer : {

  Supermarket = sm_closed and
 /* Supermarket = sm_open : supermarket is open
    Supermarket = sm_closed : supermarket is closed
    Supermarket = sm_closing : supermarket is closing
 */

 list(Client, nclients) and stable(struct(Client)) and
 (forall i<nclients : Client[i] = outside) and 
 /* Client[i] = outside : client is outside the supermarket
    Client[i] = buying  : client is in the buying area collecting items
    Client[i] = queuing  : client is in a queue waiting to pay
    Client[i] = served  : client is served at a cashpoint
 */ 

 list(Qtime, nclients) and stable(struct(Qtime)) and
 forall i<nclients: { 
  list(Qtime[i], ncashpoints) and stable(struct(Qtime[i]))
 } and
 (forall i<nclients :
   (forall j<ncashpoints :
     Qtime[i][j] = notthere
   )
 ) and
 /* Qtime[i][j] : The time that client i arrives at the queue of 
                 cashpoint j, notthere if client i is not in queue j
 */

 list(Cashpoint, ncashpoints) and stable(struct(Cashpoint)) and
 (forall j<ncashpoints : Cashpoint[j] = cp_closed) and
 /* Cashpoint[j] = cp_closed        : cashpoint is closed
    Cashpoint[j] = cp_open_busy     : cashpoint is open and busy
    Cashpoint[j] = cp_open_free     : cashpoint is open and free
    Cashpoint[j] = cp_closing_busy  : cashpoint is closing and busy
    Cashpoint[j] = cp_closing_free  : cashpoint is closing and free
 */

 Timer = 0 and
 /* Timer = k : the time is currently k
 */ 

 /* The client transactions
  *
 */
define client_enter(X) = {
  skip and 
    stable(Qtime[X]) and
    if Supermarket = sm_open then {
     if Client[X] = outside then {
      Client[X] := buying and
      format("Client %t will enter the supermarket\n",X)
     } else {
        format("Client %t is already in the supermarket\n",X) and
        stable(Client[X])
     }
    } else {
       if Supermarket = sm_closed then {
        format("Client %t can't enter the supermarket because it is closed\n",X)
       } and
       if Supermarket = sm_closing then {
        format("Client %t can't enter the supermarket because it is closing\n",X)
       } and
       stable(Client[X])
    } 
 } and

define client_leave(X) = {
  skip and 
    stable(Qtime[X]) and
    if Client[X] = buying then {
     Client[X]:= outside and 
     format("Client %t leaves the supermarket\n",X)
    } else {
       format("Client %t can't leave the supermarket\n",X) and
       stable(Client[X]) 
    } 
 } and

define client_queuing(X,Y) = {
  skip and
   if Client[X] = buying and 
     (Cashpoint[Y] = cp_open_busy or Cashpoint[Y] = cp_open_free) then { 
   Client[X]:= queuing and 
   (forall j<ncashpoints: if j=Y then Qtime[X][j]:=Timer 
                          else stable(Qtime[X][j]) ) and
   format("Client %t queues at cashpoint %t at time %t\n",X,Y,Timer+1)
 
  } else {
     format("Client %t can't queue because cashpoint %t is closing or closed\n",X,Y) and
     stable(Client[X]) and 
      stable(Qtime[X]) 
    }
 } and

define client_forgot(X) = {
  skip and
  if Client[X] = queuing then {
   format("Client %t is leaving queue %t because (s)he forgot something\n",X,in_queue(X)) and
   Client[X]:=buying and 
   (forall j<ncashpoints: if j=in_queue(X) then Qtime[X][j]:=notthere 
                          else stable(Qtime[X][j]) )
  } else {
     stable(Client[X]) and 
      stable(Qtime[X]) 
    }
 } and

define client_impatient(X,Z) = {
   skip and
    if Client[X] = queuing  and 
     (Cashpoint[Z] = cp_open_free or 
      Cashpoint[Z] = cp_open_busy) then {
    Client[X]:= queuing and 
    (forall j<ncashpoints: if j=in_queue(X) then Qtime[X][j]:=notthere 
                           else if j=Z then Qtime[X][j]:=Timer 
                           else stable(Qtime[X][j]) ) and
    format("Client %t switches from queue %t to queue %t at time %t\n",X,in_queue(X),Z,Timer+1) 
   } else {
      stable(Client[X]) and 
      stable(Qtime[X]) 
     }
 } and


define client_idle(X) = {
   skip and
    stable(Client[X]) and 
    format("Client %t remains in current state\n",X) and
     stable(Qtime[X])
 } and

 /* The cashpoint transactions
  *
 */
define cashpoint_open(Y) = {
  skip and
   if Cashpoint[Y] = cp_closed or Cashpoint[Y] = cp_closing_free then {
   Cashpoint[Y]:= cp_open_free and
   format("Cashpoint %t is to open\n",Y)
  } else {
     if  Cashpoint[Y] = cp_closing_busy then {
      Cashpoint[Y]:= cp_open_busy and
      format("Cashpoint %t is to open\n",Y)
     } else {
        stable( Cashpoint[Y]) and
        format("Cashpoint %t is already open\n",Y)
       }
  }
 } and

define cashpoint_closing(Y) = {
  skip and
  if Cashpoint[Y] = cp_open_busy then {
   format("Cashpoint %t is closing\n",Y) and
   Cashpoint[Y]:= cp_closing_busy
  } else {
   if Cashpoint[Y] = cp_open_free then {
     format("Cashpoint %t is closing\n",Y) and
     Cashpoint[Y]:= cp_closing_free
   } else {
     stable( Cashpoint[Y])
     }
  }
 } and

define cashpoint_close(Y) = {
  skip and
  if Cashpoint[Y] = cp_closing_free and queue_empty(Y) then {
   format("Cashpoint %t is to close\n",Y) and
   Cashpoint[Y]:= cp_closed
  } else {
     if not queue_empty(Y) then {
      format("The queue of cashpoint %t is not empty\n",Y)
     } and
     if Cashpoint[Y] = cp_closed then {
      format("Cashpoint %t is already closed\n",Y)
     } and
     if Cashpoint[Y] = cp_closing_busy then {
      format("Cashpoint %t is still serving\n",Y)
     } and
     if Cashpoint[Y] = cp_open_busy or 
        Cashpoint[Y] = cp_open_free then {
      format("Cashpoint %t has an open queue\n",Y)
     } and
     stable( Cashpoint[Y])
    }
 } and

define cashpoint_idle(Y) = {
  stable(Cashpoint[Y]) and
  format("Cashpoint %t remains in current state\n",Y)
 } and

 /* The synchronized client and cashpoint transactions
  *
 */
define client_served(X) = {
   skip and
   if  Client[X] = queuing and 
       (Cashpoint[in_queue(X)]=cp_open_free or 
        Cashpoint[in_queue(X)]=cp_closing_free) then {
     Client[X] := served and 
     format("Client %t is going to be served at cashpoint %t\n",X,in_queue(X))
     and stable(Qtime[X])
   } else {
      stable(Client[X]) and 
       stable(Qtime[X])
     }
 } and

define cashpoint_begin_processing(Y) = { 
  if Cashpoint[Y] = cp_closing_free then {
   Cashpoint[Y]:= cp_closing_busy and
   client_served(cl_served(Y)) 
  } else {
     if Cashpoint[Y] = cp_open_free then {
      Cashpoint[Y]:= cp_open_busy and
      client_served(cl_served(Y))
     } else {
        stable( Cashpoint[Y])
       }
    }
 } and

define client_paid(X) = {
   skip and
   if Client[X] = served  then {
    Client[X]:= outside and 
    (forall j<ncashpoints: if j=in_queue(X) then Qtime[X][j]:=notthere 
                           else stable(Qtime[X][j]) ) and
    format("Client %t is paying at cashpoint %t and leaves the supermarket\n",X,in_queue(X))   
   } else {
      stable(Client[X]) and 
      stable(Qtime[X]) 
     }
 } and

define cashpoint_end_processing(Y) = {
  skip and
  if Cashpoint[Y] = cp_closing_busy then {
   Cashpoint[Y]:= cp_closing_free and
   (forall i<nclients: 
     if Client[i]=served and in_queue(i)=Y then {
     client_paid(i) 
     } 
   )
  } else {
     if Cashpoint[Y] = cp_open_busy then {
      Cashpoint[Y]:= cp_open_free and
      (forall i<nclients: 
       if Client[i]=served and in_queue(i)=Y then {
        client_paid(i)
       }
      )
     } else {
        stable( Cashpoint[Y])
       }
    }
 } and   

 /* The supermarket transactions
  *
 */
define supermarket_open() = {
  skip and
  if Supermarket = sm_closed then {
   format("Supermarket is to open\n") and
   Supermarket:= sm_open
  } else {
     if Supermarket = sm_closing then {
      format("Can't open because supermarket is closing\n") 
     } and
     if Supermarket = sm_open then {
      format("Supermarket is already open\n") 
     } and
     stable(Supermarket)
    } 
 } and

define supermarket_closing() = {
  skip and
  if Supermarket = sm_open then {
   format("Supermarket will be closing\n") and
   Supermarket:= sm_closing
  } else {
     if Supermarket = sm_closed then {
      format("Can't closing because supermarket is closed\n") 
     } and
     if Supermarket = sm_closing then {
      format("Supermarket is already closing\n") 
     } and
     stable(Supermarket)
    } 
 } and

define supermarket_close() = {
  skip and
  if Supermarket = sm_closing and in_supermarket(nclients) = 0 then {
   format("Supermarket is to close\n") and
   Supermarket:= sm_closed
  } else {
     if Supermarket = sm_closed then {
      format("Supermarket is already closed\n") 
     } and
     if in_supermarket(nclients) ~= 0 then {
      format("Supermarket can't  close because there\n") and
      format("is(are) still %t Client(s)\n", in_supermarket(nclients))
     } and
     if Supermarket = sm_open then {
      format("Can't close because supermarket is open\n") 
     } and
     stable(Supermarket)
    } 
 } and

define supermarket_idle() = {
  stable(Supermarket) and
  format("Supermarket remains in current state\n")
 } and

 /* some help functions regarding output
  *
 */
define out_sm(X) = {
  if X=sm_open then "open"
  else if X=sm_closed then "closed"
  else if X=sm_closing then "closing"
  else "in a not currently known state"
 } and

define str_cl(X) = {
  if X=outside then "outside"
  else if X=buying then "buying"
  else if X=queuing then "queuing"
  else if X=served then "served"
  else "in a not currently known state"
 } and

define out_cl() = {
  (forall j<nclients :
    if Client[j]=queuing then { 
    format("Client %t is %s at queue %t\n",j,str_cl(Client[j]),in_queue(j))
    } else {
    format("Client %t is %s \n",j,str_cl(Client[j]))
    }
    
  )
 } and

define str_cp(X) = { 
  if X=cp_closed then "closed"
  else if X=cp_open_busy then "open and busy"
  else if X=cp_open_free then "open and free"
  else if X=cp_closing_busy then "closing and busy"
  else if X=cp_closing_free then "closing and free"
  else "in a not currently known state"
 } and
 
define out_cp() = {
   (forall i<ncashpoints : 
    format("Cashpoint %t is %s\n",i,str_cp(Cashpoint[i]))
  )
 } and

 /* some functions and predicates regarding queue processing
  *
 */
define queue_empty(Y) = {
  (forall i<nclients : if Client[i]=queuing and in_queue(i)=Y then
                         Qtime[i][Y] = notthere)
 } and

define find_served(n,Y) = {
  if n=0 then { format("didn't find a client to serve\n")}
  else { if in_queue(n-1)=Y and Client[n-1]=queuing and
          (forall i<nclients : if in_queue(i)=Y and Client[i]=queuing and
                            i~=n-1
                            then {
                             Qtime[n-1][Y]<Qtime[i][Y] or 
                             (Qtime[n-1][Y]=Qtime[i][Y] and n-1<i)
                            }
          ) 
        then { n-1 }
        else { find_served(n-1,Y) }
       } 
 } and

define cl_served(Y) = {
   find_served(nclients,Y)
 } and

define find_leave(n,Y) = {
  if n=0 then { format("didn't find a client that has paid and is to leave\n")}
  else { if Client[n-1]=served and in_queue(n-1)=Y then { n-1 }
         else { find_leave(n-1,Y)}
       }
 } and

define cl_leave(Y) = {
   find_leave(nclients,Y)
 } and

define count_prev(n,X) ={
   if n=0 then 0
   else count_prev(n-1,X)+{
         if Client[n-1]=queuing and in_queue(n-1)=in_queue(X) and
           n-1~=X and
           (Qtime[n-1][in_queue(n-1)]<Qtime[X][in_queue(X)] or
            (Qtime[n-1][in_queue(n-1)]=Qtime[X][in_queue(X)] and n-1<X))
         then 1
         else 0
        } 
  } and

define pos_queue(X) = {
   if Client[X]~=queuing then notthere
   else count_prev(nclients,X)
  } and

define find_queue(n,X) = {
  if n=0 then { format("didn't find the queue!?\n") } 
  else { if Qtime[X][n-1]~=notthere then n-1
         else find_queue(n-1,X)
       }
 } and

define in_queue(X) = {
  if (forall i<ncashpoints : Qtime[X][i] = notthere) then notthere
  else find_queue(ncashpoints,X)
 } and

 /* function to determine the number of clients in the supermarket
  *
 */
define in_supermarket(n) = {
   if n=0 then 0
   else in_supermarket(n-1)+{if Client[n-1]~=outside then 1 else 0} 
 } and

 /* the main program
  *
 */
 tcl("init",[nclients,ncashpoints]) and 
 always tclbreak() and 
 always(
  tcl("supermarket",[Supermarket]) and
  tcl("tmr",[Timer]) and
  (forall i<nclients : tcl("client",[i,Client[i],in_queue(i),pos_queue(i)])) and
  (forall j<ncashpoints : tcl("cashpoint",[j,Cashpoint[j]]))
 ) and
 /*always(
  format("\n**********************************************************\n") and
  format("Time= %t\n",Timer) and
  out_cl() and out_cp() and
  format("Supermarket is %s\n", out_sm(Supermarket)) and
  format("Number of clients in the supermarket is %t\n",in_supermarket(nclients))
   and
  format("-----------------------------------------------------------\n")
 ) and*/
 Timer gets Timer+1 and

 /* A sample run of the supermarket
  *
 */
 {
  
  {
   supermarket_open() and 
   (forall i<nclients   : client_idle(i)) and 
   (forall j<ncashpoints: cashpoint_idle(j))
  };
  {
   supermarket_idle() and  
   (forall i<nclients : client_idle(i)) and 
   (forall j<ncashpoints: if j=0 then cashpoint_open(j)
                          else cashpoint_idle(j))
  };
  {
   supermarket_idle() and 
   (forall i<nclients : if i~=0 and i~=1 then client_idle(i)
                        else client_enter(i) ) and 
   (forall j<ncashpoints: cashpoint_idle(j))
  };
  {
   supermarket_idle() and 
   (forall i<nclients : if i~=2 and i~=3 then client_idle(i)
                        else client_enter(i)) and
   (forall j<ncashpoints: cashpoint_idle(j))
  };
  {
   supermarket_idle() and
   (forall i<nclients : if i~=0 and i~=1 then client_idle(i)
                        else  client_queuing(i,0) ) and
   (forall j<ncashpoints: cashpoint_idle(j))
  };
  {
   supermarket_idle()  and
   (forall i<nclients : if i~=0 then client_idle(i)
                        else client_forgot(i)) and
   (forall j<ncashpoints: cashpoint_idle(j)) 
  }; 
  {
   supermarket_idle() and
   (forall i<nclients : if i~=0 then client_idle(i)
                        else client_queuing(i,0)) and
   (forall j<ncashpoints: cashpoint_idle(j))
  };
  { 
   supermarket_idle()  and
   (forall i<nclients: if i~=cl_served(0) 
                       then client_idle(i)) and
   (forall j<ncashpoints: if j~=0 
                          then cashpoint_idle(j)
                          else cashpoint_begin_processing(j))
   
  };
  {
   supermarket_idle() and  
   (forall i<nclients : client_idle(i)) and 
   (forall j<ncashpoints: if j=1 then cashpoint_open(j)
                          else cashpoint_idle(j))
  };
  {
   supermarket_idle() and
   (forall i<nclients : if i~=2 then client_idle(i)
                        else client_queuing(2,1)) and  
   (forall j<ncashpoints: cashpoint_idle(j))
  };
  {
   supermarket_idle() and
   (forall i<nclients : if i~=0 then client_idle(i)
                        else  client_impatient(i,1) ) and
   (forall j<ncashpoints: cashpoint_idle(j))
  };
  { 
   supermarket_idle()  and
   (forall i<nclients: if  i~=cl_served(1)
                       then client_idle(i)) and
   (forall j<ncashpoints: if  j~=1 
                          then cashpoint_idle(j)
                          else cashpoint_begin_processing(j))
   
  };
  {
   supermarket_idle()  and
   (forall i<nclients: if i~=cl_leave(0) and i~=cl_leave(1)
                       then client_idle(i) ) and
   (forall j<ncashpoints: if j~=0 and j~=1 
                          then cashpoint_idle(j)
                          else cashpoint_end_processing(j))   
  };
  {
   supermarket_idle()  and
   (forall i<nclients: if i~=cl_served(1) then
                       client_idle(i) ) and
   (forall j<ncashpoints: if j=1 
                          then cashpoint_begin_processing(j)
                          else cashpoint_idle(j)) 
  };
  {
   supermarket_idle() and 
   (forall i<nclients: client_idle(i)) and 
   (forall j<ncashpoints: if j=0 then cashpoint_closing(j)
                          else cashpoint_idle(j))
  };
  {  
   supermarket_idle()  and
   (forall i<nclients: client_idle(i)) and 
   (forall j<ncashpoints: if j=0 then cashpoint_close(j)
                          else cashpoint_idle(j))
  };
  {
   supermarket_idle()  and
   (forall i<nclients: if i~=3 then client_idle(i)
                       else client_leave(i)) and
   (forall j<ncashpoints:  cashpoint_idle(j))
  };
  {
   supermarket_closing()  and
   (forall i<nclients: client_idle(i) ) and
   (forall j<ncashpoints:  cashpoint_idle(j)) 
  };
  {
   supermarket_idle() and
   (forall i<nclients: if i~=cl_leave(1) then
                        client_idle(i) )  and
   (forall j<ncashpoints: if j=1
                          then cashpoint_end_processing(j)
                          else cashpoint_idle(j))
  };
  {
   supermarket_idle() and 
   (forall i<nclients: client_idle(i)) and 
   (forall j<ncashpoints: if j=1 then cashpoint_closing(j)
                          else cashpoint_idle(j))
  };
  {  
   supermarket_idle()  and
   (forall i<nclients: client_idle(i)) and 
   (forall j<ncashpoints: if j=1 then cashpoint_close(j)
                          else cashpoint_idle(j))
  };
  {
   supermarket_close() and 
   (forall i<nclients: client_idle(i)) and
   (forall j<ncashpoints: cashpoint_idle(j))
  }
 }

 }

}.
set print_states=true.
