/* errors.c:  error handling
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: errors.c,v 2.11 2007/05/03 09:48:46 acau Exp $
 */



#include "system.h"


/*****************\
*                 *
* Error handling  *
*                 *
\*****************/

void pop_tempura_level(b)
int b;
{
    if (TempuraLevel > 0) {
        TempuraLevel -= 1;
        Break = b;
    }
    else
        if (b) {
            printf ("Type \"exit.\" to exit from Tempura.\n");
            fflush(stdout);
        }
    
}

void tempura_prefix(s)
char *s;
{
    printf ("\n***Tempura %s:  ", s);
}

/* print out an error message and break */
void error_break2(text,v1,v2)
char *text, *v1, *v2;
{
    tempura_prefix ("error");
    printf (text, v1, v2);
    printf (".\n");
    fflush(stdout);
    show_form ();
    breakin ();
}

/* print out an error message and break */
void error_break2d(text,v1,v2)
char *text;
int v1, v2;
{
    tempura_prefix ("error");
    printf (text, v1, v2);
    printf (".\n");
    fflush(stdout);
    show_form ();
    breakin ();
}

/* print out an error message and break */
void error_break2e(text,v1)
char *text;
int v1;
{
    tempura_prefix ("error");
    printf (text, v1);
    printf (".\n");
    fflush(stdout);
    show_form ();
    breakin ();
}
/* print out a message and break */
void user_break(text)
char *text;
{
    tempura_prefix ("break");
    printf (text);
    printf (".\n");
    fflush(stdout);
    show_form ();
    breakin ();
}

/* print out a warning */
void warn1(text,v1)
char *text, *v1;
{
    tempura_prefix ("warning");
    printf (text, v1);
    printf (".\n");
    fflush(stdout);
    show_form ();
}

/* print out a warning with two data items*/
void warn2(text,v1,v2)
char *text, *v1, *v2;
{
    tempura_prefix ("warning");
    printf (text, v1, v2);
    printf (".\n");
    fflush(stdout);
    show_form ();
}



void show_state()
{
  int env_level;
  newline;
  printf ("!Begin state: %d",StateNumber);
  newline;
  fflush(stdout);
  for (env_level = EnvTop; env_level >= EnvBase; env_level -= 1) {
     NAMEPTR id = EnvStack[env_level].env -> id_list;
     while (id != NULL) {
         MEMORYPTR loc = curr_loc (mem_loc (id));
         output_str (id -> name);
         output_str ("=");
         if (!mem_defined (loc) || !val_defined (get_mem (loc))) {
                output_str ("?");
         }
         else {
             if (val_defined (get_mem (loc))) {
                     output_val(get_mem (loc));
                 }
         }
         newline;
         fflush(stdout);
         id = id -> next_node;
     }
  }
  printf ("!End state: %d",StateNumber);
  newline;
  fflush(stdout);
}

         
         

void show_undef_vars()
{
    int env_level;
    char c;
    printf ("   Undefined: ");
    if (!empty_defined ()) {
        output_str ("{ empty }");
    }
    for (env_level = EnvTop; env_level >= EnvBase; env_level -= 1) {
        NAMEPTR id = EnvStack[env_level].env -> id_list;
	output_char ('{');
	c = ' ';
        while (id != NULL) {
            MEMORYPTR loc = curr_loc (mem_loc (id));
            if (!mem_defined (loc) || !val_defined (get_mem (loc))) {
                output_char (c);
                output_str (id -> name);
		c = ',';
	    }
	    id = id -> next_node;
        }
        output_str (" }");
    }
    newline;
    fflush(stdout);
}

void state_not_defined()
{
    tempura_prefix ("error");
    printf ("state #%d (pass #%d) is not completely defined.\n",
             StateNumber, PassNumber);
    fflush(stdout);
    show_form ();
    show_undef_vars ();
    breakin ();
}

void show_form()
{
    NODEPTR e;
    if ((e = ReduceStackTop -> n) != NULL) {
        printf ("   Evaluating:  ");
        output_expn (e, PrintLevel);
        newline;
        fflush(stdout);
    }
}

void show_trace()
{
    int i=1;
    REDUCESTACKPTR rs;
    printf ("\nTrace of reductions:\n\n");
    fflush(stdout);
    for (rs = ReduceStack; rs < ReduceStackTop; rs += 1) {
        printf ("%3d. ",i++);
        output_expn (rs -> n, PrintLevel);
        newline;
        fflush(stdout);
    }
    newline;
    fflush(stdout);
}

void show_scope()
{
    int env_level;
    printf ("Environment:\n\n");
    fflush(stdout);
    for (env_level = EnvTop; env_level >= EnvBase; env_level -= 1) {
        printf ("%3d. ", EnvStack[env_level].env -> env_level);
        output_env (EnvStack[env_level].env);
        newline;
        fflush(stdout);
    }
    newline;
    fflush(stdout);
}
