/* fns.h:  
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: fns.h,v 2.12 2007/05/03 09:48:46 acau Exp $
 */

/* parser.c */
NODEPTR parse(FILE *stream);
NODEPTR resolve_refs(NODEPTR e);
NODEPTR rexpr(int n);
NODEPTR rbexpr(void);
NODEPTR rexists(void);
NODEPTR rforall(void);
NODEPTR rif(void);
NODEPTR rwhile(void);
NODEPTR rrepeat(void);
NODEPTR rchopstar(void);
NODEPTR rkeep(void);
NODEPTR rfor(void);
NODEPTR rdefine(void);
NODEPTR rset(void);
NODEPTR rlambda(void);
NODEPTR rlist(void);
NODEPTR rsublist(NODEPTR list);
NODEPTR rlistdef(unsigned int type);
NODEPTR rformat(void);
NODEPTR rhelp(void);
NODEPTR rcall(SYMBOLPTR id);
NODEPTR rnamelist(void);
NODEPTR rname(void);
NODEPTR rexprlist(void);
NODEPTR rsysvar(void);
void push_scope(NODEPTR name_list);
NODEPTR pop_scope(void);
void install_name_list(NODEPTR name_list);
NAMEPTR lookup_env(char *s);
NAMEPTR install_global_id(char *name);
void install_local_id(char *name);
void next_token(void);
void assert_syntax(int n);
void syntax_error(int n);
void syntax_msg(int n);


/* lexer.c */
int read_token(void);
int isnextchar(int expect);
int follow(int expect, int ifyes, int ifno);
int backslash(int c);
SYMBOLPTR symballoc(void);
unsigned hash(char *s);
SYMBOLPTR lookup(char *s);
SYMBOLPTR install(int type, char *name);
void install_keyword(int type, char *name);


/* init.c */
void install_keywords(void);
void tempura_init (void);


/* top-level.c */
int main(void);
void prompt(int level);
void tempura(int level);
void exec_command(NODEPTR cmd); 
NODEPTR reduce_stmt(NODEPTR e); 
NODEPTR reduce_val(NODEPTR e); 
MEMORYPTR reduce_ref(NODEPTR e); 
NODEPTR reduce_pgm(NODEPTR e); 
NODEPTR reduce_ref_and_fetch(NODEPTR e); 
MEMORYPTR make_ref(NODEPTR e);
void get_val(NODEPTR e);
void get_ref(NODEPTR e);


/* version.c */
void print_version_number(void);


/* errors.c */
void pop_tempura_level(int b);
void tempura_prefix(char *s);
void error_break2(char *text, char *v1, char *v2);
void error_break2d(char *text, int v1, int v2);
void error_break2e(char *text, int v1);
void user_break(char *text);
void warn1(char *text, char *v1);
void warn2(char *text, char *v1, char *v2);
void show_state(void);
void show_undef_vars(void);
void state_not_defined(void);
void show_form(void);
void show_trace(void);
void show_scope(void);


/* val.c */
void type_check(NODE1PTR e, unsigned type);
int val_defined(NODEPTR e);
int pointer_defined(LOCPTR p);
int list_defined(LISTPTR list);
int max_size(LISTPTR list);
int list_size_defined(LISTPTR list);
void output_val(NODEPTR e);
void output_list(LISTPTR list);
NODEPTR copy_val(NODEPTR val);
LISTPTR copy_list(LISTPTR list);
int string_size(STRINGPTR s);
NODEPTR substring(STRINGPTR s, int i, int j);
NODEPTR append_string(STRINGPTR s1, STRINGPTR s2); 
LISTPTR append_list(LISTPTR list1, LISTPTR list2);
LISTPTR new_list(int cell_type, int size_type, INTPTR n);
void assign_val (register MEMORYPTR loc, NODEPTR val);
void assign_struct(MEMORYPTR loc, LISTPTR struct1);
void assign_list(MEMORYPTR loc, LISTPTR list);
int equal_val(NODEPTR e1, NODEPTR e2);
int compare_val(INTPTR e1, INTPTR e2);
int compare_strings(STRINGPTR e1, STRINGPTR e2);
int compare_ptrs(LOCPTR e1, LOCPTR e2);
int compare_lists(LISTPTR list1, LISTPTR list2);


/* mem.c */
MEMORYPTR new_loc(register MEMORYPTR loc, int attrib);
BLOCKPTR mem_block(int attrib, register unsigned int size);
BLOCKPTR allocate_mem(unsigned int size);
int mem_defined(register MEMORYPTR loc);
NODEPTR get_mem(register MEMORYPTR loc);
NODEPTR curr_list(LISTPTR list);
NODEPTR next_list(LISTPTR list);
MEMORYPTR list_size1(LISTPTR list);
MEMORYPTR list_base1(LISTPTR list);
NODEPTR put_mem(register MEMORYPTR loc, NODEPTR val);
int is_static_mem(register MEMORYPTR loc);
void make_static_mem(register MEMORYPTR loc);
void make_indirect_mem(MEMORYPTR loc, register MEMORYPTR val);
void erase_mem(register MEMORYPTR loc);
int empty_defined(void);
void put_empty(NODEPTR val);
NODEPTR get_empty(void);
void push_env(BNDPTR bnd);
void pop_env(void);
void push_display(BNDPTR bnd);
void pop_display(void);


/* store.c */
char *allocate(unsigned int size, unsigned int elt_size);
char *strsave(char *s);
char *strapnd(char *s1, char *s2);
char *substr(char *s, int i, int j);
void more_symbol_table(unsigned size);
void more_heap(register unsigned size);
NODEPTR claim_node(void);
int free_heap_size(void);
NODEPTR node(unsigned type, NODEPTR v1, NODEPTR v2, NODEPTR v3);
NODEPTR bool_node(unsigned int val);
NAMEPTR name_node (unsigned signal, char *name, unsigned int level, unsigned int off);
NODEPTR int_node(int val);
NODEPTR string_node (char *val);
NODEPTR env_node (unsigned level, int size, NAMEPTR id);
NODEPTR bnd_node (ENVPTR env, BLOCKPTR blk);
NODEPTR list_node(BLOCKPTR blk, MEMORYPTR size_loc, MEMORYPTR base_loc);
BLOCKPTR block_node(unsigned int size, MEMORYPTR base_loc);
unsigned int elist_size(NODE2PTR elist);


/* io.c */
void output_int(int n, int f);
void output_str(char *s);
void outputf_str(char *s, int f);
void output_lambda(NODE2PTR e);
char *print_name(NODEPTR e);
void message(char *text);
void spaces(int n);
void output_env(ENVPTR env);
void output_expn(NODEPTR e, int level);


/* system.c */
RETSIGTYPE onintr(int sig);
RETSIGTYPE onfpe(int sig);
RETSIGTYPE onbus(int sig);
RETSIGTYPE onsegv(int sig);
RETSIGTYPE onsys(int sig);
RETSIGTYPE onalrm(int sig);
void catch_signals(void);
void breakin(void);
FILE *openfile(char *fullname, char *spath, char *name, char *suffix, char *mode);
void assign_sysvar(SYMBOLPTR id, NODEPTR val);
void init_sysvars(void);
void start_time(void);
void check_time(void);
int own_wait(int *statusPtr);
void check_sys_waits(void);


/* r-assign.c */
NODEPTR assign_curr(NODEPTR e1, NODEPTR e2);
NODEPTR assign_next(NODEPTR e1, NODEPTR e2);
void reduce_EQ_stmt(NODE2PTR e);
void reduce_GETS(NODE2PTR e);
void reduce_STB(NODE1PTR e);
void reduce_IS(NODE2PTR e);
void reduce_ASSN(NODE2PTR e);
void reduce_UASSN(NODE2PTR e);


/* r-int.c */
void reduce_SEMI(NODE2PTR e);
void reduce_iSEMI(NODE3PTR e);
void reduce_WHILE(NODE2PTR e);
void reduce_REPEAT(NODE2PTR e);
void reduce_CHOPSTAR(NODE1PTR e);
void reduce_KEEP(NODE1PTR e);
void reduce_FORTIMES(NODE2PTR e);
void reduce_iFORTIMES(NODE1PTR e);
void reduce_FOR(NODE3PTR e);
void reduce_iFOR(NODEPTR e);
void reduce_FORIN(NODE3PTR e);
void reduce_iFORIN(NODEPTR e);
void reduce_PROJ(NODEPTR e);
void reduce_iPROJ(NODEPTR e);


/* r-linear.c */
void reduce_AND(NODE2PTR e);
void reduce_iAND(NODE3PTR e);
void reduce_IMPLIES(NODE2PTR e);
void reduce_WHERE(NODEPTR e);
void reduce_IF_stmt(NODE3PTR e);
void reduce_IF_val(NODE3PTR e);
void reduce_iIF_val(NODE3PTR e);
void reduce_CALL(NODE2PTR e);
void reduce_iCALL_stmt(NODEPTR e);
void reduce_iCALL_val(NODEPTR e);
void reduce_iCALL_ref(NODEPTR e);
void reduce_EXISTS(NODE3PTR e);
void reduce_iEXISTS_stmt(NODE2PTR e);
void reduce_iEXISTS_val(NODE2PTR e);
void reduce_FORALL(NODE3PTR e);
void reduce_THE(NODE3PTR e);
void reduce_ALWAYS(NODE1PTR e);
void reduce_SOMETIME(NODEPTR e);
void reduce_HALT(NODE1PTR e);
void reduce_INIT(NODE1PTR e);
void reduce_FIN(NODE1PTR e);
void reduce_KEEP(NODE1PTR e);
void reduce_WNEXT(NODE1PTR e);
void reduce_NEXT_stmt(NODE1PTR e);
void reduce_NEXT_val(NODE1PTR e);
void reduce_NEXT_ref(NODE1PTR e);
void reduce_LEN(NODE1PTR e);
void reduce_iLEN(NODE1PTR e);
void reduce_PROCESS(NODE1PTR e);
void reduce_iPROCESS(NODE3PTR e);
void reduce_PREFIX(NODE1PTR e);
void reduce_iPREFIX(NODE3PTR e);


/* r-symbol.c */
void noop(NODEPTR e);
void assert_stmt(NODEPTR e);
void assert_expr(NODEPTR e);
void assert_stmt_to_true(NODEPTR e);
void assert_true(NODEPTR e);
void reduce_const(NODE2PTR e);
void reduce_id_val(NAMEPTR e);
void reduce_id_ref(NAMEPTR e);
void reduce_EMPTY_stmt(NODE0PTR e);
void reduce_MORE_stmt(NODE0PTR e);
void reduce_EMPTY_val(NODE0PTR e);
void reduce_MORE_val(NODE0PTR e);
void reduce_EMPTY_ref(NODE0PTR e);
void reduce_SKIP(NODE0PTR e);


/* r-expr.c */
void reduce_TYPE(NODE1PTR e);
void reduce_UNDEFINED(NODE1PTR e);
void reduce_EQ_val(NODE2PTR e);
int power(int i, int j);
NODEPTR calculate_val(unsigned op, NODEPTR e1, NODEPTR e2);
void reduce_bin_op(NODE2PTR e);
void reduce_u_op(NODE1PTR e);
void reduce_RANDOM(void);
void reduce_RRANDOM(void);
void reduce_ASCII(NODE1PTR e);
void reduce_REFER(NODE1PTR e);
void reduce_DEREF(NODE2PTR e);


/* r-io.c */
void reduce_INPUT(NODE1PTR e);
void reduce_OUTPUT(NODE1PTR e);
void reduce_PEEK(NODE1PTR e);
void reduce_FORMAT(NODE1PTR e);


/* r-lists.c */
void reduce_list(NODE2PTR e);
void reduce_STRUCT_val(NODEPTR e);
void reduce_STRUCT_ref(NODEPTR e);
void reduce_MAKELIST(NODE1PTR e);
int reduce_index(NODEPTR e);
void reduce_SUB(NODE2PTR e);
void reduce_SIZE(NODE1PTR e);
NODEPTR sublist(LISTPTR list, int m, int n);
void reduce_SUBLIST(NODE3PTR e);


/* r-system.c */
void reduce_RUN(NODE1PTR e);
void init_exec_stat(void);
void init_state_stat(void);
void init_pass_stat(void);
void update_exec_stat(void);
void update_state_stat(void);
void output_pass_stat(void);
void output_state_stat(void);
void output_exec_stat(void);
void reduce_LOAD(NODE1PTR e);
void reduce_DEFINE(NODE2PTR e);
void reduce_EXIT(SYMBOLPTR e);
void reduce_SET(NODE2PTR e);
void reduce_SYSVAR(NODEPTR e);
void reduce_SHOWTRACE(SYMBOLPTR e);
void show_mem(MEMORYPTR loc);
void reduce_SHOWMEM(NODE2PTR e);
void reduce_EOF(NODE0PTR e);
void reduce_RETURN(SYMBOLPTR e);
void reduce_CONTINUE(SYMBOLPTR e);
void reduce_LEAP(SYMBOLPTR e);
void reduce_STEP(SYMBOLPTR e);
void reduce_BREAK(SYMBOLPTR e);
void reduce_ABORT(SYMBOLPTR e);
void reduce_HELP(NODEPTR e);
void reduce_SYSTEM(NODEPTR e);
void reduce_iSYSTEM(NODEPTR e);
void reduce_SHOWSTATE(SYMBOLPTR e);


/* gc.c */
void mark(NODEPTR e);
void mark_mem(MEMORYPTR loc, unsigned n);
void mark_phase(void);
int is_refd_block(BLOCKPTR bp);
void reclaim_node(NODEPTR np);
void reclaim_phase(void);
void collect_garbage(void);




