/* gc.c:  garbage collection
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: gc.c,v 2.11 2007/05/03 09:48:46 acau Exp $
 */


#include "system.h"



/**********************\
*                      *
*  Garbage collector   *
*                      *
\**********************/

void mark(e)
NODEPTR e;
{
    if (e == NULL)
        return;
    if (e -> mark)
        return;
    e -> mark = T;
    switch (e -> type) {
        case STATE:
        case STATIC:
            mark (e -> next_node);
            break;
        case ENV:
            mark (nodep(envp (e) -> id_list)); /* AC pointer */
            break;
        case BND:
	    mark (nodep(bndp (e) -> env)); /* AC pointer */
	    mark (nodep(bndp (e) -> blk)); /* AC pointer */
	    break;
        case NAME:
        case NIL:
        case BOOL:
        case INT:
        case STRING:
            break;
        case LOC:
            mark_mem (locp (e) -> ptr, 1);
            break;
        case LIST:
            mark (nodep(listp (e) -> blk)); /* AC pointer */
            break;
        case MEMBLOCK:
            mark_mem (blockp (e) -> blk_base, blockp (e) -> blk_size);
            break;
        case FOR:
        case EXISTS:
        case FORALL:
        case IF:
        case SUBLIST:
        case FORTIMES:
        case FORIN:
        case iAND:
        case iCALL:
        case iEXISTS:
        case iPROCESS:
        case iPREFIX:
        case iPROJ:
        case iSEMI:
        case iFOR:
        case iFORIN:
        case iIF:    
            mark (e -> c3);
        case PROJ:
        case COMMA:
        case DEFINE:
        case LAMBDA:
        case WHILE:
        case REPEAT:
        case LT:
        case LE:
        case EQ:
        case GE:
        case GT:
        case NEQ:
        case PLUS:
        case MINUS:
        case MULT:
        case DIV:
        case POWER:
        case SEMI:
        case WHERE:
        case AND:
        case OR:
        case IMPLIES:
        case GETS:
        case IS:
        case MOD:
        case ASSN:
        case UASSN:
        case CALL:
        case SUB:
        case FIXLIST:
        case VARLIST:
        case RUN:
        case SET:
        case SHOWMEM:
        case iSYSTEM:
            mark (e -> c2);
        case NOT:
        case UPLUS:
        case UMINUS:
        case PROCESS:
        case PREFIX:
        case ALWAYS:
        case SOMETIME:
        case HALT:
        case INIT:
        case FIN:
        case CHOPSTAR:    
        case KEEP:
        case NEXT:
        case WNEXT:
        case LEN:
        case STB:
        case STRUCT:
        case INPUT:
        case OUTPUT:
        case PEEK:
        case HELP:
        case TYPE:
        case UNDEFINED:
        case ASCII:
        case FORMAT:
        case SIZE:
        case MAKELIST:
        case REFER:
        case DEREF:
        case LOAD:
        case SYSTEM:
        case iLEN:
        case iFORTIMES:
            mark (e -> c1);
            break;
        case END:
        case SKIP:
        case EMPTY:
        case MORE:    
        case EXIT:
        case SHOWTRACE:
        case ABORT:
        case BREAK:
        case RETURN:
        case CONTINUE:
        case LEAP:
        case STEP:
        case SHOWSTATE:
	case PASSLIMIT:
	case VERBOSITY:
	case PRINTLEVEL:
	case DEBUGLEVEL:
	case STATEPRINT:
	case REDUNDANCY:
	case CALLBYNAME:
	case ASSIGNAHEAD:
	case PATH:
	case HELPDIR:
	case TIMESTEP:
        case RANDOM:    
        case RANDSEED:
	case RRANDOM:    
        case RRANDSEED:
        case MAXRANDLEN:    
            break;
        default:
            error_break1d ("mark: unknown node: %d", e -> type);
            break;
    }
}


void mark_mem(loc,n)
MEMORYPTR loc;
unsigned n;
{
    while (n-- > 0) {
        loc = curr_loc (loc);
        loc -> mark = T;
        if (is_indirect_mem (loc))
            mark_mem (loc -> val.m, 1);
        else {
            if (mem_defined (loc))
                mark (loc -> val.n);
            loc = next_loc (loc);
            loc -> mark = T;
            if (mem_defined (loc))
                mark (loc -> val.n);
	}
        loc += offset(1);
    }
}

void mark_phase()
{
    REDUCESTACKPTR rs;
    NODEPTR np;
    if (store_verbosity)
        fputs ("mark, ", stdout);
    for (rs = ReduceStack; rs <= ReduceStackTop; rs += 1) {
        mark (rs -> n);
    }
    for (np = TempNodeList; np != NULL; np = np -> next_node) {
        mark (np);
    }
    mark (nodep(EnvStack[EnvBase].env -> id_list)); /* AC pointer */
    mark_mem (GlobalMemory, GlobalMemUsed);
    mark (LastCommand);
    mark (SysWaitList);
    /* Following two added by BCM.  14Dec96 */
    mark(nodep(Path));
    mark(nodep(HelpDir));
}

int is_refd_block(bp)
BLOCKPTR bp;
{
    int r = F;
    MEMORYPTR loc = bp -> blk_base;
    int n = bp -> blk_size;
    while (n-- > 0) {
        loc = curr_loc (loc);
        if (loc -> mark) {
            r = T;
            loc -> mark = F;
        }
        loc = next_loc (loc);
        if (loc -> mark) {
            r = T;
            loc -> mark = F;
        }
        loc += offset(1);
    }
    return (r);
}

void reclaim_node(np)
NODEPTR np;
{
     switch (np -> type) {
        case STRING:
            free_mem (stringp (np) -> str);
            break;
        case MEMBLOCK:
	    LocalMemUsed -= blockp (np) -> blk_size;
            free_mem (blockp (np) -> blk_base);
            break;
        default:
            break;
    }
    np -> type = 0;
    np -> next_node = FreeNodeList;
    FreeNodeList = np;
    FreeNodes += 1;
}

void reclaim_phase()
{
    HEAP h;
    NODEPTR np,base,top;
    if (store_verbosity)
        fputs("reclaim, ",stdout);
    FreeNodeList = NULL;
    for (h = Heap; h != NULL; h = h -> next_block) {
        base = h -> base;
	top = h -> top;
	for (np = base; np < top; np++) {                 /* AC pointer */
            if ((np -> type == MEMBLOCK) && is_refd_block (blockp(np)))
                np -> mark = F;
            else if (np -> mark)
                np -> mark = F;
            else 
                reclaim_node (np);
	}
    }
}

void collect_garbage()
{
    if (store_verbosity)
        fputs ("***Tempura: garbage collection: ",stdout);
    mark_phase();
    reclaim_phase();
    if (store_verbosity) {
                printf ("done (%d%% in use).\n",
		HeapSize == 0 ? 100 : ((HeapSize-FreeNodes)*100)/HeapSize);
                fflush(stdout);
    }
}
