/* init.c:  initialisation
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: init.c,v 2.12 2007/05/03 09:48:46 acau Exp $
 */


#include "system.h"





/*******************\
*                   *
* GLOBAL VARIABLES  *
*                   *
\*******************/

/* Static values (very commonly used) */
struct b_node TrueNode = {T,BOOL,NULL,T};
NODEPTR True = nodep (&TrueNode);
struct b_node FalseNode = {T,BOOL,NULL,F};
NODEPTR False = nodep (&FalseNode);
struct i_node ZeroNode = {T,INT,NULL,0};
NODEPTR Zero = nodep (&ZeroNode);
struct i_node OneNode = {T,INT,NULL,1};
NODEPTR One = nodep (&OneNode);
struct node NilNode = {T,NIL,NULL,NULL,NULL,NULL};
NODEPTR Nil = &NilNode;
struct node EmptyNode = {T,EMPTY,NULL,NULL,NULL,NULL};
NODEPTR Empty = &EmptyNode;
struct node MoreNode = {T,MORE,NULL,NULL,NULL,NULL};
NODEPTR More = &MoreNode;
struct node SkipNode = {T,SKIP,NULL,NULL,NULL,NULL};
NODEPTR Skip = &SkipNode;
struct node RandomNode = {T,RANDOM,NULL,NULL,NULL,NULL};
NODEPTR Random = &RandomNode;
struct node RRandomNode = {T,RRANDOM,NULL,NULL,NULL,NULL};
NODEPTR RRandom = &RRandomNode;
struct node EndOfFileNode = {T,END,NULL,NULL,NULL,NULL};
NODEPTR EndOfFile = &EndOfFileNode;
/*MEMORY GlobalEmpty[MEMCELLSIZE] = {{T,0,T,F,0,nodep (&TrueNode)},*/
/* 			           {T,1,T,F,0,nodep (&TrueNode)}};*/
MEMORY GlobalEmpty[MEMCELLSIZE];
MEMORYPTR EmptyLoc = GlobalEmpty;

unsigned Ready = F;                  /* True after successful reduction. */
unsigned Retry = F;                  /* Re-reduce in current pass? */
unsigned NextFlag = F;               /* Reduce next value of expression? */
unsigned RedefFlag = F;              /* Redefinition allowed? */
unsigned Clock = 0;
unsigned AlarmGone = F;

/* counter for top-level commands */
unsigned Counter = 1;

/* Memory */
MEMORY GlobalMemory[offset (GLOBALMEMSIZE)];
unsigned LocalMemUsed = 0;
/*MEMORY TempMem[MEMCELLSIZE] = {{T,0,T,F,0,NULL},*/
/* 			       {T,1,T,F,0,NULL}};*/
MEMORY TempMem[MEMCELLSIZE];
MEMORYPTR TempLoc = TempMem;

unsigned CurrBit = 0;
unsigned NextBit = 1;
unsigned OneLevelMemFlag = F;        /* if on, use one-level memory */
unsigned TimeStampCheckFlag = T;     /* if on, check time stamp */

/* environment */
struct e_node GlobalEnv = {T,ENV,NULL,0,0,NULL};
ENVSTACK EnvStack[MAXENVSTACKSIZE] = {{GlobalMemory, &GlobalEnv}};
int EnvTop = 0;
int EnvBase = 0;

/* reduce-stack */
REDUCESTACK ReduceStack[MAXREDUCESTACKSIZE];
REDUCESTACKPTR ReduceStackTop = ReduceStack+(-1);
REDUCESTACKPTR ReduceStackMax = ReduceStack+(MAXREDUCESTACKSIZE-1);
NODEPTR LastCommand = nodep (&TrueNode);

/* tracing and debugging */
jmp_buf *BreakPoint;
int TempuraLevel = 0;
int BreakDepth = -1;
unsigned Break = F;
unsigned InterruptSet = F;
int PassLimitCounter = 0;
unsigned PassProgressFlag = F;
unsigned StateProgressFlag = T;
unsigned ForceOutput = F;
unsigned RunFlag = F;
int SyntaxErrorCount = 0;

/* statistics */
unsigned StateNumber = 0;
unsigned PassNumber = 0;
unsigned TotalPassCount = 0;
unsigned TotalReduceCount = 0;
unsigned TotalReduceTimes = 0;
unsigned StateReduceCount = 0;

unsigned PassReduceCount = 0;
unsigned TotalSuccessfulReduceCount = 0;
unsigned TotalSuccessfulReduceTimes = 0;
unsigned StateSuccessfulReduceCount = 0;
unsigned PassSuccessfulReduceCount = 0;
int ReduceDepth = 0;
int MaxReduceDepth = 0;

/* Heap */
HEAP Heap = NULL;
unsigned FreeNodes = 0;
unsigned HeapSize = 0;
NODEPTR FreeNodeList = NULL;
NODEPTR TempNodeList = NULL;

/* Symbol Table */
SYMBOLPTR SymbolTableTop = 0;
SYMBOLPTR SymbolTableMax = 0;
SYMBOLPTR HashTable[HASHTABLESIZE];/* The hash table. */

/* Wait list for system calls */
NODEPTR SysWaitList = NULL;

/* System variables (user definable). */
int PassLimit = -1;
int Verbosity = 1;
int PrintLevel = 2;
int DebugLevel = 1;
unsigned StatePrintFlag = T;
unsigned RedundancyFlag = F;
unsigned CallByNameFlag = F;
unsigned AssignAheadFlag = T;
STRINGPTR Path;       /* BCM 14Dec96 */
STRINGPTR HelpDir;    /* BCM 14Dec96 */
/* Previous two lines were as follows.  Had problems with garbage collection.
                BCM 14Dec96
char *Path;
char *HelpDir = "Help";
*/
int TimeStep = 0;
/* values related to random number generator AC 29Apr1999*/
int RandSeed = 1;      /* Seed for random numbers. BCM 16Jul97 */
int MaxRandLen = 10;   /* Maximum random interval length. BCM 22Jul97 */
int RandomVal = -1;     /* random value >=0 for current state.
 			   Initially "undefined" as -1. BCM 15Aug97 */
unsigned RandomValTime = 0; /* Time of random value. BCM 15Aug97 */
int RRandSeed = 1;      /* Seed for Random numbers. AC 20/09/2000 */
int RRandomVal = -1;     /* Random value >=0.
 			   Initially "undefined" as -1. AC 20/09/2000 */
unsigned RRandomValTime = 0; /* Time of Random value. AC 20/09/2000 */
/* Input/Output */
SYMBOL_OR_NODEPTR AuxVal;
int CharsOutput = 0;
int InputLineCount = 1;
FILE *InputFile; /*={stdin};*/
char InputText[MAXTOKENSIZE];
char *InputFilename = "Standard Input";


/* AC 29Apr1999: added
   111 random
   112 random_seed
   AC 20/09/2000: added
   113 Random
   114 Random_seed
   115 max_rand_len
   AC 04/09/2001: added
   116 implies
   AC 05/09/2001: added
   117 more
   118 chopstar
   AC 06/09/2001: added
   119 _if
*/
   
/* Parser */
unsigned Token = END;
unsigned Recover = F;
char *Keyword[MAXTOKEN + 1] = {
/*  0 */                ".",
/*  1 */                "nil",
/*  2 */                "bool",
/*  3 */                "integer",
/*  4 */                "lambda",
/*  5 */                "string",
/*  6 */                "list",
/*  7 */                "state",
/*  8 */                "static",
/*  9 */                "empty",
/* 10 */                "skip",
/* 11 */                "and",
/* 12 */                "_and",
/* 13 */                "or",
/* 14 */                "not",
/* 15 */                "if",
/* 16 */                "_call",
/* 17 */                "_call",
/* 18 */                "exists",
/* 19 */                "_exists",
/* 20 */                "process",
/* 21 */                "_process",
/* 22 */                "prefix",
/* 23 */                "_prefix",
/* 24 */                "forall",
/* 25 */                "the",
/* 26 */                "always",
/* 27 */                "sometimes",
/* 28 */                "halt",
/* 29 */                "init",
/* 30 */                "fin",
/* 31 */                "keep",
/* 32 */                "next",
/* 33 */                "wnext",
/* 34 */                "len",
/* 35 */                "_len",
/* 36 */                "<",
/* 37 */                "<=",
/* 38 */                "=",
/* 39 */                ">=",
/* 40 */                ">",
/* 41 */                "~=",
/* 42 */                "gets",
/* 43 */                "stable",
/* 44 */                "where",
/* 45 */                "is",
/* 46 */                "<-",
/* 47 */                ":=",
/* 48 */                ";",
/* 49 */                "_;",
/* 50 */                "while",
/* 51 */                "repeat",
/* 52 */                "for",
/* 53 */                "_for",
/* 54 */                "_for",
/* 55 */                "_for_times",
/* 56 */                "_for",
/* 57 */                "_for_in",
/* 58 */                "proj",
/* 59 */                "_proj",
/* 60 */                "input",
/* 61 */                "output",
/* 62 */                "+",
/* 63 */                "-",
/* 64 */                "*",
/* 65 */                "div",
/* 66 */                "mod",
/* 67 */                "**",
/* 68 */                "+",
/* 69 */                "-",
/* 70 */                "_make_list",
/* 71 */                "type",
/* 72 */                "struct",
/* 73 */                "_size",
/* 74 */                "_sub",
/* 75 */                "_sub_list",
/* 76 */                "list",
/* 77 */                "List",
/* 78 */                "ref",
/* 79 */                "deref",
/* 80 */                "loc",
/* 81 */                "run",
/* 82 */                "exit",
/* 83 */                "load",
/* 84 */                "define",
/* 85 */                "set",
/* 86 */                "showtrace",
/* 87 */                "showmem",
/* 88 */                "abort",
/* 89 */                "break",
/* 90 */                "return",
/* 91 */                "continue",
/* 92 */                "leap",
/* 93 */                "step",
/* 94 */                "peek",
/* 95 */                "format",
/* 96 */                "ascii",
/* 97 */                "help",
/* 98 */                "system",
/* 99 */                "_system",
/* 98 */		"pass_limit",
/*100 */		"verbosity",
/*102 */		"print_level",
/*103 */		"debug_level",
/*104 */		"print_states",
/*105 */		"no_redundancy",
/*106 */		"call_by_name",
/*107 */		"assign_ahead",
/*108 */		"path",
/*109 */		"help_dir",
/*110 */		"time_step",
/*111 */                "random",
/*112 */                "random_seed",
/*113 */                "Random",
/*114 */                "Random_seed",
/*115 */                "max_rand_len",
/*116 */                "implies",
/*117 */                "more",
/*118 */                "chopstar",
/*119 */                "_if",
/*120 */                "undefined",
/*121 */                "showstate",
/*122 */                ",",
/*123 */                "true",
/*124 */                "false",
/*125 */                "_name",
/*126 */                "then",
/*127 */                "else",
/*128 */                "in",
/*129 */                "times",
/*130 */                "_env",
/*131 */                "_bnd",
/*132 */                "to",
/*133 */                ":",
/*134 */                "(",
/*135 */                ")",
/*136 */                "{",
/*137 */                "}",
/*138 */                "[",
/*139 */                "]",
/*140 */                "|",
/*141 */                "until",
/*142 */                "_memory"
};

/* AC 29Apr1999 added
   111 assert_expr
   112 assert_expr
   113 assert_expr
   AC 20/09/2000 added
   114 assert_expr
   115 assert_expr
   AC 04/09/2001 added
   116 implies
   AC 05/09/2001 added
   117 more
   118 chopstar
   AC 06/09/2001 added
   119 _if
*/
PFI Reduces[MAXREDUCETOKEN + 1][3] = {
/*  0 END */
    {reduce_EOF,          reduce_EOF,         reduce_EOF},
/*  1 NIL */
    {assert_expr,         reduce_const,       get_ref},
/*  2 BOOL */
    {assert_true,         reduce_const,       get_ref},
/*  3 INT */
    {assert_expr,         reduce_const,       get_ref},
/*  4 LAMBDA */
    {assert_expr,         reduce_const,       get_ref},
/*  5 STRING */
    {assert_expr,         reduce_const,       get_ref},
/*  6 LIST */
    {assert_expr,         reduce_const,       get_ref},
/*  7 STATE */
    {assert_true,         reduce_id_val,      reduce_id_ref},
/*  8 STATIC */
    {assert_true,         reduce_id_val,      reduce_id_ref},
/*  9 EMPTY */
    {reduce_EMPTY_stmt,   reduce_EMPTY_val,   reduce_EMPTY_ref},
/* 10 SKIP */
    {reduce_SKIP,         assert_stmt,        assert_stmt},
/* 11 AND */
    {reduce_AND,          reduce_bin_op,      get_ref},
/* 12 iAND */
    {reduce_iAND,         assert_stmt,        assert_stmt},
/* 13 OR */
    {assert_true,         reduce_bin_op,      get_ref},
/* 14 NOT */
    {assert_true,         reduce_u_op,        get_ref},
/* 15 IF */
    {reduce_IF_stmt,      reduce_IF_val,      get_ref},
/* 16 CALL */
    {reduce_CALL,         reduce_CALL,        reduce_CALL},
/* 17 iCALL */
    {reduce_iCALL_stmt,   reduce_iCALL_val,   reduce_iCALL_ref},
/* 18 EXISTS */
    {reduce_EXISTS,       reduce_EXISTS,      reduce_EXISTS},
/* 19 iEXISTS */
    {reduce_iEXISTS_stmt, reduce_iEXISTS_val, get_ref},
/* 20 PROCESS */
    {reduce_PROCESS,      assert_stmt,        assert_stmt},
/* 21 iPROCESS */
    {reduce_iPROCESS,     assert_stmt,        assert_stmt},
/* 22 PREFIX */
    {reduce_PREFIX,       assert_stmt,        assert_stmt},
/* 23 iPREFIX */
    {reduce_iPREFIX,      assert_stmt,        assert_stmt},
/* 24 FORALL */
    {reduce_FORALL,       reduce_FORALL,      reduce_FORALL},
/* 25 THE */
    {assert_expr,         reduce_THE,         reduce_THE},
/* 26 ALWAYS */
    {reduce_ALWAYS,       assert_stmt,        assert_stmt},
/* 27 SOMETIME */
    {reduce_SOMETIME,     assert_stmt,        assert_stmt},
/* 28 HALT */
    {reduce_HALT,         assert_stmt,        assert_stmt},
/* 29 INIT */
    {reduce_INIT,         assert_stmt,        assert_stmt},
/* 30 FIN */
    {reduce_FIN,          assert_stmt,        assert_stmt},
/* 31 KEEP */
    {reduce_KEEP,         assert_stmt,        assert_stmt},
/* 32 NEXT */
    {reduce_NEXT_stmt,    reduce_NEXT_val,    reduce_NEXT_ref},
/* 33 WNEXT */
    {reduce_WNEXT,        assert_stmt,        assert_stmt},
/* 34 LEN */
    {reduce_LEN,          assert_stmt,        assert_stmt},
/* 35 iLEN */
    {reduce_iLEN,         assert_stmt,        assert_stmt},
/* 36 LT */
    {assert_stmt_to_true, reduce_bin_op,      get_ref},
/* 37 LE */
    {assert_stmt_to_true, reduce_bin_op,      get_ref},
/* 38 EQ */
    {reduce_EQ_stmt,      reduce_EQ_val,      get_ref},
/* 39 GE */
    {assert_stmt_to_true, reduce_bin_op,      get_ref},
/* 40 GT */
    {assert_stmt_to_true, reduce_bin_op,      get_ref},
/* 41 NEQ */
    {assert_stmt_to_true, reduce_EQ_val,      get_ref},
/* 42 GETS */
    {reduce_GETS,         assert_stmt,        assert_stmt},
/* 43 STB */
    {reduce_STB,          assert_stmt,        assert_stmt},
/* 44 WHERE */
    {reduce_WHERE,        assert_stmt,        assert_stmt},
/* 45 IS */
    {reduce_IS,           assert_stmt,        assert_stmt},
/* 46 ASSN */
    {reduce_ASSN,         assert_stmt,        assert_stmt},
/* 47 UASSN */
    {reduce_UASSN,        assert_stmt,        assert_stmt},
/* 48 SEMI */
    {reduce_SEMI,         assert_stmt,        assert_stmt},
/* 49 iSEMI */
    {reduce_iSEMI,        assert_stmt,        assert_stmt},
/* 50 WHILE */
    {reduce_WHILE,        assert_stmt,        assert_stmt},
/* 51 REPEAT */
    {reduce_REPEAT,       assert_stmt,        assert_stmt},
/* 52 FOR */
    {reduce_FOR,          assert_stmt,        assert_stmt},
/* 53 iFOR */
    {assert_expr,         reduce_iFOR,        assert_stmt},
/* 54 FORTIMES */
    {reduce_FORTIMES,     assert_stmt,        assert_stmt},
/* 55 iFORTIMES */
    {assert_expr,         reduce_iFORTIMES,   assert_stmt},
/* 56 FORIN */
    {reduce_FORIN,        assert_stmt,        assert_stmt},
/* 57 iFORIN */
    {assert_expr,         reduce_iFORIN,      assert_stmt},
/* 58 PROJ */
    {reduce_PROJ,         assert_stmt,        assert_stmt},
/* 59 iPROJ */
    {reduce_iPROJ,        assert_stmt,        assert_stmt},
/* 60 INPUT */
    {reduce_INPUT,        assert_stmt,        assert_stmt},
/* 61 OUTPUT */
    {reduce_OUTPUT,       reduce_OUTPUT,      assert_stmt},
/* 62 PLUS */
    {assert_expr,         reduce_bin_op,      get_ref},
/* 63 MINUS */
    {assert_expr,         reduce_bin_op,      get_ref},
/* 64 MULT */
    {assert_expr,         reduce_bin_op,      get_ref},
/* 65 DIV */
    {assert_expr,         reduce_bin_op,      get_ref},
/* 66 MOD */
    {assert_expr,         reduce_bin_op,      get_ref},
/* 67 POWER */
    {assert_expr,         reduce_bin_op,      get_ref}, 
/* 68 UPLUS */
    {assert_expr,         reduce_u_op,        get_ref},
/* 69 UMINUS */
    {assert_expr,         reduce_u_op,        get_ref},
/* 70 MAKELIST */
    {assert_expr,         reduce_MAKELIST,    get_ref},
/* 71 TYPE */
    {assert_expr,         reduce_TYPE,        get_ref},
/* 72 STRUCT */
    {assert_expr,         reduce_STRUCT_val,  reduce_STRUCT_ref},
/* 73 SIZE */
    {assert_expr,         get_val,            reduce_SIZE},
/* 74 SUB */
    {assert_true,         get_val,            reduce_SUB},
/* 75 SUBLIST */
    {assert_expr,         get_val,            reduce_SUBLIST},
/* 76 FIXLIST */
    {reduce_list,         assert_stmt,        assert_stmt},
/* 77 VARLIST */
    {reduce_list,         assert_stmt,        assert_stmt},
/* 78 REFER */
    {assert_expr,         reduce_REFER,       get_ref},
/* 79 DEREF */
    {assert_expr,         get_val,            reduce_DEREF},
/* 80 LOC */
    {assert_expr,         reduce_const,       get_ref},
/* 81 RUN */
    {reduce_RUN,          assert_stmt,        assert_stmt},
/* 82 EXIT */
    {reduce_EXIT,         assert_stmt,        assert_stmt},
/* 83 LOAD */
    {reduce_LOAD,         assert_stmt,        assert_stmt},
/* 84 DEFINE */
    {reduce_DEFINE,       assert_stmt,        assert_stmt},
/* 85 SET */
    {reduce_SET,          assert_stmt,        assert_stmt},
/* 86 SHOWTRACE */
    {reduce_SHOWTRACE,    assert_stmt,        assert_stmt},
/* 87 SHOWMEM */
    {reduce_SHOWMEM,      assert_stmt,        assert_stmt},
/* 88 ABORT */
    {reduce_ABORT,        assert_stmt,        assert_stmt},
/* 89 BREAK */
    {reduce_BREAK,        assert_stmt,        assert_stmt},
/* 90 RETURN */
    {reduce_RETURN,       assert_stmt,        assert_stmt},
/* 91 CONTINUE */
    {reduce_CONTINUE,     assert_stmt,        assert_stmt},
/* 92 LEAP */
    {reduce_LEAP,         assert_stmt,        assert_stmt},
/* 93 STEP */
    {reduce_STEP,         assert_stmt,        assert_stmt},
/* 94 PEEK */
    {reduce_PEEK,         assert_stmt,        assert_stmt},
/* 95 FORMAT */
    {reduce_FORMAT,       assert_stmt,        assert_stmt},
/* 96 ASCII */
    {assert_expr,         reduce_ASCII,       get_ref},
/* 97 HELP */
    {reduce_HELP,         assert_stmt,        assert_stmt},
/* 98 SYSTEM */
    {reduce_SYSTEM,       assert_stmt,        assert_stmt},
/* 99 iSYSTEM */
    {reduce_iSYSTEM,      assert_stmt,        assert_stmt},
/*100 PASSLIMIT */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*101 VERBOSITY */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*102 PRINTLEVEL */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*103 DEBUGLEVEL */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*104 STATEPRINT */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*105 REDUNDANCY */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*106 CALLBYNAME */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*107 ASSIGNAHEAD */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*108 PATH */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*109 HELPDIR */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*110 TIMESTEP */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*111 RANDOM */
    {assert_expr,         reduce_RANDOM,      get_ref},
/*112 RANDSEED */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*113 RRANDOM */
    {assert_expr,         reduce_RRANDOM,     get_ref},
/*114 RRANDSEED */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*115 MAXRANDLEN */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*116 IMPLIES */
    {reduce_IMPLIES,      reduce_IMPLIES,     reduce_IMPLIES},
/*117 MORE */
    {reduce_MORE_stmt,    reduce_MORE_val,    get_ref},
/*118 CHOPSTAR */
    {reduce_CHOPSTAR,     assert_stmt,        assert_stmt},
/*119 iIF */
    {assert_expr,         reduce_iIF_val,     get_ref},
/*120 UNDEFINED */
    {assert_expr,         reduce_UNDEFINED,       get_ref},
/*121 RESET */
    {reduce_SHOWSTATE,     assert_stmt,        assert_stmt},
/*122 COMMA */
    {noop,                noop,               noop}
};







/***********************************************\
*                                               *
*  Enter reserved words into the symbol table.  *
*                                               *
\***********************************************/

void install_keywords()
{
    int code;
    for (code = 1; code <= MAXTOKEN; code += 1) {
        if (isalpha((int)*Keyword[code]))
            install_keyword(code,Keyword[code]);
    }

    /* Alternative forms for operators. */

    install_keyword(DO,strsave("do"));
    install_keyword(ALWAYS,strsave("box"));
    install_keyword(SOMETIME,strsave("sometime"));
    install_keyword(LEAP,strsave("jump"));
}


/* initialize tempura */
void tempura_init ()
{
    catch_signals ();
    install_keywords ();
    init_sysvars ();
    new_loc (TempLoc, STATIC);
    put_mem (new_loc (EmptyLoc, STATIC), True);
}
