/* io.c: input and output routines 
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: io.c,v 2.12 2007/05/03 09:48:46 acau Exp $
 */


#include "system.h"



/******************************\
*                              *
*  Auxiliary output routines.  *
*                              *
\******************************/

void output_int(n,f)
int n;
int f;
{
    char    s[MAXINTSIZE],
           *p = s + MAXINTSIZE - 1;
    int     sign;
    *p = '\0';
    if ((sign = n) < 0)
        n = -n;
    do {
        *--p = n % 10 + '0';
        n = n / 10;
    } while (n > 0);
    if (sign < 0)
        *--p = '-';
    outputf_str (p, f);
}

void output_str(s)
char *s;
{
    while (*s != '\0')
        output_char (*s++);
}

void outputf_str(s,f)
char *s;
int f;
{
    int     len = strlen (s);
    spaces (f - len);
    output_str (s);
    spaces (-f - len);
}

void output_lambda(e)
NODE2PTR e;
{
    output_expn(e,-1);
}

/* Find print-name of type. */
char *print_name(e)
NODEPTR e;
{
    unsigned t = e -> type;
    static char *undef = "undefined";
    return ((0 <= t && t <= MAXTOKEN) ? Keyword[t] : undef);
}

void message(text)
char *text;
{
    printf ("***Tempura: %s.\n", text);
    fflush(stdout);
}

void spaces(n)
int n;
{
    while (n-- > 0)
        output_char (' ');
}

void output_env(env)
ENVPTR env;
{
    NAMEPTR id = env -> id_list;
    while (id != NULL) {
        output_str (id -> name);
        if ((id = id -> next_node) != NULL)
            output_str (", ");
    }
}

void output_expn(e,level)
NODEPTR e;
int level;
{

    if (e == NULL) {
/*        output_str ("()");*/
        return;
    }

    if (level == 0) {
        output_str ("...");
        return;
    }

    switch (e -> type) {

        case STATE:
        case STATIC:
            output_str (namep (e) -> name);
            break;
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case LIST:
        case LOC:
            output_val (e);
            break;
        case BND:
            output_expn (nodep(bndp (e) -> env), level); /* AC pointer */
            break;
        case ENV:
            output_env (envp(e)); /* AC pointer */
            break;
        case MEMBLOCK:
            output_str (Keyword[e -> type]);
            output_int ((signed)blockp (e) -> blk_size, -1);
            break;
        case SKIP:
        case EMPTY:
        case MORE:    
            output_str (Keyword[e -> type]);
            break;
        case LT:
        case LE:
        case GE:
        case GT:
        case EQ:
        case NEQ:
        case PLUS:
        case MINUS:
        case MULT:
        case DIV:
        case POWER:
        case AND:
        case OR:
        case IMPLIES:
        case MOD:
            output_char ('(');
            output_expn (e -> c1, level);
	    space;
            output_str (Keyword[e -> type]);
	    space;
            output_expn (e -> c2, level - 1);
            output_char (')');
            break;
        case SEMI:
        case WHERE:
        case PROJ:
        case GETS:
        case IS:
        case ASSN:
        case UASSN:
            output_expn (e -> c1, level);
	    space;
            output_str (Keyword[e -> type]);
	    space;
            output_expn (e -> c2, level - 1);
            break;
        case NOT:
        case UPLUS:
        case UMINUS:
        case REFER:
        case DEREF:    
        case PROCESS:
        case iPROCESS:
        case PREFIX:
        case iPREFIX:
        case ALWAYS:
        case SOMETIME:
        case HALT:
        case INIT:
        case FIN:
        case CHOPSTAR:    
        case KEEP:
        case NEXT:
        case WNEXT:
        case LEN:
        case iLEN:
        case STB:
        case INPUT:
        case OUTPUT:
        case RUN:
        case TYPE:
        case ASCII:
        case STRUCT:
        case LOAD:
        case SYSTEM:
            output_str (Keyword[e -> type]);
	    space;
            output_expn (e -> c1, level);
            break;
        case FORMAT:
            output_str (Keyword[e -> type]);
	    output_char ('(');;
            output_expn (e -> c1, level);
	    output_char (')');;
            break;
        case IF:
        case iIF:    
            output_str (Keyword[e -> type]);
            space;
            output_expn (e -> c1, level);
            output_str (" then {");
            output_expn (e -> c2, level - 1);
            if (e -> c3 != NULL) {
                output_str ("} else {");
                output_expn (e -> c3, level - 1);
            }
            output_char ('}');
            break;
        case CALL:
        case iCALL:
            output_expn (e -> c1, level);
            output_char ('(');
            output_expn (e -> c2, level);
            output_char (')');
            break;
        case LAMBDA:
            output_str (Keyword[e -> type]);
            output_str (" (");
            output_expn (e -> c1, level);
            output_str (") : {");
            output_expn (e -> c2, level - 1);
            output_char ('}');
            break;
        case EXISTS:
        case FORALL:
            output_str (Keyword[e -> type]);
            space;
            output_expn (e -> c1, level);
            if (e -> c2 != NULL) {
                output_char ('<');
                output_expn (e -> c2, level);
            }
            output_str (": {");
            output_expn (e -> c3, level - 1);
            output_char ('}');
            break;
        case WHILE:
            output_str (Keyword[e -> type]);
            output_char (' ');
            output_expn (e -> c1, level);
            output_str (" do {");
            output_expn (e -> c2, level - 1);
            output_char ('}');
            break;
        case FOR:
            output_str (Keyword[e -> type]);
            space;
            output_expn (e -> c1, level);
            output_char ('<');
            output_expn (e -> c2, level);
            output_str (" do {");
            output_expn (e -> c3, level - 1);
            output_char ('}');
            break;
        case FORIN:
            output_str (Keyword[e -> type]);
            space;
            output_expn (e -> c1, level);
            output_str (" in ");
            output_expn (e -> c2, level);
            output_str (" do {");
            output_expn (e -> c3, level - 1);
            output_char ('}');
            break;
        case FORTIMES:
            output_str ("for ");
            output_expn (e -> c1, level);
            output_str (" times do {");
            output_expn (e -> c2, level - 1);
            output_char ('}');
            break;
        case DEFINE:
        case SET:
            output_str (Keyword[e -> type]);
            space;
            output_expn (e -> c1, level);
            output_char ('=');
            output_expn (e -> c2, level - 1);
            break;
        case COMMA:
	    output_expn (e -> c1, level);
            if (e -> c2 != NULL) {
                output_char (',');
		output_expn (e -> c2, level);
            }
            break;
        case iEXISTS:
            output_str (Keyword[e->type]);
            space;
            output_expn (e -> c1, level);
            output_str (" : {");
            output_expn (e -> c2, level - 1);
            output_char ('}');
            break;
        case iAND:
        case iSEMI:
        case iPROJ:
            output_expn (e -> c2, level);
            if (e -> c3 != NULL) {
                space;
                output_str (Keyword[e -> type]);
                space;
                output_expn (e -> c3, level - 1);
            }
            break;
        case MAKELIST:
            output_char ('[');
            output_expn (e -> c1, level - 1);
            output_char (']');
            break;
        case SUB:
            output_expn (e -> c1, level);
            output_char ('[');
            output_expn (e -> c2, level - 1);
            output_char (']');
            break;
        case SIZE:
            output_char ('|');
            output_expn (e -> c1, level);
            output_char ('|');
            break;
        case SUBLIST:
            output_expn (e -> c1, level);
            output_char ('[');
            output_expn (e -> c2, level - 1);
            output_str ("..");
            output_expn (e -> c3, level - 1);
            output_char (']');
            break;
        case FIXLIST:
        case VARLIST:
        case iSYSTEM:
            output_str (Keyword[e -> type]);
            output_char ('(');
            output_expn (e -> c1, level - 1);
            output_char (',');
            output_expn (e -> c2, level - 1);
            output_char (')');
            break;    
        default:
            output_str (print_name (e));
            break;
    }
}
