/* lexer.c: lexical analyser
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: lexer.c,v 2.11 2007/05/03 09:48:46 acau Exp $
 */


#include "system.h"



#define BS '\\'
#define ERROR 256
/*********************\
*                     *
*  Lexical analyser   *
*                     *
\*********************/

int read_token()
{
    register int     c;
    register char   *p = InputText;
    register char   *pmax = InputText + MAXTOKENSIZE - 1;
    char     *strsave ();

    AuxVal.n = EndOfFile;
    while (isspace (*p = c = getc (InputFile)) ||
    	   (c == '/' && isnextchar ('*'))) {
        if (c == '/') {         /* Discard comment. */
            do {
                while ((c = getc (InputFile)) != '*') {
                   if (c == EOF) {
                       syntax_error (IDN);
                       return (END);
                   }
                   if (c == '\n')
                        InputLineCount++;
                }
            } while (!isnextchar ('/'));
        }
        if (c == '\n')
            InputLineCount++;
    }
    InputText[1] = '\0';

    if (isalpha (c)) {
        do {
            *(p < pmax ? p++ : pmax) = c;
        } while ((c = getc (InputFile)) != EOF && (isalnum (c) || (c == '_')));
        ungetc (c, InputFile);
        *(p < pmax ? p : pmax) = '\0';
        if (p > pmax)
            warn1 ("identifier truncated: \"%s\"", InputText);
        if ((AuxVal.s = lookup (InputText)) == NULL) {
            AuxVal.s = install (NAME, strsave (InputText));
        }
        return (AuxVal.s -> type);
    }

    if (isdigit (c)) {
        do {
            *(p < pmax ? p++ : pmax) = c;
        } while ((c = getc (InputFile)) != EOF && isdigit (c));
        ungetc (c, InputFile);
        *(p < pmax ? p : pmax) = '\0';
        AuxVal.n = ((c = atoi (InputText)) == 0) ?
            Zero :
            (c == 1) ?
            One :
            int_node (c);
        return (INT);
    }

    if (c == '"') {
        while ((c = getc (InputFile)) != '"') {
            if (c == EOF) {
                *p = '\0';
                syntax_error (MQS);
                return (END);
            }
            *(p < pmax ? p++ : pmax) = backslash (c);
        }
        *(p < pmax ? p : pmax) = '\0';
        if (p >= pmax)
            syntax_error (STL);
        AuxVal.n = string_node (strsave (InputText));
        return (STRING);
    }

    switch (c) {
        case EOF:
            return (END);
        case '.':
            return (follow ('.', TO, END));
        case '>':
            return (follow ('=', GE, GT));
        case '<':
            return (follow ('-', ASSN, follow ('=', LE, LT)));
        case '~':
            return (follow ('=', NEQ, NOT));
        case BS:
            return (follow ('/', OR, ERROR));
        case ':':
            return (follow ('=', UASSN, DO));
        case '/':
            return (follow (BS, AND, DIV));
        case '*':
            return (follow ('*', POWER, MULT));
        case '+':
            return (PLUS);
        case '-':
            return (MINUS);
        case '=':
            return (EQ);
        case ';':
            return (SEMI);
        case '(':
            return (LPAREN);
        case ')':
            return (RPAREN);
        case '{':
            return (LBRACE);
        case '}':
            return (RBRACE);
        case '[':
            return (LBRACK);
        case ']':
            return (RBRACK);
        case '|':
            return (BAR);
        case ',':
            return (COMMA);
        case '@':
            return (REFER);
        case '!':
	    AuxVal.n = LastCommand;
            return (follow ('!', END, DEREF));
        case '?':
            return (HELP);
        default:
            break;
    }
    return (ERROR);
}

int isnextchar (expect)
int expect;
{
    int     c;
    if ((c = getc (InputFile)) == expect) {
        return (T);
    }
    else {
        ungetc (c, InputFile);
        return (F);
    }
}

int follow (expect, ifyes, ifno)
int expect, ifyes, ifno;
{
    int     c = getc (InputFile);
    if (c == expect) {
        InputText[1] = 'c';
        InputText[2] = '\0';
        return (ifyes);
    }
    ungetc (c, InputFile);
    return (ifno);
}

int backslash(c)
int c;
{
    if (c == BS) {
        switch (c = getc (InputFile)) {
            case 'n':
                return ('\n');
            case 't':
                return ('\t');
            case 'b':
                return ('\b');
            case 'r':
                return ('\r');
            case 'f':
                return ('\f');
            default:
                return (c);
        }
    }
    return (c);
}

SYMBOLPTR symballoc()           /* allocate a new symbol */
{
    if (SymbolTableTop >= SymbolTableMax)
        more_symbol_table (MAXSYMBOLTABLESIZE);
    return (SymbolTableTop++);
}

unsigned hash(s)                         /* form hash value for string s */
char *s;
{
    register unsigned hashval;

    for (hashval = 0; *s != '\0'; )
        hashval = (hashval << 1) ^ (*s++);
    return(hashval % HASHTABLESIZE);
}

SYMBOLPTR lookup(s)
char *s;
{
    register SYMBOLPTR sp;
    for (sp = HashTable[hash (s)]; sp != NULL; sp = sp -> next)
        if (strcmp (s, sp -> name) == 0)
            return (sp);
    return (NULL);
}

SYMBOLPTR install(type,name)
int type;
char *name;
{
    register SYMBOLPTR sp;
    unsigned hashval;
    sp = symballoc ();
    sp -> type = type;
    sp -> name = name;
    hashval = hash (sp -> name);
    sp -> next = HashTable[hashval];
    HashTable[hashval] = sp;
    return (sp);
}


void install_keyword(type,name)
int type;
char *name;
{
    install (type, name);
}
/*
uninstall(s)
char *s;
{
    SYMBOLPTR * spp;
    for (spp = &HashTable[hash (s)]; *spp != NULL; spp = &((*spp) -> next)) {
        if (strcmp (s, (*spp) -> name) == 0) {
            *spp = (*spp) -> next;
            return (T);
        }
    }
    return(F);
}
*/

/* End of lexer.c. */
