/* mem.c: Memory management routines 
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: mem.c,v 2.10 2001/09/10 15:17:33 acau Exp $
 */


#include "system.h"





/*******************************\
*                               *
*  Memory management routines   *
*                               *
\*******************************/

/* new_loc:  initialise a memory cell. */
MEMORYPTR new_loc (loc,attrib)
register MEMORYPTR loc;
int attrib;
{
    register MEMORYPTR loc1;
    loc -> mem_no = 0;
    (loc1 = loc + 1) -> mem_no = 1;
    loc -> mark = loc1 -> mark = F;
    loc -> is_static = loc1 -> is_static = (attrib == STATIC);
    loc -> is_indirect = loc1 -> is_indirect = F;
    loc -> time = loc1 -> time = UNDEF;
    loc -> val.n = loc1 -> val.n = NULL;
    return (loc);
}

BLOCKPTR mem_block(attrib,size)
int attrib;
register unsigned int size;
{
    BLOCKPTR blk;
    register MEMORYPTR loc;

    blk = allocate_mem (size);
    loc = blk -> blk_base;
    while (size-- > 0) {
        new_loc (loc, attrib);
        loc += offset (1);
    }
    return (blk);
}

BLOCKPTR allocate_mem(size)
unsigned int size;
{
    MEMORYPTR base_loc;

    if (size == 0)
        return (block_node (0, NULL));
    base_loc = memoryp (allocate (offset (size), sizeof (MEMORY)));
    LocalMemUsed += size;
    return (block_node (size, base_loc));
}

int mem_defined(loc)
register MEMORYPTR loc;
{
    dereference (loc);
    return (loc -> is_static ?
            (loc -> time != UNDEF) :
            (loc -> time == (int)(is_next (loc) ? Clock + 1 : Clock)));
}

NODEPTR get_mem (loc)
register MEMORYPTR loc;
{
    dereference (loc);
    return (loc -> val.n);
}

NODEPTR curr_list(list)
LISTPTR list;
{
    return ((is_next (list -> size) || is_next (list -> base)) ?
             list_node (list -> blk,
                        curr_loc (list_size (list)),
                        curr_loc (list_base (list))) :
             nodep (list));
}

NODEPTR next_list(list)
LISTPTR list;
{
    return ((is_curr (list -> size) || is_curr (list -> base)) ?
             list_node (list -> blk,
                        next_loc (list_size (list)),
                        next_loc (list_base (list))) :
             nodep (list));
}

MEMORYPTR list_size1(list)
LISTPTR list;
{
    MEMORYPTR loc = list -> size;
    dereference (loc);
    return (loc);
}

MEMORYPTR list_base1(list)
LISTPTR list;
{
    MEMORYPTR loc = list -> base;
    dereference (loc);
    return (loc);
}


NODEPTR put_mem (loc,val)
register MEMORYPTR loc;
NODEPTR val;
{
    NODEPTR adj_val;
    dereference (loc);
    if (is_next (loc)) {
        loc -> val.n = adj_val = next_val (val);
        loc -> time = Clock + 1;
        if (loc -> is_static) {
            (loc = curr_loc (loc)) -> val.n = curr_val (val);
            loc -> time = Clock;
	}
    }
    else {
        loc -> val.n = adj_val = curr_val (val);
        loc -> time = Clock;
	if (loc -> is_static) {
	    (loc = next_loc (loc)) -> val.n = next_val (val);
            loc -> time = Clock + 1;
	}
    }
    return (adj_val);
}

int is_static_mem(loc)
register MEMORYPTR loc;
{
    dereference (loc);
    return (loc -> is_static);
}

void make_static_mem(loc)
register MEMORYPTR loc;
{
    dereference (loc);
    next_loc (loc) -> is_static = curr_loc (loc) -> is_static = T;
}

void make_indirect_mem(loc,val)
MEMORYPTR loc;
register MEMORYPTR val;
{
    MEMORYPTR cloc, nloc;
    dereference (val);
    cloc = curr_loc (loc);
    nloc = next_loc (loc);
    cloc -> val.m = val;
    nloc -> val.m = is_curr (val) ? next_loc (val) : curr_loc (val);
    cloc -> is_indirect = nloc -> is_indirect = T;
}

void erase_mem(loc)
register MEMORYPTR loc;
{
    dereference(loc);
    loc -> time = UNDEF;
    loc -> val.n = NULL;
}



/******************************\
*                              *
*  Routines for the empty-flag *
*                              *
\******************************/

int empty_defined()
{
    return (mem_defined (EmptyLoc));
}

void put_empty(val)
NODEPTR val;
{
    if (mem_defined (EmptyLoc)) {
        assert0 (equal_val (val, get_mem (EmptyLoc)),
                "attempt to re-assign interval length");
    }
    else
        put_mem (EmptyLoc, val);
}

NODEPTR get_empty()
{
    return (EmptyLoc -> val.n);
}



/*************************\
*                         *
*  Environment handling   *
*                         *
\*************************/

void push_env(bnd)
BNDPTR bnd;
{
    if (EnvTop >= MAXENVSTACKSIZE)
        error_break0("environment stack full");
    else {
        EnvTop += 1;
        EnvStack[EnvTop].env = bnd -> env;
        EnvStack[EnvTop].env_base = bnd -> blk -> blk_base;
    }
}

void pop_env()
{
    EnvTop -= 1;
}

void push_display(bnd)
BNDPTR bnd;
{
    register unsigned l = bnd -> env -> env_level;
    unsigned newenvbase;
    if (EnvTop + l + 1 >= MAXENVSTACKSIZE)
        error_break0 ("environment stack full");
    else {
	newenvbase = (EnvTop += 1);
        for (; l > 0; l -= 1) {
            EnvStack[EnvTop].env = EnvStack[EnvBase].env;
            EnvStack[EnvTop].env_base = EnvStack[EnvBase].env_base;
            EnvTop += 1;
            EnvBase += 1;
	}
        EnvStack[EnvTop].env_base = bnd -> blk -> blk_base;
        EnvStack[EnvTop].env = bnd -> env;
        EnvBase = newenvbase;
    }
}

void pop_display()
{
    EnvTop = --EnvBase;
    EnvBase = EnvBase - (EnvStack[EnvTop].env -> env_level);
}


