/* parser.c: parser routines 
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: parser.c,v 2.12 2007/05/03 09:48:45 acau Exp $
 */

#include "system.h"


NODEPTR parse(stream)
FILE *stream;
{
    NODEPTR p;
    InputFile = stream;
    SyntaxErrorCount = 0;
    Recover = F;
    next_token ();
    p = resolve_refs (rexpr (0));
    while (Token != END) {
        syntax_error (END);
        next_token ();
    }
    return (SyntaxErrorCount == 0 ? p : True);
}

NODEPTR resolve_refs (e)
NODEPTR e;
{
    if (e == NULL || e == LastCommand)
        return (e);
    switch (e -> type) {
        case NAME:
	    return (nodep(lookup_env (symbolp (e) -> name)));
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case LOC:
            return (e);
        case EXISTS:
        case FORALL:
        case FORIN:
        case FOR:
	    push_scope (e -> c1);
	    e -> c3 = resolve_refs (e -> c3);
	    e -> c2 = resolve_refs (e -> c2);
	    e -> c1 = pop_scope ();
	    return (e);
        case LAMBDA:
	    push_scope (e -> c1);
	    e -> c2 = resolve_refs (e -> c2);
	    e -> c1 = pop_scope ();
	    return (e);
        case FORTIMES:
        case IF:
        case SUBLIST:
            e -> c3 = resolve_refs (e -> c3);
        case PROJ:
        case COMMA:
        case DEFINE:
        case WHILE:
        case REPEAT:
        case LT:
        case LE:
        case EQ:
        case GE:
        case GT:
        case NEQ:
        case PLUS:
        case MINUS:
        case MULT:
        case DIV:
        case POWER:
        case SEMI:
        case WHERE:
        case AND:
        case OR:
        case IMPLIES:    
        case GETS:
        case IS:
        case MOD:
        case ASSN:
        case UASSN:
        case CALL:
        case SUB:
        case FIXLIST:
        case VARLIST:
        case RUN:
        case SET:
        case SHOWMEM:
            e -> c2 = resolve_refs (e -> c2);
        case NOT:
        case UPLUS:
        case UMINUS:
        case PROCESS:
        case PREFIX:
        case ALWAYS:
        case SOMETIME:
        case HALT:
        case INIT:
        case FIN:
        case CHOPSTAR:    
        case KEEP:    
        case NEXT:
        case WNEXT:
        case LEN:
        case STB:
        case STRUCT:
        case INPUT:
        case OUTPUT:
        case PEEK:
        case HELP:
        case SYSTEM:
        case TYPE:
        case UNDEFINED:
        case ASCII:
        case FORMAT:
        case SIZE:
        case MAKELIST:
        case REFER:
        case DEREF:
        case LOAD:
            e -> c1 = resolve_refs (e -> c1);
        default:
            return (e);
    }
}

NODEPTR rexpr(n)
int n;
{
    NODEPTR p = rbexpr ();
    unsigned tk;
    int lp = 0;
    int rp = 0;
    do {
        switch (tk = Token)  {
            case LPAREN:
                p = rcall (symbolp(p)); /* AC pointer */
                continue;
            case LBRACK:
                p = rsublist (p);
                continue;
            case WHERE:
                lp = 36;
                rp = 15;
/*                lp = rp = 15;*/
                break;
            case SEMI:
            case PROJ:
                lp = rp = 20;
                break;
            case IMPLIES:
                lp = rp = 22;
                break;    
            case OR:
                lp = rp = 25;
                break;
            case AND:
                lp = rp = 30;
                break;
            case GETS:
            case IS:
            case ASSN:
            case UASSN:
                lp = 50; rp = 49;
                break;
            case EQ:
            case LT:
            case LE:
            case GE:
            case GT:
            case NEQ:
                lp = 55; rp = 54;
                break;
            case PLUS:
            case MINUS:
                lp = rp = 60;
                break;
            case MULT:
            case DIV:
            case MOD:
                lp = rp = 65;
                break;
            case POWER:
                lp = rp = 70;
                break;
            default:
                return (p);
        }
        Recover = F;
        if (n >= lp)
            return (p);
        next_token ();
        p = node2 (tk, p, rexpr (rp));
    } while (T);
}

NODEPTR rbexpr()
{
    NODEPTR p = AuxVal.n;
    unsigned tk;
    if ((tk = Token) == END)
        return (p);
    next_token ();
    switch (tk) {
        case LPAREN:
            p = rexpr (0);
            assert_syntax (RPAREN);
            return (p);
        case LBRACE:
            p = rexpr (0);
            assert_syntax (RBRACE);
            return (p);
        case NIL:
            return (Nil);
        case TRUE:
            return (True);
        case FALSE:
            return (False);
        case INT:
        case STRING:
        case NAME:
	case PASSLIMIT:
	case VERBOSITY:
	case PRINTLEVEL:
	case DEBUGLEVEL:
	case STATEPRINT:
	case REDUNDANCY:
	case CALLBYNAME:
	case ASSIGNAHEAD:
	case PATH:
	case HELPDIR:
	case TIMESTEP:
            return (p);
        case LBRACK:
            return (rlist ());
        case FIXLIST:
        case VARLIST:
            return (rlistdef (tk));
        case EMPTY:
            return (Empty);
        case MORE:
            return (More);       /* AC added 05/09/2001 */    
        case SKIP:
            return (Skip);
        case RANDOM:             /* Added AC 29Apr1999 */
            return (Random);   
        case RRANDOM:            /* Added AC 20/09/2000 */
            return (RRandom);
        case EXISTS:
            return (rexists ());
        case FORALL:
            return (rforall ());
        case IF:
            return (rif ());
        case WHILE:
            return (rwhile ());
        case REPEAT:
            return (rrepeat ());
        case CHOPSTAR:
            return (rchopstar ());
        case KEEP:
            return (rkeep ());    
        case FOR:
            return (rfor ());
        case LAMBDA:
            return (rlambda ());
        case DEFINE:
            return (rdefine ());
        case SET:
            return (rset ());
        case FORMAT:
            return (rformat ());
        case RUN:
            return (node2 (tk, rexpr (5), True));
        case PROCESS:
        case PREFIX:
        case ALWAYS:
        case SOMETIME:
        case HALT:
        case FIN:
        case INIT:
        case STB:
        case LEN:
            return (node1 (tk, rexpr (40)));
        case INPUT:
        case OUTPUT:
        case PEEK:
            return (node1 (tk, rexpr (45)));
        case NOT:
        case NEXT:
        case WNEXT:
        case TYPE:
        case UNDEFINED:
        case ASCII:
        case STRUCT:
        case LOAD:
        case SYSTEM:
            return (node1 (tk, rexpr (75)));
        case MINUS:
            return (node1 (UMINUS, rexpr (75)));
        case PLUS:
            return (node1 (UPLUS, rexpr (75)));
        case REFER:
        case DEREF:
            return (node1 (tk, rexpr (80)));
        case BAR:
            p = rexpr (0);
            assert_syntax (BAR);
            return (node1 (SIZE, p));
        case SHOWMEM:
            return (node2(SHOWMEM, rexpr(0), rexpr(0)));
                              /* AC 1999-12-01: with this showmem works */
        case EXIT:
        case ABORT:
        case BREAK:
        case RETURN:
        case LEAP:
        case STEP:
        case SHOWSTATE:
        case CONTINUE:
        case SHOWTRACE:
            return (p);
        case HELP:
            return (rhelp ());
        default:
            syntax_error (-1);
            return (True);
    }
}

NODEPTR rexists()
{
    NODEPTR e1,e2;
    NODEPTR nl = rnamelist ();
    switch (Token) {
    case LT:
        if (nl == NULL || nl -> type == COMMA)
	    syntax_error(N1);
        next_token ();
        e1 = rexpr (0);
        assert_syntax (DO);
        e2 = rexpr (35);
        return (node3 (EXISTS, nl, e1, e2));
    case DO:
        next_token ();
        e2 = rexpr (35);
        return (node3 (EXISTS, nl, NULL, e2));
    default:
        syntax_error (-1);
        return (True);
    }
}

NODEPTR rforall()
{
    NODEPTR e1,e2;
    NODEPTR id = rname ();
    switch (Token) {
    case LT:
        next_token ();
        e1 = rexpr (0);
        assert_syntax (DO);
        e2 = rexpr (35);
        return (node3 (FORALL, id, e1, e2));
    default:
        syntax_error (-1);
        return (True);
    }
}

NODEPTR rif()
{
    NODEPTR b, e1, e2;
    b = rexpr (0);
    assert_syntax (THEN);
    e1 = rexpr (35);
    if (Token == ELSE) {
        next_token ();
        e2 = rexpr (35);
    }
    else
        e2 = True;
    return (node3 (IF, b, e1, e2));
}

NODEPTR rwhile()
{
    NODEPTR b;
    b = rexpr (0);
    assert_syntax (DO);
    return (node2 (WHILE, b, rexpr (35)));
}

NODEPTR rrepeat()
{
    NODEPTR e;
    e = rexpr (35);
    assert_syntax (UNTIL);
    return (node2 (REPEAT, e, rexpr (0)));
}

NODEPTR rchopstar()   
{
    return (node1 (CHOPSTAR, rexpr (35)));
}

NODEPTR rkeep()   
{
    return (node1 (KEEP, rexpr (35)));
}

NODEPTR rfor()
{
    NODEPTR e1, e2;
    e1 = rexpr (0);
    switch (e1 -> type) {
    case LT:
        assert_syntax (DO);
        return (node3 (FOR, e1 -> c1, e1 -> c2, rexpr (35)));
    default:
        switch (Token) {
        case IN:
            next_token ();
            e2 = rexpr (0);
            assert_syntax (DO);
            return (node3 (FORIN, e1, e2, rexpr (35)));
        case TIMES:
            next_token ();
	    assert_syntax (DO);
            return (node2 (FORTIMES, e1, rexpr (35)));
        default:
            syntax_error (-1);
            return (True);
        }
    }
}

NODEPTR rdefine()
{
    NODEPTR e;
    NODEPTR nl;
    NODEPTR id = rname ();
    if (Token == LPAREN) {
        assert_syntax (LPAREN);
        nl = rnamelist ();
        assert_syntax (RPAREN);
        assert_syntax (EQ);
        e = rexpr (35);
/*        e = rexpr (10);*/
        return (node2 (DEFINE, id, node2 (LAMBDA, nl, e)));
    }
    else {
        assert_syntax (EQ);
        e = rexpr (35);
/*        e = rexpr (10);*/
        return (node2 (DEFINE, id, e));
    }
}

NODEPTR rset()
{
    NODEPTR id = rsysvar ();
    assert_syntax (EQ);
    return (node2 (SET, id, rexpr (35)));
}

NODEPTR rlambda()
{
    NODEPTR e;
    NODEPTR nl;
    assert_syntax (LPAREN);
    nl = rnamelist ();
    assert_syntax (RPAREN);
    switch (Token) {
    case DO:
        next_token ();
        e = rexpr (35);
        return (node2 (LAMBDA, nl, e));
    default:
        syntax_error (-1);
        return (True);
    }
}

NODEPTR rlist()
{
    NODEPTR e;
    if (Token == RBRACK)
        e = NULL;
    else
        e = rexprlist ();
    assert_syntax (RBRACK);
    return (node1 (MAKELIST, e));
}

NODEPTR rsublist(list)
NODEPTR list;
{
    NODEPTR e1, e2;
    assert_syntax (LBRACK);
    e1 = rexpr (0);
    if (Token == TO) {
        next_token ();
        e2 = rexpr (0);
        assert_syntax (RBRACK);
        return (node3 (SUBLIST, list, e1, e2));
    }
    else {
        assert_syntax (RBRACK);
        return (node2 (SUB, list, e1));
    }
}

NODEPTR rlistdef(type)
unsigned int type;
{
    NODEPTR e1, e2;
    assert_syntax (LPAREN);
    e1 = rexpr (0);
    assert_syntax (COMMA);
    e2 = rexpr (0);
    assert_syntax (RPAREN);
    return (node2 (type, e1, e2));
}

NODEPTR rformat()
{
    NODEPTR e;
    assert_syntax (LPAREN);
    e = rexprlist ();
    assert_syntax (RPAREN);
    return (node1 (FORMAT, e));
}

NODEPTR rhelp()
{
    NODEPTR s;
    switch (Token) {
        case STRING:
            s = AuxVal.n;
            break;
        case NAME:
            s = string_node (AuxVal.s -> name);
            break;
        case END:
            s = NULL;
            break;
        default:
            s = string_node (strsave (Keyword[Token]));
    }
    while (Token != END)
        next_token ();
    return (node1 (HELP, s));
}

NODEPTR rcall(id)
SYMBOLPTR id;
{
    NODEPTR e;
    assert_syntax (LPAREN);
    if (Token != RPAREN)
        e = rexprlist ();
    else
        e = NULL;
    assert_syntax (RPAREN);
    return (node2 (CALL, nodep(id), e)); /* AC pointer */
}

NODEPTR rnamelist()
{
    if (Token == NAME) {
	NODEPTR id = rname ();
        if (Token != COMMA)
            return (id);
        next_token ();
	return (node2 (COMMA, id, rnamelist ()));
    }
    return (NULL);
}

NODEPTR rname()
{
    NODEPTR id = AuxVal.n;
    assert_syntax (NAME);
    return (id);
}

NODEPTR rexprlist()
{
    NODEPTR e = rexpr (0);
    if (Token != COMMA)
        return (node2 (COMMA, e, NULL));
    next_token ();
    return (node2 (COMMA, e, rexprlist ()));
}

NODEPTR rsysvar()
{
    NODEPTR id = AuxVal.n;
    switch (Token) {
	case PASSLIMIT:
	case VERBOSITY:
	case PRINTLEVEL:
	case DEBUGLEVEL:
	case STATEPRINT:
	case REDUNDANCY:
	case CALLBYNAME:
	case ASSIGNAHEAD:
	case PATH:
	case HELPDIR:
	case TIMESTEP:
        case RANDSEED:
        case RRANDSEED:
	case MAXRANDLEN:
	    next_token ();
	    return (id);
	default:
	    syntax_error (-1);
	    return (NULL);
    }
}

void push_scope(name_list)
NODEPTR name_list;
{
    unsigned int level = EnvStack[EnvTop].env -> env_level + 1;
    if (EnvTop >= MAXENVSTACKSIZE)
         error_break0("environment stack full");
    else {
        EnvTop += 1;
        EnvStack[EnvTop].env = envp (env_node(level, 0, NULL));
        EnvStack[EnvTop].env_base = NULL;
        install_name_list (name_list);
    }
}

NODEPTR pop_scope()
{
    return (nodep (EnvStack[EnvTop--].env));
}

void install_name_list(name_list)
NODEPTR name_list;
{
    SYMBOLPTR id;
    if (name_list == NULL)
        return;
    while (name_list -> type == COMMA) {
	id = symbolp (name_list -> c1);
	install_local_id (id -> name);
	name_list = name_list -> c2;
    }
    install_local_id (symbolp (name_list) -> name);
}

NAMEPTR lookup_env(s)
char *s;
{
    register NAMEPTR np;
    register int level;

    for (level = EnvTop; level >= EnvBase; level--)
        for (np = namep (EnvStack[level].env -> id_list);
	     np != NULL; np = np -> next_node)
            if (strcmp (s, namep (np) -> name) == 0)
                return (np);
    return (install_global_id (s));
}

NAMEPTR install_global_id(name)
char *name;
{
    NAMEPTR *np; /* = &(EnvStack[EnvBase].env -> id_list);*/
    register    NAMEPTR id;
    np = &(EnvStack[EnvBase].env -> id_list);  /* instead */
    id = name_node ((unsigned)isupper ((int)*name), name, 0, offset (GlobalMemUsed++));
    if (GlobalMemUsed >= GLOBALMEMSIZE)
        error_break0("global memory full");
    else
	new_loc (mem_loc (id), id -> type);
    while (*np != NULL) {
        np = &((*np) -> next_node);
    }
    return (*np = id);
}

void install_local_id(name)
char *name;
{
    NAMEPTR *np; /*= & (EnvStack[EnvTop].env -> id_list);*/
    register    NAMEPTR id;
    np = &(EnvStack[EnvTop].env -> id_list);  /* instead */
    id = name_node ((unsigned)isupper ((int)*name), name, EnvStack[EnvTop].env -> env_level,
    		    offset (EnvStack[EnvTop].env -> env_size++));
    while (*np != NULL) {
        np = &((*np) -> next_node);
    }
    *np = id;
} 

void next_token()
{
    Token = read_token ();
}

void assert_syntax(tk)
int tk;
{
    if ((int)Token != tk)
        syntax_error (tk);
    else
        next_token ();
}

void syntax_error(n)
int n;
{
    if (Recover)
        return;
    Recover = T;
    SyntaxErrorCount += 1;
    printf ("***Syntax Error");
    if (InputFile != stdin)
        printf (" on line %d of \"%s\"", InputLineCount, InputFilename);
    printf (": \n");
    fflush(stdout);
    syntax_msg (n);
    printf (".\n");
    fflush(stdout);
}

void syntax_msg(n)
int n;
{
    switch (n) {
        case N1:
            printf ("one identifier expected after quantifier");
            break;
        case MQS:
            printf ("missing quote in string");
            break;
        case STL:
            printf ("string too long");
            break;
        case IDN:
            printf ("syntax error, comment doesn't end properly");
            break;
        case END:
            printf ("\"%s\" unexpected (command must be terminated by \".\")",
                                                                   InputText);
            break;
        default:
            printf ("\"%s\" unexpected", InputText);
            if (0 <= n && n <= MAXTOKEN)
                printf (" (\"%s\" expected)", Keyword[n]);
    }
}

