/* r-assign.c:  assignment reduction routines
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: r-assign.c,v 2.10 2001/09/10 15:17:36 acau Exp $
 */


#include "system.h"



NODEPTR assign_curr (e1, e2)
NODEPTR e1, e2;
{
    NODEPTR val;
    MEMORYPTR loc;

    val = reduce_val (e2);
    if (!Ready) {
        val = reduce_val (e1);
        e1 = e2;
    }
    if (Ready) {
        loc = reduce_ref (e1);
	if (Ready) {
	    if (!AssignAheadFlag && is_next (loc))
	        return (assign (e1, val));
	    else {
		assign_val (loc, val);
	        return (True);
	    }
	}
    }
    return (True);
}


NODEPTR assign_next (e1, e2)
NODEPTR e1, e2;
{
    NODEPTR val;
    MEMORYPTR loc;
    
    val = reduce_val (e2);
    if (Ready) {
	if (!AssignAheadFlag)
	    return (assign (e1, val));
	NextFlag = T;
        loc = reduce_ref (e1);
	NextFlag = F;
	if (Ready) {
	    assign_val (loc, val);
	    return (True);
	}
    }
    return (True);
}

/************************\
*                        *
* assignment statements  *
*                        *
\************************/

/* v = e (assignment) */
void reduce_EQ_stmt(e)
NODE2PTR e;
{
    e = assign_curr (e -> c1, e -> c2);
    if (Ready)
        result (e);
}

/* e1 gets e2 */
void reduce_GETS(e)
NODE2PTR e;
{
    keep (assign_next (e -> c1, e -> c2));
}

/* STB e */
void reduce_STB(e)
NODE1PTR e;
{
    keep (assign_next (e -> c1, e -> c1));
}

/* IS e1 e2 */
void reduce_IS(e)
NODE2PTR e;
{
    always (assign_curr (e -> c1, e -> c2));
}

/* ASSN e1 e2 */
void reduce_ASSN(e)
NODE2PTR e;
{
    NODEPTR val = reduce_val (e -> c2);
    if (Ready) {
            Retry = T;
            result (node1 (FIN, assign (e -> c1, val)));
    }
}

/* UASSN e1 e2 */
void reduce_UASSN(e)
NODE2PTR e;
{
    put_empty (False);
    e = assign_next (e -> c1, e -> c2);
    if (Ready)
        result (e);
}
