/* r-expr.c:  expression reduction routines
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: r-expr.c,v 2.11 2007/05/03 09:48:45 acau Exp $
 */


#include "system.h"




/****************************************\
*                                        *
*  statements and expressions for types  *
*                                        *
\****************************************/


/* Get the type of something. */
/* TYPE e */
void reduce_TYPE(e)
NODE1PTR e;
{
    MEMORYPTR loc =reduce_ref  (e -> c1);

    if (Ready) {
       if ((Ready = mem_defined (loc)))
           result (string_node (strsave (print_name (get_mem (loc)))));
    }
}


/* Check whether e is undefined. */
/* UNDEFINED e */
void reduce_UNDEFINED(e)
NODE1PTR e;
{
     MEMORYPTR loc=reduce_ref (e -> c1);

     if ((Ready && mem_defined (loc)) || ForceOutput) {
         if (Ready)
             if (mem_defined (loc))
                 result(False);
             else
                 result(True);
     }
     
    
    Ready = (Ready &&  mem_defined (loc)) || ForceOutput;
}

/**************************************\
*                                      *
*  Simple Boolean, arithmetic, string  *
*  and list expressions.               *
*                                      *
\**************************************/

/* Equality and Non-equality. */
void reduce_EQ_val(e)
NODE2PTR e;
{
    NODEPTR e1, e2;
    e1 = reduce_val (e -> c1);
    if (Ready) {
        e2 = reduce_val (e -> c2);
        if (Ready) {
            if (e -> type == EQ)
                result (equal_val (e1, e2) ? True : False);
            else
                result (equal_val (e1, e2) ? False : True);
        }
    }
}

int power(i,j)
int i,j;
{
    int p = 1;
    while (j-- > 0)
        p = p * i;
    return (p);
}

#define int_plus(X,Y) int_node (intp (X) -> ival + intp (Y) -> ival)
#define int_minus(X,Y) int_node (intp (X) -> ival - intp (Y) -> ival)
#define int_mult(X,Y) int_node (intp (X) -> ival * intp (Y) -> ival)
#define int_div(X,Y) int_node (intp (X) -> ival / intp (Y) -> ival)
#define int_mod(X,Y) int_node (mod (intp (X) -> ival, intp (Y) -> ival))
#define int_exp(X,Y) int_node (power (intp (X) -> ival, intp (Y) -> ival))

NODEPTR calculate_val(op, e1, e2)
unsigned op;
NODEPTR e1, e2;
{
    unsigned type = e1 -> type;
        type_check (e2, type);
        switch (op) {
        case AND:
            switch (type) {
                case BOOL:
                case INT:
                    return (boolp (e1) -> bval ? e2 : e1);
                default:
                    break;
            }
            break;
        case OR:
            switch (type) {
                case BOOL:
                case INT:
                    return (boolp (e1) -> bval ? e1 : e2);
                default:
                    break;
            }
            break;
        case LT:
            return (compare_val (intp(e1), intp(e2)) < 0 ? True : False);
        case LE:
            return (compare_val (intp(e1), intp(e2)) <= 0 ? True : False);
        case GE:
            return (compare_val (intp(e1), intp(e2)) >= 0 ? True : False);
        case GT:
            return (compare_val (intp(e1), intp(e2)) > 0 ? True : False);
        case PLUS:
            switch (type) {
                case INT:
                    return (int_plus (e1, e2));
                case LIST:
                    return (nodep (append_list (listp(e1), listp(e2))));
                case STRING:
                    return (append_string (stringp(e1), stringp(e2)));
                default:
                    break;
            }
            break;
        case MINUS:
            switch (type) {
                case INT:
                    return (int_minus (e1, e2));
                default:
                    break;
            }
            break;
        case MULT:
            switch (type) {
                case INT:
                    return (int_mult (e1, e2));
                default:
                    break;
            }
            break;
        case DIV:
            switch (type) {
                case INT:
                    return (int_div (e1, e2));
                default:
                    break;
            }
            break;
        case MOD:
            switch (type) {
                case INT:
                    return (int_mod (e1, e2));
                default:
                    break;
            }
            break;
        case POWER:
            switch (type) {
                case INT:
                    return (int_exp (e1, e2));
                default:
                    break;
            }
            break;
        default:
            break;
        }
        error_break2 ("\"%s\" is not a %s operator",
                      Keyword[op], Keyword[type]);
        return (False);
}

/* Binary expressions: AND, OR, PLUS, MINUS, ... */
void reduce_bin_op(e)
NODE2PTR e;
{
    NODEPTR calculate_val ();
    NODEPTR e1, e2;
    e1 = reduce_val (e -> c1);
    if (Ready) {
        switch (e -> type) {
            case AND:
                e2 = (boolp (e1) -> bval == F) ? e1 : reduce_val (e -> c2);
                break;
            case OR:
                e2 = (boolp (e1) -> bval == T) ? e1 : reduce_val (e -> c2);
                break;
            default:
                e2 = reduce_val (e -> c2);
        }
    }
    if (Ready)
        result (calculate_val (e -> type, e1, e2));
}

#define int_uminus(X) int_node (- (intp (X) -> ival))
#define int_not(X) (intp (X) -> ival == 0 ? One : Zero)

/* unary operators: NOT, UPLUS, UMINUS, SIZE, ... */
void reduce_u_op(e)
NODE1PTR e;
{
    NODEPTR e1;

    e1 = reduce_val (e -> c1);
    if (Ready) {
        switch (e1 -> type) {
            case BOOL:
                switch (e -> type) {
                    case NOT:
                        result (boolp (e1) -> bval ? False : True);
                        break;
                    default:
                        error_break1 ("\"%s\" is not a Boolean operator",
                                print_name (e));
                }
                break;
            case INT:
                switch (e -> type) {
                    case UPLUS:
                        result (e1);
                        break;
                    case UMINUS:
                        result (int_uminus (e1));
                        break;
                    case NOT:
                        result (int_not (e1));
                        break;
                    default:
                        error_break1 ("\"%s\" is not an integer operator",
                                   print_name (e));
                }
                break;
            case STRING:
                switch (e -> type) {
                    case SIZE:
                        result (int_node ((int)e -> c2));
                        break;
                    default:
                        error_break1 ("\"%s\" is not a string operator",
                                print_name (e));
                }
                break;
            default:
                error_break2 ("\"%s\" is not a %s operator",
                        print_name (e), print_name (e1));
        }
    }
}

/* random  */
/* Reduces to random integer >=0  AC 29Apr1999 */
void reduce_RANDOM()
{
  if ( RandomValTime != Clock || RandomVal == -1 ) {
    RandomVal = rand ();
    RandomValTime = Clock;
  };
  Ready = T;
  result (int_node (RandomVal));
}

/* Random  */
/* Reduces to random integer >=0  AC 20/09/2000 */
/* This will  allow for more than 1 random number
   in the same state */
void reduce_RRANDOM()
{
  if ( RRandomValTime != Clock || RRandomVal == -1 ) {
    RRandomValTime = Clock;
  };
  Ready = T;
  result (int_node (rand()));
}


/********************\
*                    *
*  string operators  *
*                    *
\********************/

void reduce_ASCII(e)
NODE1PTR e;
{
    char   *str;
    NODEPTR e1 = reduce_val (e -> c1);
    if (Ready) {
        type_check (e1, STRING);
        assert1 (strlen (str = stringp (e1) -> str) == 1,
                "cannot get ascii code of string \"%s\" (length not 1)", str);
        result (int_node (*str));
    }
}



/**********************\
*                      *
*  pointer operations  *
*                      *
\**********************/

/* REFER e */
void reduce_REFER(e)
NODE1PTR e;
{
    MEMORYPTR loc = reduce_ref (e -> c1);
    if (Ready) {
        assert0 (loc != TempLoc, "cannot reference a value");
        result (node2 (LOC, nodep(loc), nodep(is_next(loc)))); /* AC pointer */
    }
}

/* DEREF e */
void reduce_DEREF(e)
NODE2PTR e;
{
    LOCPTR p = locp (reduce_val (e -> c1));
    if (Ready) {
        type_check (nodep(p), LOC); /* AC pointer */
        ref_result (p -> to_next ? next_loc (p -> ptr) : curr_loc (p -> ptr));
    }
}
