/* r-int.c: interval formulae reduction routines
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: r-int.c,v 2.10 2001/09/10 15:17:36 acau Exp $
 */



#include "system.h"




/****************************************\
* 					*
* interval-dependent control statements *
* 	                                *
\****************************************/


/* SEMI w1 w2 */
void reduce_SEMI(e)
NODE2PTR e;
{
    NODE2PTR e1 = e;
    BLOCKPTR empty_loc = local_mem (STATE);
    for (e = NULL; e1 -> type == SEMI; e1 = e1 -> c1)
        e = node3 (iSEMI, nodep(empty_loc), e1 -> c2, e); /* AC pointer */
    Retry = T;
    result (node3 (iSEMI, nodep(empty_loc), e1, e)); /* AC pointer */
}

/* iSEMI empty_loc w next */
void reduce_iSEMI(e)
NODE3PTR e;
{
    MEMORYPTR empty_loc;
    push_empty (empty_loc = blockp (e -> c1) -> blk_base);
    e -> c2 = reduce_pgm (e -> c2);
    pop_empty;
    if (mem_defined (empty_loc = curr_loc (empty_loc))) {
        if (get_mem (empty_loc) == False) {
            put_empty (False);
        }
        else if (Ready) {
            if ((e = e -> c3) == NULL) {
                put_empty (True);
                result (True);
            }
            else {
                Retry = T;
                if (!RedundancyFlag && e -> c3 == NULL)
                    result (e -> c2);
                else {
                    erase_mem (empty_loc);
                    result (e);
                }
            }
        }
    }
}

/* WHILE b s */
void reduce_WHILE(e)
NODE2PTR e;
{
    NODEPTR b = reduce_val (e -> c1);
    if (Ready) {
        type_check (b, BOOL);
        result (boolp (b) -> bval ? node2 (SEMI, e -> c2, e) : Empty);
        Retry = T;
    }
}

/* REPEAT s b */
void reduce_REPEAT(e)
NODE2PTR e;
{
    Retry = T;
    result (node2 (SEMI, e -> c1,
    	node2 (WHILE, node1 (NOT, e -> c2), e -> c1)));
}

/* FORTIMES e1 e2 */
void reduce_FORTIMES(e)
NODE2PTR e;
{
    NODEPTR e1 = reduce_val (e -> c1);
    if (Ready) {
        type_check (e1, INT);
        Retry = T;
        result (node2 (WHILE,
	          node1(iFORTIMES, int_node (intp (e1) -> ival)), e -> c2));
    }
}

/* iFORTIMES i e2 */
void reduce_iFORTIMES(e)
NODE1PTR e;
{
    NODEPTR e1 = e -> c1;
    if (intp (e1) -> ival == 0)
        result (False);
    else {
        intp (e1) -> ival -= 1;
        result (True);
    }
}

/* iFORTIMES i e2
reduce_iFORTIMES(e)
NODE2PTR e;
{
    if (e -> c1 -> ival == 0)
    result (Empty);
    else {
    e -> c1 -> ival -= 1;
    result (node2 (SEMI, e -> c2, e));
    }
    Retry = T;
}
*/
/* FOR id n pgm */
void reduce_FOR(e)
NODE3PTR e;
{
    NODEPTR env = e -> c1;
    BLOCKPTR blk;
    NODEPTR e2 = reduce_val (e -> c2);
    if (Ready) {
        type_check (e2, INT);
        blk = local_mem (envp (env) -> id_list -> type);
        result (node2 (iEXISTS, bnd_node (envp(env), blk), /* AC pointer */
                       node2 (WHILE,
                              node3 (iFOR, int_node (0), e2, nodep(blk)),
                              e -> c3)));                 /* AC pointer */
        Retry = T;
    }
}

/* iFOR i n loc */
void reduce_iFOR(e)
NODEPTR e;
{
    MEMORYPTR loc = blockp (e -> c3) -> blk_base;
    int       i;
    if ((i = intp (e -> c1) -> ival++) >= intp (e -> c2) -> ival)
        result (False);
    else {
        put_mem (loc, int_node (i));
        result (True);
    }
}

/* FORIN id list expn */
void reduce_FORIN(e)
NODE3PTR e;
{
    NODEPTR env = e -> c1;
    BLOCKPTR blk;
    NODEPTR list = reduce_val (e -> c2);
    if (Ready) {
        blk = local_mem (envp (env) -> id_list -> type);
        Retry = T;
        result (node2 (iEXISTS, bnd_node (envp(env), blk), /* AC pointer */
                       node2 (WHILE,
                              node3 (iFORIN, int_node (0), list, nodep(blk)),
                              e -> c3)));                 /* AC pointer */
    }
}

/* iFORIN i list base_loc */
void reduce_iFORIN(e)
NODEPTR e;
{
    MEMORYPTR loc = blockp (e -> c3) -> blk_base;
    NODEPTR list = e -> c2;
    unsigned int i;
    switch (list -> type) {
    case LIST:
        if ((i = (intp (e -> c1) -> ival)++) >= (unsigned int) get_int (list_size (list)))
            result (False);
        else {
            make_indirect_mem (loc, list_base (list) + offset (i));
            result (True);
        }
        break;
        case STRING:      /* AC pointer */
        if ((i = (intp (e -> c1) -> ival)++) >= (unsigned int) string_size (stringp(list)))
            result (False);
        else {        /* AC pointer */
            put_mem (curr_loc (loc), substring (stringp(list), (int)i, (int)i));
            result (True);
        }
        break;
    default:
        error_break1 ("cannot index expression of type: %s",
            print_name (list));
    }
}


void reduce_CHOPSTAR(NODE1PTR e)
{
  if ((Ready = empty_defined ())) {
    if (get_empty () == True)
      result (True);
    else {
      Retry = T;
      result (node2 (SEMI, node2 (AND, e -> c1, More), e));
    }
  }
}

/* KEEP s */
/* BCM 15Aug97 */
void reduce_KEEP(NODE1PTR e)
{
  if ((Ready = empty_defined ())) {
    if (get_empty () == True)
      result (True);
    else {
      Retry = T;
      result (node2 (SEMI, node2 (AND, Skip, e -> c1), e));
    }
  }
}

/************************\
*               	*
*  Projection operators	*
* 			*
\************************/

/* PROJ e1 e2 */
void reduce_PROJ(e)
NODEPTR e;
{
/*    assert0 (!AssignAheadFlag,*/
/*        "projection requires \"assign_ahead\" to be false");*/
    Retry = T;
    result (node3 (iPROJ, False, e -> c1, e -> c2));
}

/* iPROJ ready e1 e2 */
void reduce_iPROJ(e)
NODEPTR e;
{
    if (!(Ready = boolp (e -> c1) -> bval)) {
        push_assign_ahead;
        e -> c3 = reduce_stmt (e -> c3);
        e -> c1 = bool_node (Ready);
        pop_assign_ahead;
    }
    if (Ready) {
        if ((Ready = empty_defined ())) {
            if (get_empty () == True) {
                result (True);
            }
            else {
                Retry = T;
		e -> c1 = False;
                result (node2 (SEMI, e -> c2, e));
	    }
        }
    }
}


