/* r-linear.c: linear-time formulae reduction routines
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: r-linear.c,v 2.10 2001/09/10 15:17:37 acau Exp $
 */


#include "system.h"



/*****************************************\
*                                         *
* tempura linear-time control statements  *
*                                         *
\*****************************************/

/* AND e1 e2 */
void reduce_AND (e)
NODE2PTR e;
{
    NODE3PTR e1,e2;
    e1 = e2 = node3 (iAND, False, e, NULL);
    while (e1 != NULL) {
	while (e1 -> c2 -> type == AND) {
	    e1 -> c3 = node3 (iAND, False, (e1 -> c2) -> c2, e1 -> c3);
	    e1 -> c2 = (e1 -> c2) -> c1;
	}
	e1 = e1 -> c3;
    }
    Retry = T;
    result (e2);
}

/* iAND ready e next */
void reduce_iAND (e)
NODE3PTR e;
{
    NODE3PTR * cnode_ptr = &e,
        cnode,
        temp;
    unsigned allready = T;

    while ((cnode = *cnode_ptr) != NULL) {
        if (cnode -> c1 == True) 
	    cnode_ptr = &(cnode -> c3);
	else if ((temp = cnode -> c2) -> type == iAND) {
	    NODE3PTR * temp_ptr = &(cnode -> c2);
	    while (*temp_ptr != NULL)
	        temp_ptr = &((*temp_ptr) -> c3);
            *temp_ptr = cnode -> c3;
	    cnode -> c1 = temp -> c1;
	    cnode -> c2 = temp -> c2;
	    cnode -> c3 = temp -> c3;
	}
	else {
            if ((cnode -> c2 = reduce_stmt (temp)) -> type == BOOL) {
		if (cnode -> c2 == True) 
		    *cnode_ptr = cnode -> c3;
		else {
		    Ready = T;
		    result (False);
		    return;
		}
	    }
            else {
		if (Ready)
                    cnode -> c1 = True;
	        else
	            allready = F;
                cnode_ptr = &(cnode -> c3);
	    }
        }
    }
    if ((Ready = allready)) {
        for (cnode = e; cnode != NULL; cnode = cnode -> c3)
            cnode -> c1 = False;
    }
    else {
	cnode_ptr = &e;
	temp = NULL;
        while ((cnode = *cnode_ptr) != NULL) {
            if (cnode -> c1 == True) {
                cnode_ptr = &(cnode -> c3);
            }
            else {
                *cnode_ptr = cnode -> c3;
                cnode -> c3 = temp;
                temp = cnode;
	    }
	}
        *cnode_ptr = temp;
    }
    result (e == NULL ? True : e -> c3 == NULL ? e -> c2 : e);
}


/* IMPLIES e1 e2   */
/* added 04/09/2001 */
void reduce_IMPLIES (e)
NODE2PTR e;
{
  Retry = T;
  result (node3 (IF, e -> c1, e -> c2, True));
}

void reduce_WHERE(e)
NODEPTR e;
{
    Retry = T;
    result (conj (e -> c2, e -> c1));
}

/* IF b e1 e2  */
/* added 06/09/2001 */
void reduce_IF_stmt(e)
NODE3PTR e;
{
    NODEPTR b = reduce_val (e -> c1);
    if (Ready) {
        type_check (b, BOOL);
        result (boolp (b) -> bval ? e -> c2 : e -> c3);
        Retry = T;
    }
}

/* Added 06/09/2001 */
void reduce_IF_val(e)
NODE3PTR e;
{
    NODEPTR b = reduce_val (e -> c1);
    if (Ready) {
        type_check (b, BOOL);
        result (boolp (b) -> bval ? e -> c2 : e -> c3);
        Retry = T;
    }
}

/* iIF b e1 e2 */
/* Used by reduce_IF_val when test is a NEXT term. */
void reduce_iIF_val(e)
NODE3PTR e;
{
  NODEPTR e1,e2;
  Ready = T;
  e1 = reduce_val(e -> c2);
  if ( Ready ) {
      e1 = copy_val (e1);
  };
  if ( Ready ) {
    e2 = reduce_val(e -> c3);
    if ( Ready ) {
	e2 = copy_val (e2);
    };
  };
  if ( Ready )
    result (node1 (NEXT, node3 (IF, (e -> c1) -> c1 ,e1,e2)));
}


/* CALL f elist */
void reduce_CALL(e)
NODE2PTR e;
{
    NODEPTR f = e -> c1;
    NODE2PTR lambda;
    NODEPTR env;
    NAMEPTR parm;
    NODE2PTR elist = e -> c2;
    BLOCKPTR blk;
    MEMORYPTR base_loc;
    MEMORYPTR arg_loc;

    lambda = reduce_val (f);
    if (Ready) {
        type_check (lambda, LAMBDA);
        blk = mem_block (STATIC, envp (env = lambda -> c1) -> env_size);
        base_loc = blk -> blk_base;
        for (parm = envp (env) -> id_list; (parm != NULL) && (elist != NULL);
                parm = parm -> next_node, elist = elist -> c2) {
            arg_loc = reduce_ref (elist -> c1);
            if (Ready) {
                if (arg_loc == TempLoc)
                    put_mem (new_loc (base_loc + parm -> off, STATIC),
                            get_mem (arg_loc));
                else if (parm -> type == STATIC && !is_static_mem (arg_loc)) {
		    NODEPTR val = reduce_val (elist -> c1);
		    if (Ready)
                        put_mem (new_loc (base_loc + parm -> off,STATIC),val);
		    else
		        return;
		}
                else
                    make_indirect_mem (
                            new_loc (base_loc + parm -> off, parm -> type),
                            arg_loc);
            }
            else
                return;
        }
        assert0 ((elist == NULL) && (parm == NULL),
                "wrong number of parameters in function call");
        Retry = T;
        result (node3 (iCALL, f, bnd_node (envp(lambda -> c1), blk), lambda -> c2));
    }
}

/* iCALL env body baseloc */
void reduce_iCALL_stmt(e)
NODEPTR e;
{
    push_display (bndp(e -> c2));
    e -> c3 = reduce_stmt (e -> c3);
    pop_display ();
    if (e -> c3 -> type == BOOL)
        result (e -> c3);
/*    else if (e -> c3 -> type == iCALL)*/
/*        result (e -> c3);             */  /* Tail recursion */
}

/* iCALL env body baseloc */
void reduce_iCALL_val(e)
NODEPTR e;
{
    NODEPTR val;
    push_display (bndp(e -> c2));
    val = reduce_val (e -> c3);
    pop_display ();
    if (Ready)
        result (val);
}

/* iCALL env body baseloc */
/* added memoryp 06/09/2001 */
void reduce_iCALL_ref(e)
NODEPTR e;
{
    MEMORYPTR loc;
    push_display (bndp(e -> c2));
    loc = reduce_ref (e -> c3);
    pop_display ();
    if (Ready)
        ref_result (loc);
}

/* EXISTS env e w  (this is for "exists v<e:w") */
/* EXISTS env w */
void reduce_EXISTS(e)
NODE3PTR e;
{
    NODEPTR env = e -> c1;
    NAMEPTR id;
    BLOCKPTR env_blk;
    MEMORYPTR env_base;

    if (e -> c2 != NULL) {
        Retry = T;
        result (node1 (NOT,
                    node3 (FORALL, env, e -> c2, node1 (NOT, e -> c3))));
    }
    else {
        env_base = (env_blk = allocate_mem (envp (env) -> env_size)) -> blk_base;
        for (id = envp (env) -> id_list; id != NULL; id = id -> next_node)
            new_loc (env_base + id -> off, id -> type);
        Retry = T;
        result (node2 (iEXISTS, bnd_node (envp(env), env_blk), e -> c3));
    }
}

/* iEXISTS idlist body baseloc */
void reduce_iEXISTS_stmt(e)
NODE2PTR e;
{
    push_env (bndp(e -> c1));
    e -> c2 = reduce_stmt (e -> c2);
    pop_env ();
    if (e -> c2 -> type == BOOL) 
        result (e -> c2);
}

/* iEXISTS idlist body baseloc */
void reduce_iEXISTS_val(e)
NODE2PTR e;
{
    NODEPTR val;
    push_env (bndp(e -> c1));
    val = reduce_val (e -> c2);
    pop_env ();
    if (Ready) {
	type_check (val, BOOL);
        result (val);
    }
}

/* FORALL id n expn */
void reduce_FORALL(e)
NODE3PTR e;
{
    BLOCKPTR blk;
    int nval;
    NODEPTR env = e -> c1;
    NODEPTR expn = e -> c3;
    NODEPTR n = reduce_val (e -> c2);
    if (Ready) {
        type_check (n, INT);
        nval = intp (n) -> ival;
        e = True;
        for (nval -= 1; nval >= 0; nval -= 1) {
            blk = local_mem (envp (env) -> id_list -> type);
            put_mem (adj_loc (blk -> blk_base), int_node (nval));
            e = node2 (AND, node2 (iEXISTS, bnd_node (envp(env), blk), expn), e);
        }
        Retry = T;
        result (e);
    }
}

/* THE x : expn */
void reduce_THE(e)
NODE3PTR e;
{
    noop (e);
}

/* ALWAYS w */
void reduce_ALWAYS(e)
NODE1PTR e;
{
    always (reduce_stmt (e -> c1));
}

/* SOMETIME w */
void reduce_SOMETIME(e)
NODEPTR e;
{
    Retry = T;
    result (node3 (IF, e -> c1, True, node1 (NEXT, e)));
}

/* HALT w */
void reduce_HALT(e)
NODE1PTR e;
{
    NODEPTR e1;

    e1 = reduce_val (e -> c1);
    if (Ready) {
        type_check (e1, BOOL);
        if (e1 == True) {
            put_empty (True);
            result (True);
        }
        else
            put_empty (False);
    }
    else {
	if (empty_defined ()) {
	    Retry = T;
	    result (get_empty () == True ?
	            e -> c1 :
		    conjtf (e, node1 (NOT, e -> c1)));
	}
    }
}

/* INIT w */
void reduce_INIT(e)
NODE1PTR e;
{
    Retry = T;
    result (node2 (SEMI, conj (Empty, e -> c1), True));
}

/* FIN w */
void reduce_FIN(e)
NODE1PTR e;
{
    if ((Ready = empty_defined ())) {
        if (get_empty () == True) {
            Retry = T;
            result (e -> c1);
        }
    }
}

/* KEEP w */
/* void reduce_KEEP(e)
NODE1PTR e;
{
    keep (reduce_stmt (e -> c1));
}
*/


/* WNEXT w */
void reduce_WNEXT(e)
NODE1PTR e;
{
    Ready = T;
    result (e -> c1);
}

/* NEXT w */
void reduce_NEXT_stmt(e)
NODE1PTR e;
{
    put_empty (False);
    Ready = T;
    result (e -> c1);
}

/* NEXT e */
void reduce_NEXT_val(e)
NODE1PTR e;
{
    NODEPTR val;
    put_empty (False);
    assert0 (!NextFlag, "two levels of next in expression");
    NextFlag = T;
    val = reduce_val (e -> c1);
    NextFlag = F;
    if (Ready)
        result (copy_val (val));
    
}

/* NEXT e */
void reduce_NEXT_ref(e)
NODE1PTR e;
{
    MEMORYPTR loc;
    put_empty (False);
    assert0 (!NextFlag, "two levels of next in expression");
    NextFlag = T;
    loc = reduce_ref (e -> c1);
    NextFlag = F;
    if (Ready)
        ref_result (loc);
}


/* LEN e */
void reduce_LEN(e)
NODE1PTR e;
{
    NODEPTR e1 = reduce_val (e -> c1);
    if (Ready) {
        type_check (e1, INT);
        Retry = T;
        result (node1 (iLEN, int_node (intp (e1) -> ival)));
    }
}

/* iLEN c */
void reduce_iLEN(e)
NODE1PTR e;
{
    INTPTR c = intp (e -> c1);
    if (c -> ival == 0) {
        put_empty (True);
        result (True);
    }
    else {
        c -> ival -= 1;
        put_empty (False);
    }
    Ready = T;
}

/* PROCESS e */
void reduce_PROCESS(e)
NODE1PTR e;
{
    Retry = T;
    result (RedundancyFlag ?
            node3 (iPROCESS, e -> c1, False, nodep(local_mem (STATE))) :
            e -> c1);
}

/* iPROCESS e ready empty_loc */
void reduce_iPROCESS(e)
NODE3PTR e;
{
    NODE1PTR local_empty, global_empty;
    MEMORYPTR empty_loc = blockp (e -> c3) -> blk_base;

    if (!(Ready = boolp (e -> c2) -> bval)) {
        push_empty (empty_loc);
        e -> c1 = reduce_pgm (e -> c1);
        e -> c2 = Ready ? True : False;
        pop_empty;
    }
    if (Ready) {
        if ((Ready = empty_defined ())) {
            global_empty = get_empty ();
            local_empty = get_mem (curr_loc (empty_loc));
            assert0 (!(local_empty == False && global_empty == True),
                    "calling process terminated before subprocess");
            if (local_empty == True || e -> c1 -> type == BOOL)
                result (e -> c1);
            else 
                e -> c2 = False;
        }
    }
}

/* PREFIX e */
void reduce_PREFIX(e)
NODE1PTR e;
{
/*    assert0 (!AssignAheadFlag,*/
/*            "prefix operator requires \"assign_ahead\" to be false");*/
    Retry = T;
    result (node3 (iPREFIX, e -> c1, False, nodep(local_mem (STATE))));
}

/* iPREFIX e ready empty_loc */
void reduce_iPREFIX(e)
NODE3PTR e;
{
    NODE1PTR local_empty, global_empty;
    MEMORYPTR empty_loc = blockp (e -> c3) -> blk_base;
    push_assign_ahead;

    if (!(Ready = boolp (e -> c2) -> bval)) {
        push_empty (empty_loc);
        e -> c1 = reduce_stmt (e -> c1);
        e -> c2 = Ready ? True : False;
        pop_empty;
    }
    if (Ready) {
        if ((Ready = empty_defined ())) {
            global_empty = get_empty ();
            local_empty = mem_defined (empty_loc = curr_loc (empty_loc)) ?
                         get_mem(empty_loc) :
                         global_empty;
            assert0 (local_empty == False || global_empty == True,
                    "prefix terminated before enclosing interval");
            if (global_empty == True)
                result (True);
            else if (e -> c1 -> type == BOOL)
	        result (e -> c1);
	    else
                e -> c2 = False;
        }
    }
    pop_assign_ahead;
}

