/* r-lists.c: list expression reduction routines
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: r-lists.c,v 2.10 2001/09/10 15:17:37 acau Exp $
 */


#include "system.h"




/* FIXLIST e n */
/* VARLIST e n */
void reduce_list(e)
NODE2PTR e;
{
    MEMORYPTR loc;
    int     cell_type,
            size_type;
    NODEPTR n = reduce_val (e -> c2);
    if (Ready) {
        type_check (n, INT);
        loc = reduce_ref (e -> c1);
        if (Ready) {
            cell_type = is_static_mem (loc) ? STATIC : STATE;
            size_type = (e -> type == VARLIST) ? STATE : STATIC;
            assign_struct (loc, new_list (cell_type, size_type, intp(n)));
            result (True);
        }
    }
}


/******************\
*                  *
*  list operators  *
*                  *
\******************/

void reduce_STRUCT_val(e)
NODEPTR e;
{
    NODEPTR e1;
    if ((e1 = e -> c1) -> type == LIST)
        Ready = T;
    else {
        e1 = reduce_ref_and_fetch (e1);
        if (Ready)
            result (node1 (STRUCT, e1));
    }
}

void reduce_STRUCT_ref(e)
NODEPTR e;
{
    ref_result (reduce_ref (e -> c1));
}

/* produce a static list */
/* MAKELIST elist */
void reduce_MAKELIST(e)
NODE1PTR e;
{
    NODE2PTR elist = e -> c1;
    NODEPTR item;
    unsigned int     size = elist_size (elist);
    BLOCKPTR blk = mem_block (STATIC, size + 1);
    MEMORYPTR size_loc = blk -> blk_base,
            base_loc = size_loc + offset(1),
            loc = curr_loc (base_loc);

    Ready = T;
    put_mem (size_loc, int_node ((int)size));

    for (elist = e -> c1; elist != NULL; elist = elist -> c2) {
        item = reduce_val (elist -> c1);
        if (Ready) {
            put_mem (loc, copy_val (item));
            loc += offset (1);
        }
        else
            return;
    }
    if (Ready)
        result (list_node (blk, size_loc, base_loc));
}

int reduce_index(e)
NODEPTR e;
{
    e = reduce_val (e);
    if (Ready) {
        type_check (e, INT);
        return (intp (e) -> ival);
    }
    return (0);
}

/* SUB e1 e2 */
void reduce_SUB(e)
NODE2PTR e;
{
    NODEPTR list;
    int     n = reduce_index (e -> c2);
    if (Ready) {
        list = reduce_ref_and_fetch (e -> c1);
        if (Ready) {
            switch (list -> type) {
                case LIST:
                    assert2d ((0 <= n) && (n < max_size (listp(list))),
                             "subscript should be less than %d: %d",
			     max_size (listp(list)), n);
                    ref_result (list_base (list) + offset (n));
                    break;
                case STRING:
                    ref_result (make_ref (substring (stringp(list), n, n)));
                    break;
                default:
                    error_break1 ("cannot subscript expression of type %s",
                            print_name (list));
	    }
	}
    }
}

/* SIZE e, |e| */
void reduce_SIZE(e)
NODE1PTR e;
{
    NODEPTR list = reduce_ref_and_fetch (e -> c1);
    if (Ready) {
        switch (list -> type) {
            case LIST:
                ref_result (list_size (list));
                break;
            case STRING:
                ref_result (make_ref (int_node (string_size (stringp(list)))));
                break;
            default:
                error_break1 ("cannot get size of %s", print_name (list));
	}
    }
}


NODEPTR sublist(list,m,n)
LISTPTR list;
int m,n;
{
    int     sub_len, current_size;
    BLOCKPTR blk;
    MEMORYPTR size_loc, sub_base_loc;

    size_loc = list_size (list);
    current_size = get_int (size_loc);
    assert2d ((m >= 0) && (n <= current_size),
            "sublist [%d..%d] not within list", m, n);
    sub_len = (n > m) ? n - m : 0;
    sub_base_loc = list_base (list) + offset (m);
    blk = local_mem (STATIC);
    size_loc = is_next (size_loc) ?
        next_loc (blk -> blk_base) :
        curr_loc (blk -> blk_base);
    put_mem (size_loc, int_node (sub_len));
    return (list_node (blk, size_loc, sub_base_loc));
}

/* sublist of e from e1 to e2-1 */
/* SUBLIST e e1 e2 */
void reduce_SUBLIST(e)
NODE3PTR e;
{
    NODEPTR list;
    int     m,
            n;
    m = reduce_index (e -> c2);
    if (Ready) {
        n = reduce_index (e -> c3);
        if (Ready) {
            list = reduce_ref_and_fetch (e -> c1);
            if (Ready) {
                switch (list -> type) {
                    case LIST:
                        if ((Ready = list_size_defined (listp(list))))
                            ref_result (make_ref (sublist (listp(list), m, n)));
                        break;
                    case STRING:
                        ref_result (make_ref (substring (stringp(list), m, n - 1)));
                        break;
                    default:
                        error_break1 ("cannot subscript expression: %s",
                                      print_name (list));
		}
	    }
	}
    }
}
