/* r-system.c: system reduction routines
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: r-system.c,v 2.12 2007/05/03 09:48:45 acau Exp $
 */


#include "system.h"





/************************\
*                        *
* top-level run command  *
*                        *
\************************/

/* run a tempura program. */
void reduce_RUN (e)
NODE1PTR e;
{
    NODEPTR emptyflag;
    unsigned old_run_flag = RunFlag;
    MEMORY local_empty[MEMCELLSIZE];

    push_empty (new_loc (local_empty, STATE));
    RunFlag = T;
    e -> c2 = e -> c1;
    
    init_exec_stat ();
    do {
        if (TimeStep > 0)
            start_time ();
        init_state_stat ();
        do {
            init_pass_stat ();
            e -> c2 = reduce_pgm (e -> c2);
            output_pass_stat ();
            update_state_stat ();
        } while (!Ready);

        output_state_stat ();
        update_exec_stat ();
        emptyflag = get_empty ();
        if (emptyflag == False) {
            StateNumber += 1;
            Clock += 1;
            if (!OneLevelMemFlag) {
                switch_memory;
    	        EmptyLoc = curr_loc (EmptyLoc);
            }
	    if (TimeStep > 0)
	        check_time ();
        }
    } while (emptyflag != True);

    output_exec_stat ();
    pop_empty;
    RunFlag = old_run_flag;
    result (e -> c2);
}

/* initialize statistics, etc. for execution */
void init_exec_stat ()
{
    TotalReduceCount = 0;
    MaxReduceDepth = 0;
    StateNumber = 0;
    TotalSuccessfulReduceCount = 0;
    TotalPassCount = 0;
}

/* initialize statistics, etc. for state */
void init_state_stat ()
{
    StateReduceCount = 0;
    StateSuccessfulReduceCount = 0;
    PassNumber = 0;
    PassLimitCounter = 0;
    StateProgressFlag = T;
    ForceOutput = F;
}

/* initialize statistics, etc. for state */
void init_pass_stat ()
{
    PassReduceCount = 0;
    PassSuccessfulReduceCount = 0;
    PassProgressFlag = F;
}

/* update execution statistics */
void update_exec_stat ()
{
   TotalReduceCount += StateReduceCount;
    TotalSuccessfulReduceCount += StateSuccessfulReduceCount;
    TotalPassCount += PassNumber;
}

/* update statistics, etc. for state */
void update_state_stat ()
{
    StateReduceCount += PassReduceCount;
    StateSuccessfulReduceCount += PassSuccessfulReduceCount;
    PassNumber += 1;
    PassLimitCounter += 1;
    StateProgressFlag = !ForceOutput;
    ForceOutput = !PassProgressFlag;
}

/* output statistics after each pass. */
void output_pass_stat ()
{
    if (state_debug) {
        output_state_number;
        output_str ("Program: ");
        output_expn (ReduceStackTop -> n -> c2, PrintLevel);
        newline;
    }
    if (pass_verbosity) {
        printf ("Pass #%d complete.  Reductions:  %d  (%d successful).\n",
                PassNumber, PassReduceCount, PassSuccessfulReduceCount);
        fflush(stdout);
    }
}

/* output out some statistics at the end of each state */
void output_state_stat ()
{
    if (state_verbosity) {
        printf ("State #%d ready.  Passes:  %d.  ", StateNumber, PassNumber);
        printf ("Reductions:  %d  (%d successful).\n",
                StateReduceCount, StateSuccessfulReduceCount);
        fflush(stdout);
        if (store_verbosity) {
            printf ("Memory in use:  ");
            printf ("Global:  %d ", GlobalMemUsed);
            printf ("(%d%%).  ",(GlobalMemUsed*100)/GLOBALMEMSIZE);
            printf ("Local:  %d.  ", LocalMemUsed);
            printf ("Heap:  %d ", HeapSize-FreeNodes);
            printf ("(%d%%).\n",((HeapSize-FreeNodes)*100)/HeapSize);
            fflush(stdout);
        }
        newline;
        fflush(stdout);
    }
    else 
	fputs ("", stdout);
}

void output_exec_stat ()
{
    if (exec_verbosity) {
        printf ("\nDone!  Computation length:  %d.  Total Passes:  %d.\n",
                StateNumber, TotalPassCount);
        fflush(stdout);
        printf ("Total reductions:  %d  (%d successful).  ",
                TotalReduceCount, TotalSuccessfulReduceCount);
        printf ("Maximum reduction depth:  %d.\n", MaxReduceDepth);
        fflush(stdout);
        if (store_verbosity) {
            printf ("Memory in use:  ");
            printf ("Global:  %d ", GlobalMemUsed);
            printf ("(%d%%).  ",(GlobalMemUsed*100)/GLOBALMEMSIZE);
            printf ("Local:  %d.  ", LocalMemUsed);
            printf ("Heap:  %d ", HeapSize-FreeNodes);
            printf ("(%d%%).\n",((HeapSize-FreeNodes)*100)/HeapSize);
            fflush(stdout);
        }
    }
}



/*************************\
*                         *
*  other system commands  *
*                         *
\*************************/

/* read in a file */
/* LOAD e */
void reduce_LOAD (e)
NODE1PTR e;
{
    int     old_input_line_count = InputLineCount;
    char   *old_filename = InputFilename,
            name[MAXFILENAMESIZE];
    FILE * fp = NULL,
        *openfile ();
    NODE2PTR e1 = reduce_val (e -> c1);

    if (Ready) {
        type_check (e1, STRING);
        /* Next line changed to avoid problems with garbage collector.
	   BCM  14Dec96 */
        if ((fp = openfile (name, Path->str, stringp(e1) -> str, FX, "r")) == NULL) {
        /* if ((fp = openfile (name, Path, e1 -> c1, FX, "r")) == NULL) {*/
            warn1 ("cannot open file \"%s\"",  stringp(e1) -> str);
        }
        else {
            InputFilename = name;
            InputLineCount = 1;
            printf ("[Reading file %s]\n", InputFilename);
            fflush(stdout);
            while (!feof (fp))
                exec_command (parse (fp));
            fclose (fp);
        }
        result (True);
    }
    InputFilename = old_filename;
    InputLineCount = old_input_line_count;
}

/* define a value or function */
/* {define e1 e2} */
void reduce_DEFINE(e)
NODE2PTR e;
{
    NAMEPTR id = namep (e -> c1);
    NODEPTR e2;
    assert1 (is_identifier (id), "cannot redefine operator: \"%s\"",
            print_name (nodep(id)));
    e2 = reduce_val (e -> c2);
    if (Ready) {
        make_static_mem (mem_loc (id));
        put_mem (curr_loc (mem_loc (id)), e2);
        Ready = T;
        result (True);
    }
}

/* EXIT */
/* exit tempura to system. */
void reduce_EXIT(e)
SYMBOLPTR e;
{
/*    intrrest (CNTRL_C);*/
    exit(0);
}

/* SHOWSTATE */
/* show the value (both defined and undefined) of the variables
   in the current state. */
void reduce_SHOWSTATE(e)
SYMBOLPTR e;
{
    show_state();
    Ready = T;
    result (True);
    
}

/* SET sysid e2 */
void reduce_SET(e)
NODE2PTR e;
{
    NODEPTR old_val, new_val;
    old_val = reduce_val (e -> c1);
    if (Ready) {
	new_val = reduce_val (e -> c2);
	if (Ready) {
	    type_check (new_val, old_val -> type);
            printf ("%s was previously set to ", print_name (e -> c1));
            output_val (old_val);
            newline;
            fflush(stdout);
	    assign_sysvar (symbolp(e -> c1), new_val);
	    result (True);
	}
    }
}

void reduce_SYSVAR(e)
NODEPTR e;
{
    switch (e -> type) {
	case PASSLIMIT:
	    e = int_node (PassLimit);
	    break;
	case VERBOSITY:
	    e = int_node (Verbosity);
	    break;
	case PRINTLEVEL:
	    e = int_node (PrintLevel);
	    break;
	case DEBUGLEVEL:
	    e = int_node (DebugLevel);
	    break;
	case STATEPRINT:
	    e = bool_node (StatePrintFlag);
	    break;
	case REDUNDANCY:
	    e = bool_node (RedundancyFlag);
	    break;
	case CALLBYNAME:
	    e = bool_node (CallByNameFlag);
	    break;
	case ASSIGNAHEAD:
	    e = bool_node (AssignAheadFlag);
	    break;
	case PATH:
	    e = nodep(Path);/*  Avoid problems with garbage collector.
			    BCM  14Dec96 */
                /* e = string_node (Path);*/
	    break;
	case HELPDIR:
	    e = nodep(HelpDir);/*  Avoid problems with garbage collector.
			      BCM  14Dec96 */
                /* e = string_node (HelpDir);*/
	    break;
	case TIMESTEP:
	    e = int_node (TimeStep);
	    break;
        case RANDSEED:
            e = int_node (RandSeed); /* AC 29Apr1999 */
            break;
        case RRANDSEED:
            e = int_node (RRandSeed); /* AC 20/09/2000 */
            break;
        case MAXRANDLEN:
            e = int_node (MaxRandLen); /* AC 29Apr1999 */
            break;               
	default:
	    error_break1 ("not a system variable: %s", print_name (e));
    }
    Ready = T;
    result (e);
}


/******************\
*                  *
* system commands  *
*                  *
\******************/


/* SHOWTRACE */
void reduce_SHOWTRACE(e)
SYMBOLPTR e;
{
    show_scope();
    show_trace();
    Ready = T;
    result (True);
}

void show_mem(loc)
MEMORYPTR loc;
{
    printf (is_static_mem (loc) ? "*  " : "   ");
    printf ("%4d  ", loc->time - Clock);
    while (is_indirect_mem (loc)) {
        loc = loc -> val.m;
        printf ("%p%s -> ", cell_no (loc), is_next (loc) ? "n" : "c");
    }
    if (mem_defined (loc))
        output_expn (loc -> val.n, -1);
    else
        printf ("undefined");
}

/* SHOWMEM e1 e2 */
void reduce_SHOWMEM(e)
NODE2PTR e;
{
    NODE1PTR e1, e2;
    MEMORYPTR loc;
    e1 = reduce_val (e -> c1);
    if (Ready) {
        type_check (e1, INT);
        e2 = reduce_val (e -> c2);
        if (Ready) {
            type_check (e2, INT);
            printf ("\nLocation  S  Time  Value\n");
            fflush(stdout);
            for (loc = memoryp (e1 -> c1); loc <= memoryp (e2 -> c1); loc++) {
                printf ("%7pc  ", loc);
                show_mem (curr_loc (loc));
                printf ("\n%8s  ", "n");
                show_mem (next_loc (loc));
                printf ("\n");
                fflush(stdout);
            }
            printf("\n");
            fflush(stdout);
            result (True);
        }
    }
}

/* EOF (NULL) */
void reduce_EOF(e)
NODE0PTR e;
{
    if (InputFile == stdin)
        pop_tempura_level (T);
    Ready = T;
    result (True);
}

/* Pop up to next higher level. */
/* RETURN */
void reduce_RETURN(e)
SYMBOLPTR e;
{
    pop_tempura_level (T);
    Ready = T;
    result (True);
}

/* Pop up to next higher level. */
/* CONTINUE */
void reduce_CONTINUE(e)
SYMBOLPTR e;
{
    pop_tempura_level (F);
    Ready = T;
    result (True);
}

/* LEAP */
void reduce_LEAP(e)
SYMBOLPTR e;
{
    BreakDepth = ReduceDepth - 1;/* Already reducing LEAP. */
    pop_tempura_level (F);
    Ready = T;
    result (True);
}

/* STEP */
void reduce_STEP(e)
SYMBOLPTR e;
{
    BreakDepth = ReduceDepth;   /* Already reducing STEP. */
    pop_tempura_level (F);
    Ready = T;
    result (True);
}

/* Call top-level recursively. */
/* BREAK */
void reduce_BREAK(e)
SYMBOLPTR e;
{
    user_break ("program break");
    Ready = T;
    result (True);
}

/* ABORT */
void reduce_ABORT(e)
SYMBOLPTR e;
{
    TempuraLevel = 0;
    Break = T;
    Ready = T;
    result (True);
}

/* print a help message. */
void reduce_HELP(e)
NODEPTR e;
{
    char    ch,
           *subject = "summary",
            name[MAXFILENAMESIZE];
    FILE * fp, *openfile ();
    STRINGPTR sp;

    Ready = T;
    if (e -> c1 != NULL) {
        sp = stringp (reduce_val (e -> c1));
        if (Ready && sp -> type == STRING)
            subject = sp -> str;
    }
    /* Next line changed to avoid problems with garbage collector.
      BCM  14Dec96 */
    if ((fp = openfile (name, HelpDir->str, subject, "", "r")) == NULL) {
/*    if ((fp = openfile (name, HelpDir, subject, "", "r")) == NULL) {*/
        printf ("***No help available");
        if (e -> c1 == NULL)
            printf ("\n");
        else
            printf (" on \"%s\"\n", subject);
        fflush(stdout);
    }
    else {
        while ((ch = getc (fp)) != EOF)
            putchar (ch);
        fclose (fp);
    }
    result (True);
}

void reduce_SYSTEM(e)
NODEPTR e;
{
    NODEPTR e1 = reduce_val (e -> c1);
#ifdef HAVE_WORKING_FORK
    if (Ready) {
        int pid;
	type_check (e1, STRING);
	pid = fork ();
        if (pid == 0)
            exit (system (stringp (e1) -> str));
	put_empty (False);
	result (node1 (iSYSTEM, int_node (pid)));
    }
#endif
}

void reduce_iSYSTEM(e)
NODEPTR e;
{
    int pid = intp (e -> c1) -> ival;
    NODEPTR *wl = &SysWaitList;
    check_sys_waits ();
    Ready = T;
    while (*wl != NULL) {
	if (intp ((*wl) -> c1) -> ival == pid) {
	    *wl = (*wl) -> c2;
	    put_empty (True);
	    result (True);
	    return;
	}
	wl = &((*wl) -> c2);
    }
    put_empty (False);
}

	    
