/* store.c: Storage management
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: store.c,v 2.11 2007/05/03 09:48:45 acau Exp $
 */


#include "system.h"



/***********************\
*                       *
*  Storage management   *
*                       *
\***********************/

char *allocate(size,elt_size)
unsigned int size,elt_size;
{
    char *p;
    if ((p = calloc (size, elt_size)) != NULL)
        return (p);
        collect_garbage (); 
    while ((p = calloc (size, elt_size)) == NULL)
        error_break0 ("insufficient memory!");
    return (p);
}


char *strsave(s)                /* save string s somewhere */
char *s;
{
    char   *p;

    p = allocate (strlen (s) + 1, sizeof (char));
    strcpy (p, s);
    return (p);
}

char *strapnd(s1,s2)            /* save append of s1 and s2 */
char *s1,*s2;
{
    char   *p;

    p = allocate (strlen (s1) + strlen (s2) + 1, sizeof (char));
    strcpy (p, s1);
    strcat (p, s2);
    return (p);
}

char *substr(s,i,j)             /* sub-string of s from i to j */
char *s;
int i,j;
{
    int len = strlen(s);
    char   *p,
           *q;

    j = max (i - 1, min (j, len - 1));
    p = q = allocate ((unsigned)j - i + 2, sizeof (char));
    while (i <= j) {
        *p++ = *(s + i++);
    }
    *p = '\0';
    return (q);
}

void more_symbol_table(size)
unsigned size;
{
    SYMBOLPTR sp;
    if (store_verbosity)
        message ("extending symbol table");
    sp = symbolp (allocate (size, sizeof (SYMBOL)));
    SymbolTableTop = sp;
    SymbolTableMax = SymbolTableTop + size;
}

void more_heap(size)
register unsigned size;
{
    HEAP h;
    register NODEPTR new_block;
    register unsigned i;
    if (store_verbosity)
        message("extending heap");
    h = heapp (allocate (1, sizeof (HEAPNODE)));
    h -> base = new_block = nodep (allocate (size, sizeof (NODE)));
    h -> top = h -> base + size;
    h -> next_block = Heap;
    for (i = 0; i < size - 1; i += 1)
        new_block[i].next_node = &(new_block[i+1]);
    new_block[i].next_node = NULL;
    FreeNodeList = new_block;
    Heap = h;
    HeapSize += size;
    FreeNodes += size;
}

NODEPTR claim_node()
{
    register NODEPTR np;

    if (FreeNodeList == NULL) {
            collect_garbage (); 
        if (FreeNodeList == NULL) 
            more_heap (MAXHEAPSIZE);
    }
    (np = FreeNodeList) -> mark = F;
    FreeNodeList = np -> next_node;
    np -> next_node = TempNodeList;
    TempNodeList = np;
    FreeNodes -= 1;
    return (np);
}

int free_heap_size()
{
    int     size = 0;
    NODEPTR np = FreeNodeList;
    while (np != NULL) {
        size += 1;
        np = np -> next_node;
    }
    return (size);
}

NODEPTR node(type,v1,v2,v3)          /* Make a node with four children. */
unsigned type;
NODEPTR v1,v2,v3;
{
    register NODEPTR np;
    np = claim_node();
    np->type = type;
    np->c1 = v1;
    np->c2 = v2;
    np->c3 = v3;
    return (np);
}

NODEPTR bool_node(val)
unsigned int val;
{
    return (val ? True : False);
}

NAMEPTR name_node (signal,name,level,off)
unsigned signal;
char *name;
unsigned int level;
unsigned int off;
{
    register NODEPTR np;

    np = claim_node();
    namep (np) -> type = signal ? STATE : STATIC;
    namep (np) -> name = name;
    namep (np) -> level = level;
    namep (np) -> off = off;
    TempNodeList = nodep (np -> next_node);
    namep (np) -> next_node = NULL;
    return(namep(np));
}

NODEPTR int_node(val)
int val;
{
    register NODEPTR ip = claim_node();
    intp (ip) -> type = INT;
    intp (ip) -> ival = val;
    return (ip);
}

NODEPTR string_node (val)
char *val;
{
    register NODEPTR sp = claim_node();
    stringp (sp) -> type = STRING;
    stringp (sp) -> str = val;
    return (sp);
}

NODEPTR env_node (level,size,id)
unsigned level;
int size;
NAMEPTR id;
{
    register NODEPTR ep;

    ep = claim_node();
    envp (ep) -> type = ENV;
    envp (ep) -> env_level = level;
    envp (ep) -> env_size = size;
    envp (ep) -> id_list = id;
    return(ep);
}

NODEPTR bnd_node (env,blk)
ENVPTR env;
BLOCKPTR blk;
{
    register NODEPTR bp;

    bp = claim_node();
    bndp (bp) -> type = BND;
    bndp (bp) -> env = env;
    bndp (bp) -> blk = blk;
    return(bp);
}

NODEPTR list_node(blk,size_loc,base_loc)
BLOCKPTR blk;
MEMORYPTR size_loc,base_loc;
{
    register NODEPTR lp;

    lp = claim_node();
    listp (lp) -> type = LIST;
    listp (lp) -> blk = blk;
    listp (lp) -> size = size_loc;
    listp (lp) -> base = base_loc;
    return(lp);
}

BLOCKPTR block_node(size,base_loc)
unsigned int size;
MEMORYPTR base_loc;
{
    register BLOCKPTR bp;

    bp = blockp (claim_node());
    bp -> type = MEMBLOCK;
    bp -> blk_size = size;
    bp -> blk_base = base_loc;
    return(bp);
}

unsigned int elist_size(elist)
NODE2PTR elist;
{
    register unsigned int     size = 0;
    while (elist != NULL) {
        size += 1;
        elist = elist -> c2;
    }
    return (size);
}

/* end of store.c */
