/* system.c: system routines
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: system.c,v 2.11 2007/05/03 09:48:45 acau Exp $
 */


#include "system.h"



#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#else /* HAVE_SYS_WAIT_H */
#if !defined(WNOHANG) /* && (defined(bsd43) || defined(vax)) */
#define WNOHANG 1
#endif
#ifndef WEXITSTATUS
#define WEXITSTATUS(X) ((unsigned)(X) >> 8)
#endif
#ifndef WIFEXITED
#define WIFEXITED(X) (((X) & 255) == 0)
#endif
#ifndef WIFSIGNALED
#define WIFSIGNALED(X) ((((X) & 255) != 0x255 && ((X) & 255) != 0))
#endif
#ifndef WTERMSIG
#define WTERMSIG(X) ((X) & 255)
#endif
#endif /* HAVE_SYS_WAIT_H */

/* For profiling on SUN */
/* char *strrchr(a,b)
char *a,b;
{return (char *) rindex(a,b);}
*/
/****************************\
*                            *
*  System specific routines  *
*                            *
\****************************/


/******************\
*                  *
*  Error handling  *
*                  *
\******************/
RETSIGTYPE onintr(sig)
int sig;
{
    signal (SIGINT, onintr);
    InterruptSet = T;
/*    user_break ("interrupt");*/
}

RETSIGTYPE onfpe(sig)
int sig;
{
    signal (SIGFPE, onfpe);
    error_break0 ("floating point exception");
}

/* SIGBUS for djgpp */
#ifdef SIGBUS
RETSIGTYPE onbus(sig)
int sig;
{
    signal (SIGBUS, onbus);
    error_break0 ("bus error");
}
#endif

RETSIGTYPE onsegv(sig)
int sig;
{
    signal (SIGSEGV, onsegv);
    error_break0 ("segmentation violation");
}
/*  SIGSYS (for bad system calls) not available on Linux?  BCM, 4Dec96 */
#ifdef SIGSYS
RETSIGTYPE onsys(sig)
int sig;
{
    signal (SIGSYS, onsys);
    error_break0 ("bad argument to system call");
}
#endif

#ifdef SIGALRM
RETSIGTYPE onalrm(sig)
int sig;
{
    signal (SIGALRM, onalrm);
    AlarmGone = T;
}
#endif

void catch_signals()
{
    if (signal (SIGINT, SIG_IGN) != SIG_IGN)
        signal (SIGINT, onintr);
    if (signal (SIGFPE, SIG_IGN) != SIG_IGN)
        signal (SIGFPE, onfpe);
/* SIGBUS for djgpp */
#ifdef SIGBUS
    if (signal (SIGBUS, SIG_IGN) != SIG_IGN)
        signal (SIGBUS, onbus);
#endif    
    if (signal (SIGSEGV, SIG_IGN) != SIG_IGN)
        signal (SIGSEGV, onsegv);
/*  SIGSYS (for bad system calls) not available on Linux?  BCM, 4Dec96 */
#ifdef SIGSYS    
    if (signal (SIGSYS, SIG_IGN) != SIG_IGN)
        signal (SIGSYS, onsys);
#endif
#ifdef SIGALRM    
    if (signal (SIGALRM, SIG_IGN) != SIG_IGN)
        signal (SIGALRM, onalrm);
#endif
/*    intrinit (onintr, INTERRUPTSTACK, CNTRL_C); FOR PC */
}

void breakin()
{
    int c,val;
    InterruptSet = F;
    fputs ("   Abort (a), Break (b) or Continue (c)?\n ", stdout);
    fflush(stdout);
    while (isspace (c = getchar ()));
    while (T) {
        while (getchar () != '\n');
                                /* Discard up to newline. */
        switch (c) {
            case 'a':
            case 'A':
                putchar ('\n');
                TempuraLevel = 0;
                restore_state (BreakPoint,val);
            case 'b':
            case 'B':
                putchar ('\n');
                tempura (++TempuraLevel);
                return;
            case 'c':
            case 'C':
                putchar ('\n');
                return;
            default:
                fputs ("   Please type a, b or c: ", stdout);
                fflush(stdout);
        }
        c = getchar ();
    }
}



/******************\
*                  *
*   File handling  *
*                  *
\******************/
/* Modified by 05/09/2001 to handle absolute paths better */
FILE *openfile(fullname,spath,name,suffix,mode)
char *fullname, *spath, *name, *suffix, *mode;
{
    char   *p ;
    FILE   *fp, *fopen ();
    free (allocate (1000, 1));  /* check enough space for buffer */
    if ( name[0]==FILENAMESEP ) {  /* check for absolute path.  BCM 28Jul97 */
      strcpy(fullname,name);
      strcat(fullname,suffix);
      fp = fopen (fullname, mode);
    }
    else {
        do {
            *(p = fullname) = '\0';
            while (*spath != '\0' && !fisspace ((int)*spath)) {
                *p++ = *spath++;
                *p = FILENAMESEP;
            }
            *++p = '\0';
            while (fisspace ((int)*spath))
                spath++;
            strcat (fullname, name);
            strcat (fullname, suffix);
        } while (((fp = fopen (fullname, mode)) == NULL) && (*spath != '\0'));
    };
    return (fp);
}

int fisspace (s)
int s;
{
    if (s == ';') return(1);
    else return(0);
}




/*******************************\
*                               *
*  Initialise system variables  *
*                               *
\*******************************/

void assign_sysvar(id,val)
SYMBOLPTR id;
NODEPTR val;
{
    switch (id -> type) {
	case PASSLIMIT:
            PassLimit = intp (val) -> ival;
            break;
	case VERBOSITY:
            Verbosity = intp (val) -> ival;
            break;
	case PRINTLEVEL:
            PrintLevel = intp (val) -> ival;
            break;
	case DEBUGLEVEL:
            DebugLevel = intp (val) -> ival;
            break;
	case STATEPRINT:
            StatePrintFlag = boolp (val) -> bval;
            break;
	case REDUNDANCY:
            RedundancyFlag = boolp (val) -> bval;
            break;
	case CALLBYNAME:
            CallByNameFlag = boolp (val) -> bval;
            break;
	case ASSIGNAHEAD:
            AssignAheadFlag = boolp (val) -> bval;
            break;
	case PATH:
            /* Next line changed to avoid garbarge collection problems
	       BCM 14Dec96 */
            Path = stringp(val);
                /* Path = stringp (val) -> str;*/
            break;
	case HELPDIR:
            /* Next line changed to avoid garbarge collection problems
	       BCM 14Dec96 */
            HelpDir = stringp(val);
                /* HelpDir = stringp (val) -> str;*/
	    break;
	case TIMESTEP:
            TimeStep = intp (val) -> ival;
            break;
        case RANDSEED:      /* Added AC 29APR 1999 */
            RandSeed = intp (val) -> ival;
	    srand((unsigned int) RandSeed);  /* srand wants unsigned integer */
	    RandomVal = -1;       /* Reset random.    */
	    RandomValTime = 0;
            break;
        case RRANDSEED:      /* Added AC 20/09/2000 */
            RRandSeed = intp (val) -> ival;
	    srand((unsigned int) RRandSeed);  /* srand wants unsigned integer */
	    RRandomVal = -1;       /* Reset Random.    */
	    RRandomValTime = 0;
            break;
	case MAXRANDLEN:    /* Added AC 29APR 1999  */
	    if ( (intp (val) -> ival) > 0)
	      MaxRandLen = intp (val) -> ival;
	    else {
	      printf("\n***Maximum random length must be >= 1.\n");
              fflush(stdout);
            }            
            break;    
    }
}

void init_sysvars()
{
    /* The following was changed to avoid garbage collection problems.
       BCM 14Dec96 */
    Path = stringp(
      string_node(strsave (getenv ("TPATH") == NULL ? "" : getenv ("TPATH"))));
    /* Old version:
    Path = strsave (getenv ("TPATH") == NULL ? "" : getenv ("TPATH"));
    */
    /* Following added by BCM, 4Dec96, 14Dec96 */
    HelpDir = stringp(
      string_node(strsave (getenv ("THELPDIR") == NULL ? "help"
		       : getenv ("THELPDIR"))));
}

void start_time()
{
    AlarmGone = F;
#ifdef SIGALRM
    alarm ((unsigned)TimeStep);
#endif
}


void check_time()
{   
#ifdef SIGALRM
    if (!AlarmGone) 
        pause ();
    else
        warn0 ("falling behind (real) time");
#endif
}


int own_wait(statusPtr)
int *statusPtr;
{
#ifdef HAVE_WORKING_FORK
#ifdef NO_WAIT3
    return(waitpid (-1,statusPtr,WNOHANG));
#else
    return(wait3 (statusPtr,WNOHANG, 0));
#endif
#endif
}


void check_sys_waits ()
{
    int pid;
    /*AC: the wait problem
      WAIT_STATUS_TYPE s;  
      WAIT_STATUS_TYPE status = &s */
    WAIT_STATUS_TYPE status; 
    /*int *statusp;*/
        /* waitpid (-1,&status, WNOHANG) for those
           that haven't got wait3 (&status, WNOHANG, 0)*/
#ifdef HAVE_WORKING_FORK
    while ((pid = own_wait (&status)) != -1 && pid != 0) {
        SysWaitList = node2 (iSYSTEM, int_node (pid), SysWaitList);
	assert2d (WIFEXITED(status), 
	         "system call %d terminated with non-zero status: %d",
		 pid, WEXITSTATUS(status) ); 
    }
#endif
}
