/* tempura.h:  tempura definitions
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: tempura.h,v 2.12 2007/05/03 09:48:45 acau Exp $
 */

/********************\
*                    *
*  Type definitions  *
*                    *
\********************/

typedef struct symbol_table_entry {
    unsigned mark : 1;		/* Don't need this */
    unsigned char type;
    char *name;
    struct symbol_table_entry *next;
} SYMBOL, *SYMBOLPTR;
#define symbolp(X) ((SYMBOLPTR) (X))

typedef struct node {
    unsigned mark : 1;
    unsigned char type;
    struct node *next_node;
    struct node *c1;
    struct node *c2;
    struct node *c3;
} NODE0, *NODE0PTR, NODE1, *NODE1PTR, NODE2, *NODE2PTR, NODE3, *NODE3PTR,
  NODE, *NODEPTR;
/*  NODE, *NODEPTR, *REDUCESTACK, **REDUCESTACKPTR;*/
#define nodep(X) ((NODEPTR) (X))

typedef union symbol_or_node {
    SYMBOLPTR s;
    NODEPTR n;
} SYMBOL_OR_NODEPTR;

typedef struct memory {
    unsigned mark : 1;
    unsigned mem_no : 1;
    unsigned is_static : 1;
    unsigned is_indirect : 1;
    int time;
/*    struct node *val;*/
    union {
	NODEPTR n;
	struct memory * m;
    } val;
} MEMORY, *MEMORYPTR;
#define memoryp(X) ((MEMORYPTR) (X))

typedef struct b_node {
    unsigned mark : 1;
    unsigned char type;
    struct node *next_node;
    int bval;
} *BOOLPTR;
#define boolp(X) ((BOOLPTR) (X))

typedef struct i_node {
    unsigned mark : 1;
    unsigned char type;
    struct node *next_node;
    int ival;
} *INTPTR;
#define intp(X) ((INTPTR) (X))

typedef struct s_node {
    unsigned mark : 1;
    unsigned char type;
    struct node *next_node;
    char *str;
} *STRINGPTR;
#define stringp(X) ((STRINGPTR) (X))

typedef struct n_node {
    unsigned mark : 1;
    unsigned char type;
    struct n_node *next_node;
    char *name;
    int level;
    int off;
/*    struct n_node *next_name;*/
} *NAMEPTR;
#define namep(X) ((NAMEPTR) (X))

typedef struct e_node {
    unsigned mark : 1;
    unsigned char type;
    struct node *next_node;
    unsigned env_level;
    unsigned int env_size;
    NAMEPTR id_list;
} *ENVPTR;
#define envp(X) ((ENVPTR) (X))

typedef struct loc_node {
    unsigned mark : 1;
    unsigned char type;
    struct node *next_node;
    MEMORYPTR ptr;
    unsigned to_next;
} *LOCPTR;
#define locp(X) ((LOCPTR) (X))

typedef struct memory_block {
    unsigned mark : 1;
    unsigned char type;
    struct node *next_node;
    unsigned int blk_size;
    MEMORYPTR blk_base;
} *BLOCKPTR, BLOCK;
#define blockp(X) ((BLOCKPTR) (X))

typedef struct bnd_node {
    unsigned mark : 1;
    unsigned char type;
    struct node *next_node;
    ENVPTR env;
    BLOCKPTR blk;
} *BNDPTR;
#define bndp(X) ((BNDPTR) (X))

typedef struct list_node {
    unsigned mark : 1;
    unsigned char type;
    struct node *next_node;
    BLOCKPTR blk;
    MEMORYPTR size;
    MEMORYPTR base;
} *LISTPTR;
#define listp(X) ((LISTPTR) (X))

typedef void (*PFI)();
/*typedef NODEPTR (*PFN)();*/

typedef struct heap_block {
    struct heap_block *next_block;
    struct node *base;
    struct node *top;
} HEAPNODE, *HEAP;
#define heapp(X) ((HEAP) (X))

typedef struct es_item {
    MEMORYPTR env_base;
    ENVPTR env;
} ENVSTACK, *ENVSTACKPTR;

typedef union memory_or_node {
    MEMORYPTR m;
    NODEPTR n;
} MEMORY_OR_NODEPTR, REDUCESTACK, *REDUCESTACKPTR;

#define save_state(BP) \
        int old_env_base = EnvBase;\
        int old_env_top = EnvTop;\
        REDUCESTACKPTR old_reduce_stack_top = ReduceStackTop;\
        int old_reduce_depth = ReduceDepth;\
        MEMORYPTR old_empty_loc = EmptyLoc;\
        unsigned old_next_flag = NextFlag;\
        unsigned old_assign_ahead_flag = AssignAheadFlag;\
        unsigned old_run_flag = RunFlag;\
        jmp_buf break_point;\
        BP = &break_point;\
        setjmp(break_point);\
        EnvBase = old_env_base;\
        EnvTop = old_env_top;\
        ReduceStackTop = old_reduce_stack_top;\
        ReduceDepth = old_reduce_depth;\
        EmptyLoc = old_empty_loc;\
        NextFlag = old_next_flag;\
        AssignAheadFlag = old_assign_ahead_flag;\
        RunFlag = old_run_flag

#define restore_state(BP,val) longjmp(*(BP),val)

#define F 0
#define T 1
#define UNDEF -1

#define VAL 1
#define REF 2
#define STMT 3

#define syntax_error0(S) syntax_error(S,NULL)
#define syntax_error1(S,X) syntax_error(S,X)

#define error_break1(S,X) error_break2(S,X,NULL)
#define error_break1d(S,X) error_break2e(S,X)
#define error_break0(S) error_break2(S,NULL,NULL)

#define warn0(S) warn1(S,NULL)

#define assert0(C,S) if(!(C)) error_break0(S)
#define assert1(C,S,X) if(!(C)) error_break1(S,X)
#define assert2(C,S,X,Y) if(!(C)) error_break2(S,X,Y)
#define assert2d(C,S,X,Y) if(!(C)) error_break2d(S,X,Y)

#define result(E) ReduceStackTop->n=(E)
#define ref_result(E) ReduceStackTop->m=(E) 

#define node0(U) node(U,NULL,NULL,NULL)
#define node1(U,V) node(U,V,NULL,NULL)
#define node2(U,V,W) node(U,V,W,NULL)
#define node3(U,V,W,X) node(U,V,W,X)
#define conj(U,V) node3(iAND,False,U,node3(iAND,False,V,NULL))
#define conjtf(U,V) node3(iAND,True,U,node3(iAND,False,V,NULL))
#define assign(L,V) node2(EQ,L,copy_val(V))

#define max(A,B) (((A) > (B)) ? (A) : (B))
#define min(A,B) (((A) < (B)) ? (A) : (B))

#define mod(A,B) (A<0?(A+((B-A)/B)*B)%B:A%B)

#define MEMCELLSIZE 2
#define switch_memory CurrBit = CurrBit^1;NextBit = NextBit^1
/*#define switch_memory CurrBit = 1-CurrBit;NextBit = 1-NextBit*/
#define zero_loc(X) ((X)-(X)->mem_no)
#define curr_loc0(X) ((X)+CurrBit)
#define curr_loc(X) ((X)+(int)(CurrBit-((X)->mem_no)))
#define next_loc(X) ((X)+(int)(NextBit-((X)->mem_no)))
#define is_curr(X) ((X)->mem_no ^ NextBit)
#define is_next(X) ((X)->mem_no ^ CurrBit)
#define adj_loc(X) ((NextFlag)?next_loc(X):curr_loc(X))
#define offset(X) ((X)<<1)
#define cell_no(X) (X)
#define id_list_size(X) (((X)==NULL)?0:(X)->offset+1)
#define mem_loc(X) (EnvStack[EnvBase+namep(X)->level].env_base+namep(X)->off)
#define is_indirect_mem(X) ((X)->is_indirect)
/*#define dereference(L) while (is_indirect_mem (L)) (L) = memoryp((L)->val)*/
#define dereference(L) while (is_indirect_mem (L)) (L) = (L)->val.m
#define local_mem(A) mem_block(A,1)
#define GlobalMemUsed EnvStack[EnvBase].env -> env_size

#define get_int(X) (intp(get_mem(X)) -> ival)
#define get_list(X) listp(get_mem(X))
#define put_list(L,X) listp(put_mem(L,X))

#define list_size(L) listp(L)->size
#define list_base(L) listp(L)->base

#define is_identifier(X) (((X)->type == STATE)||((X)->type == STATIC))
#define is_boolean(X) ((X)->type == BOOL)
#define is_integer(X) ((X)->type == NAT)
#define is_string(X) ((X)->type == STRING)
#define is_lambda(X) ((X)->type == LAMBDA)
#define is_list(X) ((X)->type == LIST)

#define curr_val(X) is_list(X) ? curr_list(listp(X)) : (X)
#define next_val(X) is_list(X) ? next_list(listp(X)) : (X)

#define pass_verbosity (Verbosity > 3)
#define state_verbosity (Verbosity > 2)
#define store_verbosity (Verbosity > 1)
#define exec_verbosity (Verbosity > 0)
#define state_debug DebugLevel == 2
#define reduce_debug DebugLevel > 2

#define output_char(C) {putchar(C);CharsOutput++;}
#define space output_char(' ')
#define newline output_char('\n')
#define output_state_number if (RunFlag && StatePrintFlag) { \
                                printf("State %3d: ",StateNumber); \
                                CharsOutput += 11;}

#define push_empty(X) MEMORYPTR old_empty_loc=EmptyLoc;EmptyLoc=curr_loc(X)
#define pop_empty EmptyLoc=old_empty_loc

#define always(EXP) \
{\
    NODEPTR e1 = EXP;\
    if (e1 != True) {\
        if (Ready)\
            result (conj (e1, e));\
        else\
            result (conjtf (e, e1));\
    }\
}

#define keep(EXP) \
{\
    if ((Ready = empty_defined ())) {\
        if (get_empty () == True)\
            result (True);\
        else {\
            NODEPTR e1 = EXP;\
            if (e1 != True) {\
                if (Ready)\
                    result (conj (e1, e));\
	        else\
                    result (conjtf (e, e1));\
	    }\
        }\
    }\
}

#define push_assign_ahead unsigned old_assign_ahead_flag = AssignAheadFlag;\
        AssignAheadFlag = F;
#define pop_assign_ahead AssignAheadFlag = old_assign_ahead_flag;
