/* default.h:  SET RUN TIME DEFAULT VALUES
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: top-lev.c,v 2.11 2007/05/03 09:48:45 acau Exp $
 * 									 
 *  History:								 
 * 									 
 *   The first Tempura interpreter was programmed in Prolog by Ben	 
 *   Moszkowski, and was operational around December 2, 1983.  Subsequently
 *   he rewrote the interpreter in Lisp (mid Mar, 1984), and in late 1984  
 *   modified the program to handle a two-level memory and multi-pass	   
 *   scanning. The C-Tempura interpreter was written in early 1985 by	   
 *   Roger Hale at Cambridge University.				  
 * 									  
 * Some known bugs to fix:						  
 *   "struct" is rather flaky, but believed to be OK if used as		  
 *   "stable(struct(L))".
 */


#include "system.h"



/*************************************\
* 				      *
* top-level read-eval-print routines  *
* 				      *
\*************************************/


/* C top level */
int main ()
{
    tempura_init ();
/*    printf ("PC-Tempura (Version ");*/
    printf("Tempura: Interval Temporal Logic interpreter\n\n");
    fflush(stdout);
    printf("Copyright (C) 1987  Roger Hale\n\n");
    fflush(stdout);
    printf("This program is free software; you can redistribute it and/or modify\n");
    fflush(stdout);
    printf("it under the terms of the GNU General Public License as published by\n");
    fflush(stdout);
    printf("the Free Software Foundation; either version 2 of the License, or\n");
    fflush(stdout);
    printf("(at your option) any later version.\n\n");
    fflush(stdout);
    printf("This program is distributed in the hope that it will be useful,\n");
    fflush(stdout);
    printf("but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
    fflush(stdout);
    printf("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
    fflush(stdout);
    printf("GNU General Public License for more details.\n\n");
    fflush(stdout);
    printf("You should have received a copy of the GNU General Public License\n");
    fflush(stdout);
    printf("along with this program; if not, write to the Free Software\n");
    fflush(stdout);
    printf("Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n\n");
    fflush(stdout);
    printf("Current maintainers:\n");
    fflush(stdout);
    printf("  Antonio Cau and Ben Moszkowski, (tempura@dmu.ac.uk)\n\n");
    fflush(stdout);
    printf ("C-Tempura (Version ");
    fflush(stdout);
    print_version_number();
    printf(")\n\n");
    fflush(stdout);

    tempura (TempuraLevel = 0);
    exit(0);
}

void prompt (level)
int level;
{
    int     i;
    printf ("Tempura %d", Counter++);
    for (i = 0; i <= level; i++)
    putchar ('%');
    fputs (" ", stdout);
    newline;
    fflush(stdout);
}

/* main read_run loop */
void tempura (level)
int level;
{   
    int val;
    jmp_buf *OldBreakPoint= BreakPoint; 
    save_state (BreakPoint);
    while (TempuraLevel >= level) {
        prompt (level);
        exec_command (LastCommand = parse (stdin));
    }
    BreakPoint = OldBreakPoint;
    if (Break)
        restore_state (BreakPoint,val);
}


/***********************\
*                      *
*  command interpreter *
*                      *
\***********************/

void exec_command (cmd)
NODEPTR cmd;
{
    BreakDepth = -1;
    cmd = reduce_stmt (cmd);
    assert0 (cmd != False, "the command is false!");
    assert0 (Ready, "unable to evaluate top-level command in one pass");
}

/*********************\
*                    *
* reduction routines *
* 	             *
\*********************/

#define push_expn(E)\
    if (++ReduceDepth > MAXREDUCESTACKSIZE) {\
        ReduceStackTop -> n = E;\
        error_break0 ("reduction stack full");\
    }\
    else {\
        (++ReduceStackTop) -> n = E;\
        MaxReduceDepth = max (MaxReduceDepth, ReduceDepth);\
    }\
    PassReduceCount += 1

#define pop_expn (ReduceDepth--,(ReduceStackTop--) -> n)
#define pop_ref (ReduceDepth--,(ReduceStackTop--) -> m)

#define applystmt(R) (*(R[ReduceStackTop -> n -> type][0]))(ReduceStackTop -> n)
#define applyval(R) (*(R[ReduceStackTop -> n -> type][1]))(ReduceStackTop -> n)
#define applyref(R) (*(R[ReduceStackTop -> n -> type][2]))(ReduceStackTop -> n)

/* reduce a formula */
NODEPTR reduce_stmt (e)
NODEPTR e;
{
    push_expn(e);
    TempNodeList = NULL;  /* This was moved from within loop. BCM 25Jul97 */
    do {
        if (InterruptSet)
            user_break ("interrupt");
        Retry = F;
        if (ReduceDepth <= BreakDepth) {
            show_form ();
            tempura (++TempuraLevel);
        }
            /* TempNodeList = NULL;*/
        if (DebugLevel>=5) {  /* DEBUG  BCM 29Jul97 */
	  printf("reduce_stmt 1: ");
	  output_expn(ReduceStackTop -> n,PrintLevel);
	  printf("\n");
          fflush(stdout);
	};
        
        applystmt (Reduces);
        
        if (DebugLevel>=5) {  /* DEBUG  BCM 29Jul97 */
	  printf("reduce_stmt 2, Rdy=%1d, Rtry=%1d: ",Ready,Retry);
	  output_expn(ReduceStackTop -> n,PrintLevel);
	  printf("\n");
          fflush(stdout);
	};
        
    } while (Retry);

    if (Ready) {
        PassSuccessfulReduceCount += 1;
        PassProgressFlag = T;
    }
    else if (PassLimitCounter == PassLimit) {
        PassLimitCounter = 0;
	state_not_defined ();
    }
    else
        if (!StateProgressFlag)
	    state_not_defined ();

    return (pop_expn);
}

/* reduce a value */
NODEPTR reduce_val (e)
NODEPTR e;
{
    push_expn (e);
    do {
        Retry = F;
        if (DebugLevel>=5) {  /* DEBUG  BCM 29Jul97 */
	  printf("reduce_val 1: ");
	  output_expn(ReduceStackTop -> n,PrintLevel);
	  printf("\n");
          fflush(stdout);
	};
        applyval (Reduces);
        
        if (DebugLevel>=5) {  /* DEBUG  BCM 29Jul97 */
	  printf("reduce_val 2, Rdy=%1d, Rtry=%1d: ",Ready,Retry);
	  output_expn(ReduceStackTop -> n,PrintLevel);
	  printf("\n");
          fflush(stdout);
	};
    } while (Retry);

    if (Ready)
        PassSuccessfulReduceCount += 1;
    return (pop_expn);
}

/* reduce a reference */
MEMORYPTR reduce_ref (e)
NODEPTR e;
{
    push_expn (e);
    do {
        Retry = F;
        if (DebugLevel>=5) {  /* DEBUG  BCM 29Jul97 */
	  printf("reduce_ref 1: ");
	  output_expn(ReduceStackTop -> n,PrintLevel);
	  printf("\n");
          fflush(stdout);
	};
        
        applyref (Reduces);
        if (DebugLevel>=5) {  /* DEBUG  BCM 29Jul97 */
	  printf("reduce_ref 2, Rdy=%1d, Rtry=%1d: ",Ready,Retry);
	  output_expn(ReduceStackTop -> n,PrintLevel);
	  printf("\n");
          fflush(stdout);
	};
        
    } while (Retry);

    if (Ready)
        PassSuccessfulReduceCount += 1;
    return (pop_ref);
}



/***********************************************\
* 	                                       *
*  reduce a subprogram with its own empty-flag *
* 		                               *
\***********************************************/


NODEPTR reduce_pgm (e)
NODEPTR e;
{
    e = reduce_stmt (e);
    if (Ready) {
        assert0 (e != False, "the program is false!");
        assert0 (empty_defined (), "the interval length is undefined");
    }
    return (e);
}

/*******************************************\
*                                           *
*  Reduce a reference and fetch its value.  *
* 		                            *
\*******************************************/
NODEPTR reduce_ref_and_fetch(e)
NODEPTR e;
{
    MEMORYPTR loc = reduce_ref (e);
    
    if (Ready) {
        if ((Ready = mem_defined (loc)))
            e = get_mem (loc);
    }
    return (e);
}

/* Convert a value to a reference. */
MEMORYPTR make_ref (e)
NODEPTR e;
{
    TempLoc = adj_loc (TempLoc);
    put_mem (TempLoc, e);
    return (TempLoc);
}

void get_val(e)
NODEPTR e;
{
    NODEPTR val = reduce_ref_and_fetch (e);
    if (Ready) {
        if ((Ready = val_defined (val))) 
            result (val);
    }
}

void get_ref(e)
NODEPTR e;
{
    NODEPTR val = reduce_val (e);
    if (Ready)
        ref_result (make_ref (val));
}



/* end of top-level.c */

